/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#define PILOT_PRECOMPILED_HEADERS_OFF
#include "ROM Transfer.h"
#include <SysEvtMgr.h>


#define kXmodemBlockSize	1024
#define kXmodemSoh			'\002'
#define kXmodemEof			'\004'
#define kXmodemAck			'\006'
#define kXmodemNak			'\025'
#define kXmodemCan			'\030'
#define kXmodemNakCrc		'C'


static char*	GetROMStart (UInt32 aROMVersion);
static UInt32	GetROMSize (UInt32 aROMVersion);

// ---------------------------------------------------------------------------
//		 PilotMain
// ---------------------------------------------------------------------------
// The main entry point for ROM Transfer.

UInt32 PilotMain (	UInt16		iCommand,
					Ptr			iCommandParams,
					UInt16		iLaunchFlags)
{
	#pragma unused (iCommandParams)
	#pragma unused (iLaunchFlags)

	if (iCommand != sysAppLaunchCmdNormalLaunch)
		return 0;

	ErrTry
	{
		InitializeApplication ();
		ExecuteApplication ();
	}

	// Catch any unhandled exceptions.

	ErrCatch (iError)
	{
	}
	ErrEndCatch

	// Clean up.

	DisposeApplication ();

	return 0;
}


#pragma mark -


// ---------------------------------------------------------------------------
//		 InitializeApplication
// ---------------------------------------------------------------------------
// Initializes the application globals and prepares to execute the event loop.

void InitializeApplication (void)
{
	CharPtr			aLabel;
	ListPtr			aList;
	Word			aListItem;
	ControlPtr		aControl;

	// Initialize and draw the main form.

	FormPtr aMainForm = FrmInitForm (MainForm);
	Throw_IfNil (aMainForm, 0);

	FrmSetActiveForm (aMainForm);
	FrmDrawForm	(aMainForm);

	// Set up the speed popup.

	aListItem = kItem_115200;
	aList = FrmGetObjectPtr (aMainForm, FrmGetObjectIndex (aMainForm, MainSpeedList));
	aLabel = LstGetSelectionText (aList, aListItem);
	aControl = FrmGetObjectPtr (aMainForm, FrmGetObjectIndex (aMainForm, MainSpeedPopTrigger));
	CtlSetLabel (aControl, aLabel);
	LstSetSelection (aList, aListItem);
}


// ---------------------------------------------------------------------------
//		 DisposeApplication
// ---------------------------------------------------------------------------
// Disposes the application globals and prepares to exit.

void DisposeApplication (void)
{
}


// ---------------------------------------------------------------------------
//		 ExecuteApplication
// ---------------------------------------------------------------------------
// Executes an event loop until the application quits.

void ExecuteApplication (void)
{
	EventType aEvent;

	aEvent.eType = nilEvent;

	// Repeat until the application quits.

	while (aEvent.eType != appStopEvent)
	{
		// Get the next available event.

		EvtGetEvent (&aEvent, evtWaitForever);

		// Give the system a chance to handle the event.

		if (!SysHandleEvent (&aEvent))

			// Try to handle the event ourselves.

			if (!ProcessEvent (&aEvent))

				// Let the form provide default handling of the event.

				FrmHandleEvent (FrmGetActiveForm (), &aEvent);
	}
}


// ---------------------------------------------------------------------------
//		 ProcessEvent
// ---------------------------------------------------------------------------
// Attempts to process the specified event.

Boolean ProcessEvent (EventPtr iEvent)
{
	Boolean		aWasHandled = false;

	if (iEvent->eType == ctlSelectEvent)
	{
		switch (iEvent->data.ctlEnter.controlID)
		{
			case MainBeginTransferButton:

			TransferROM ();
	   		aWasHandled = true;
	   		break;
		}
	}

	return aWasHandled;
}


// ---------------------------------------------------------------------------
//		 TransferROM
// ---------------------------------------------------------------------------
// Downloads the ROM via the serial connection.

void TransferROM (void)
{
	Int16	aError;
	UInt32	aROMVersion;
	UInt32	aHeaderCode;
	Int32	aOffset;
	char*	aROMStart;
	Int32	aROMSize;
	Word	aROMChecksum;

	Int16	aSpeedChoice;
	ULong	aSpeed;
	FormPtr aForm;

	Byte	c;
	Byte	block;
	UInt32	i;
	Boolean	bXmodemCrc;
	Byte	header[kXmodemBlockSize];
	Boolean bXmodemHeaderBlock;
	
	aForm = FrmGetActiveForm ();

	ErrTry
	{
		Throw_IfNil (aForm, 0);

		aSpeedChoice = LstGetSelection (FrmGetObjectPtr (aForm, FrmGetObjectIndex (aForm, MainSpeedList)));
		switch (aSpeedChoice)
		{
			case kItem_1200: aSpeed = kBaud_1200; break;
			case kItem_2400: aSpeed = kBaud_2400; break;
			case kItem_4800: aSpeed = kBaud_4800; break;
			case kItem_9600: aSpeed = kBaud_9600; break;
			case kItem_14400: aSpeed = kBaud_14400; break;
			case kItem_19200: aSpeed = kBaud_19200; break;
			case kItem_28800: aSpeed = kBaud_28800; break;
			case kItem_38400: aSpeed = kBaud_38400; break;
			case kItem_57600: aSpeed = kBaud_57600; break;
			case kItem_115200: aSpeed = kBaud_115200; break;
			default: aSpeed = kBaud_14400; break;
		}

		aError = Comm_Initialize (kPort_Pilot, aSpeed);
		Throw_IfError (aError, aError);

		// Determine which ROM we're transferring.

		FtrGet (sysFtrCreator, sysFtrNumROMVersion, &aROMVersion);

		// Tell the Emulator that we're going to start.

		aHeaderCode		= 0xDEADC0DE;
		aROMStart		= GetROMStart (aROMVersion);
		aROMSize		= GetROMSize (aROMVersion);
		aROMChecksum	= Crc16CalcBlock (aROMStart, aROMSize, 0);

		// Change the status text.

		FrmCopyLabel (aForm, MainStatusTextLabel, "Waiting...          ");

		// First, wait for the receiver to send us a nak. If it's
		// a real nak, then we'll send in checksum mode. If it's
		// a 'C', we'll send in CRC mode.

		Comm_ClearError();
		bXmodemCrc = false;
		while (true)
		{
			aError = SerReceive10(gPortReference, (void*) &c, 1, 30*sysTicksPerSecond);
			Throw_IfError(aError, aError);
			if (c == kXmodemNakCrc)
			{
				bXmodemCrc = true;
				break;
			}
			else if (c == kXmodemNak)
			{
				break;
			}
			else if (c == kXmodemCan)
			{
				ErrThrow(c);
			}
		}

		// Build the Ymodem header block.

		MemSet(header, kXmodemBlockSize, 0);
		StrCopy((char*) header, "PalmOS.ROM"); // What's in a name?
		StrIToA((char*) &header[StrLen((char*) header)+1], aROMSize);

		// Send the ROM one block at a time until it's done.

		aOffset	= 0;
		block	= 0;
		bXmodemHeaderBlock = true;

		while (aOffset < aROMSize)
		{
			Byte *blockStart;

			EvtResetAutoOffTimer ();

			// Change the status text.
			
			{
				DWord	percentComplete = (aOffset * 100) / (aROMSize - kXmodemBlockSize);
				char	buffer1[100];
				char	buffer2[20];

				StrCopy (buffer1, "Sending...");
				StrIToA (buffer2, percentComplete);
				StrCat (buffer1, buffer2);
				StrCat (buffer1, "%          ");

				FrmCopyLabel (aForm, MainStatusTextLabel, buffer1);
			}

			// Send each part of the block individually.

			c = kXmodemSoh;
			SerSend10(gPortReference, (void*) &c, 1);
			SerSend10(gPortReference, (void*) &block, 1);
			c = block ^ 0xFF;
			SerSend10(gPortReference, (void*) &c, 1);

			if (bXmodemHeaderBlock)
			{
				blockStart = header;
			}
			else
			{
				blockStart = (Byte*) (aROMStart + aOffset);
			}
			aError = SerSend10 (gPortReference, (void*) blockStart, kXmodemBlockSize);

			if (bXmodemCrc)
			{
				UInt16 crc = Crc16CalcBlock (blockStart, kXmodemBlockSize, 0);
				
				Byte	zeros[2] = {0, 0};
				crc = Crc16CalcBlock (zeros, 2, crc);

				c = crc >> 8;
				aError = SerSend10 (gPortReference, (void*) &c, 1);
				c = crc;
				aError = SerSend10 (gPortReference, (void*) &c, 1);
			}
			else
			{
				c = 0;
				for (i = 0; i < kXmodemBlockSize; i++)
				{
					c += blockStart[i];
				}
				aError = SerSend10 (gPortReference, (void*) &c, 1);
			}
			Throw_IfError (aError, aError);

			// Wait for the other end to ack or nak us. Leave
			// plenty of timeout so the receiver has time to
			// decide whether it got a partial block.

			aError = SerReceive10 (gPortReference, (void*) &c, 1, 20*sysTicksPerSecond);
			Throw_IfError (aError, aError);
			if (c == kXmodemNak || c == kXmodemNakCrc)
			{
				continue;
			}
			else if (c != kXmodemAck)
			{
				ErrThrow(c);
			}

			if (!bXmodemHeaderBlock)
			{
				// Only do this if we're into sending the file
				// (don't increment on the file header block).

				aOffset += kXmodemBlockSize;
			}
			block++;
			bXmodemHeaderBlock = false;
		}

		// Send an EOF and wait for an acknowledgement.

		while (true)
		{
			c = kXmodemEof;
			SerSend10(gPortReference, (void*) &c, 1);
			aError = SerReceive10(gPortReference, (void*) &c, 1, 20*sysTicksPerSecond);
			Throw_IfError (aError, aError);
			if (c == kXmodemNak)
			{
				continue;
			}
			else if (c == kXmodemAck)
			{
				break;
			}
			else
			{
				ErrThrow(c);
			}
		}

		// Send another one for good measure.

		c = kXmodemEof;
		SerSend10(gPortReference, (void*) &c, 1);

		// Change the status text.

		FrmCopyLabel (aForm, MainStatusTextLabel, "Finished               ");
	}

	// Catch any exceptions.

	ErrCatch (iError)
	{
		// Change the status text.

		if (aForm != NULL)
		{
			char	buffer1[100];
			char	buffer2[20];

			StrCopy (buffer1, "Error #");
			StrIToA (buffer2, iError);
			StrCat (buffer1, buffer2);
			StrCat (buffer1, "           ");

			FrmCopyLabel (aForm, MainStatusTextLabel, buffer1);
		}
	}
	ErrEndCatch

	Comm_Dispose ();
}


#define NON_PORTABLE
#include <MemoryPrvNew.h>	// CardHeaderType

// ---------------------------------------------------------------------------
//		 GetROMStart
// ---------------------------------------------------------------------------

char* GetROMStart (UInt32 aROMVersion)
{
	DWord	bigROMOffset;
	char*	bigROMStart;
	char*	card0base;

	if (sysGetROMVerMajor(aROMVersion) == 1)
		return (char*) 0x10C00000;

	if (sysGetROMVerMajor(aROMVersion) == 2)
		return (char*) 0x10C00000;

	card0base		= (char*) (HwrMemWritable ((VoidPtr) -2) & 0xF0000000);
	bigROMOffset	= ((CardHeaderType*) 0x10C00000)->bigROMOffset;
	bigROMStart		= card0base + bigROMOffset;

	return bigROMStart;
}


// ---------------------------------------------------------------------------
//		 GetROMSize
// ---------------------------------------------------------------------------

UInt32 GetROMSize (UInt32 aROMVersion)
{
	char*	bigROMStart;

	if (sysGetROMVerMajor(aROMVersion) == 1)
		return 512 * 1024L;

	if (sysGetROMVerMajor(aROMVersion) == 2)
		return 1024 * 1024L;

	bigROMStart = GetROMStart (aROMVersion);

	return ((CardHeaderType*) bigROMStart)->checksumBytes;
}
