/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "Bank_MC68681.h"

#include "CPU_REG.h"			// Software::BusError
#include "RAM_ROM.h"			// Memory::InitializeBanks


// ===========================================================================
//		 MC68681 Bank Accessors
// ===========================================================================
// These functions provide fetch and store access to the emulated M68681
// serial port.

AddressBank	MC68681Bank::fgAddressBank =
{
	MC68681Bank::GetLong,
	MC68681Bank::GetWord,
	MC68681Bank::GetByte,
	MC68681Bank::SetLong,
	MC68681Bank::SetWord,
	MC68681Bank::SetByte,
	MC68681Bank::GetRealAddress,
	MC68681Bank::ValidAddress
};

const uae_u32	MC68681Bank::kMemoryStart = hwrDuartBase;

HwrDuartType	MC68681Bank::fgDuart;

// ---------------------------------------------------------------------------
//		 MC68681Bank::Initialize
// ---------------------------------------------------------------------------

void MC68681Bank::Initialize (void)
{
	// Memory bank 0x10E0 is managed by the functions in MC68681Bank.

	Memory::InitializeBanks (	fgAddressBank,
								bankindex (kMemoryStart), 1);
}


// ---------------------------------------------------------------------------
//		 MC68681Bank::Reset
// ---------------------------------------------------------------------------

void MC68681Bank::Reset (void)
{
}


// ---------------------------------------------------------------------------
//		 MC68681Bank::GetLong
// ---------------------------------------------------------------------------

uae_u32 MC68681Bank::GetLong (uaecptr iAddress)
{
	InvalidAccess (iAddress, 4, true);

	return 0;
}

// ---------------------------------------------------------------------------
//		 MC68681Bank::GetWord
// ---------------------------------------------------------------------------

uae_u32 MC68681Bank::GetWord (uaecptr iAddress)
{
	InvalidAccess (iAddress, 2, true);

	return 0;
}

// ---------------------------------------------------------------------------
//		 MC68681Bank::GetByte
// ---------------------------------------------------------------------------

uae_u32 MC68681Bank::GetByte (uaecptr iAddress)
{
	return ((uae_u8*) &fgDuart)[iAddress - kMemoryStart];
}

// ---------------------------------------------------------------------------
//		 MC68681Bank::SetLong
// ---------------------------------------------------------------------------

void MC68681Bank::SetLong (uaecptr iAddress, uae_u32 iLongValue)
{
	InvalidAccess (iAddress, 4, false);
}

// ---------------------------------------------------------------------------
//		 MC68681Bank::SetWord
// ---------------------------------------------------------------------------

void MC68681Bank::SetWord (uaecptr iAddress, uae_u32 iWordValue)
{
	InvalidAccess (iAddress, 2, false);
}

// ---------------------------------------------------------------------------
//		 MC68681Bank::SetByte
// ---------------------------------------------------------------------------

void MC68681Bank::SetByte (uaecptr iAddress, uae_u32 iByteValue)
{
	((uae_u8*) &fgDuart)[iAddress - kMemoryStart] = iByteValue;
}

// ---------------------------------------------------------------------------
//		 MC68681Bank::ValidAddress
// ---------------------------------------------------------------------------

int MC68681Bank::ValidAddress (uaecptr iAddress, uae_u32 iSize)
{
	const uaecptr	offset = iAddress - kMemoryStart;

	int	result = (offset >= 0) && (offset < sizeof (fgDuart));

	assert (result);

	return result;
}

// ---------------------------------------------------------------------------
//		 MC68681Bank::GetRealAddress
// ---------------------------------------------------------------------------

uae_u8* MC68681Bank::GetRealAddress (uaecptr iAddress)
{
	return &(((uae_u8*) &fgDuart)[iAddress - kMemoryStart]);
}

// ---------------------------------------------------------------------------
//		 MC68681Bank::InvalidAccess
// ---------------------------------------------------------------------------

void MC68681Bank::InvalidAccess (uaecptr iAddress, long size, Bool forRead)
{
	Software::BusError ();
}

// ---------------------------------------------------------------------------
//		 MC68681Bank::PreventedAccess
// ---------------------------------------------------------------------------

void MC68681Bank::PreventedAccess (uaecptr iAddress, long size, Bool forRead)
{
	Software::BusError ();
}


