/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#ifndef _LOGGING_H_
#define _LOGGING_H_

#include "AppPreferences.h"		// gPrefs (for the LogFoo functions)
								// We don't need them here, but out clients likely need them.
#include "Miscellaneous.h"		// StMemory

class LogStream
{
	public:
								LogStream		(const char* baseName);
								~LogStream		(void);

		int						Printf			(const char* msg, ...);
		int						DataPrintf		(const void*, long dataLen, const char* msg, ...);
		int						VPrintf			(const char* fmt, va_list args);
		int						Write			(const void* buffer, long size);

		void					Clear			(void);

		long					GetLogSize		(void);
		void					SetLogSize		(long);

		void					EnsureNewFile	(void);
		void					DumpToFile		(void);

	private:
		FILE*					OpenFile		(void);
		void					Timestamp		(void);
		void					NewLine			(void);
		void					Append			(const char* buffer, long size);
		void					DumpToFile		(FILE*, const char*, long size);

		const char*				fBaseName;
		long					fFileIndex;

		StMemory				fBuffer;
		long					fBufferSize;
		long					fBufferOffset;

		uae_u32					fLastGremlinEventCounter;
		uae_u32					fLastTimestampTime;
		uae_u32					fBaseTimestampTime ;
		char					fLastTimestampString[30];
};

void	LogEvtAddEventToQueue		(uaecptr);
void	LogEvtAddUniqueEventToQueue	(uaecptr, DWord, Boolean);
void	LogEvtEnqueueKey			(UInt ascii, UInt keycode, UInt modifiers);
void	LogEvtEnqueuePenPoint		(uaecptr);

void	LogEvtGetEvent				(uaecptr, SDWord timeout);
void	LogEvtGetPen				(uaecptr, uaecptr, uaecptr);
void	LogEvtGetSysEvent			(uaecptr, Long timeout);

extern LogStream	gStdLog;

#define AppendMsg	gStdLog.Printf
#define AppendData	gStdLog.DataPrintf
#define ClearLog	gStdLog.Clear
#define DumpLog		gStdLog.DumpToFile
#define StartNewLog	gStdLog.DumpToFile (); gStdLog.Clear (); gStdLog.EnsureNewFile

#define LogErrorMessages()		gPrefs->GetLogErrorMessages ()
#define LogWarningMessages()	gPrefs->GetLogWarningMessages ()
#define LogGremlins()			gPrefs->GetLogGremlins ()
#define LogCPUOpcodes()			gPrefs->GetLogCPUOpcodes ()
#define LogEnqueuedEvents()		gPrefs->GetLogEnqueuedEvents ()
#define LogDequeuedEvents()		gPrefs->GetLogDequeuedEvents ()
#define LogSystemCalls()		gPrefs->GetLogSystemCalls ()
#define LogApplicationCalls()	gPrefs->GetLogApplicationCalls ()
#define LogSerial()				gPrefs->GetLogSerial ()
#define LogSerialData()			gPrefs->GetLogSerialData ()
#define LogNetLib()				gPrefs->GetLogNetLib ()
#define LogNetLibData()			gPrefs->GetLogNetLibData ()
#define LogExgMgr()				gPrefs->GetLogExgMgr ()
#define LogExgMgrData()			gPrefs->GetLogExgMgrData ()
#define LogHLDebugger()			gPrefs->GetLogHLDebugger ()
#define LogHLDebuggerData()		gPrefs->GetLogHLDebuggerData ()
#define LogLLDebugger()			gPrefs->GetLogLLDebugger ()
#define LogLLDebuggerData()		gPrefs->GetLogLLDebuggerData ()


#endif /* _LOGGING_H_ */
