/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#ifndef _METAMEMORY_H_
#define _METAMEMORY_H_

#include "PalmHeap.h"			// PalmHeap
#include "ErrorHandling.h"		// Errors::EAccessType


class MetaMemory
{
	public:
		static void				Initialize				(uae_u32 ramSize);
		static void				Reset					(void);
		static void				Dispose					(void);

		// Called to mark and unmark some areas of memory.

		static void				MarkLowMemory			(uaecptr begin, uaecptr end);
		static void				UnmarkLowMemory			(uaecptr begin, uaecptr end);

		static void				MarkSystemGlobals		(uaecptr begin, uaecptr end);
		static void				UnmarkSystemGlobals		(uaecptr begin, uaecptr end);

		static void				MarkHeapHeader			(uaecptr begin, uaecptr end);
		static void				UnmarkHeapHeader		(uaecptr begin, uaecptr end);

		static void				MarkLowStack			(uaecptr begin, uaecptr end);
		static void				UnmarkLowStack			(uaecptr begin, uaecptr end);

		static void				MarkScreen				(uaecptr begin, uaecptr end);
		static void				UnmarkScreen			(uaecptr begin, uaecptr end);

		static void				MarkUnlockedChunk		(uaecptr begin, uaecptr end);
		static void				UnmarkUnlockedChunk		(uaecptr begin, uaecptr end);

		static void				MarkFreeChunk			(uaecptr begin, uaecptr end);
		static void				UnmarkFreeChunk			(uaecptr begin, uaecptr end);

		// Called when memory needs to be marked as initialized or not.

		static void				MarkUninitialized		(uaecptr begin, uaecptr end);
		static void				MoveUninitialized		(uaecptr source, uaecptr dest, uae_u32 size);
		static void				MarkInitialized			(uaecptr begin, uaecptr end);
		static void				MarkLongInitialized		(uaecptr);	// Inlined, defined below
		static void				MarkWordInitialized		(uaecptr);	// Inlined, defined below
		static void				MarkByteInitialized		(uaecptr);	// Inlined, defined below

		// Called on Chunk-based operations.

		static void				MemChunkNew				(VoidPtr, UInt attributes);
		static void				MemChunkFree			(VoidPtr);

		// Called on Pointer-based operations.

		static void				MemPtrNew				(VoidPtr);
		static void				MemPtrResize			(VoidPtr, UInt oldSize);

		// Called on Handle-based operations.

		static void				MemHandleNew			(VoidHand);
		static void				MemHandleResize			(VoidHand, UInt oldSize);
		static void				MemHandleFree			(VoidHand);

		// Called when a chunk is locked or unlocked.

		static void				MemLocalIDToLockedPtr	(VoidPtr);
		static void				MemHandleLock			(VoidHand);
		static void				MemHandleUnlock			(VoidHand);
		static void				MemHandleResetLock		(VoidHand);
		static void				MemPtrResetLock			(VoidPtr);
		static void				MemPtrUnlock			(VoidPtr);

		// Called to brute-force sync with the heap state.  Called
		// after things like MemHeapCompact, MemHeapInit, MemHeapScramble,
		// etc.  Implicitly called after above functions like MemFooNew.
		//
		// As an optimization, I don't resync non-relocatable blocks.
		// If that synchronization is needed, set fullSync to true.

		static void				SyncWithHeap			(UInt heapID, Bool fullSync);

		// Can a normal application access this memory location?
		// Forbidden locations are low-memory, system globals, screen
		// memory, memory manager structures, unlocked handles, and
		// uninitialized memory.

		static Bool				CanAppGetLong			(uaecptr);	// Inlined, defined below
		static Bool				CanAppGetWord			(uaecptr);	// Inlined, defined below
		static Bool				CanAppGetByte			(uaecptr);	// Inlined, defined below

		static Bool				CanAppSetLong			(uaecptr);	// Inlined, defined below
		static Bool				CanAppSetWord			(uaecptr);	// Inlined, defined below
		static Bool				CanAppSetByte			(uaecptr);	// Inlined, defined below

		// Can the non-Memory Manager parts of the system access this memory location?
		// Forbidden locations are low-memory, memory manager structures,
		// unlocked handles, and uninitialized memory.

		static Bool				CanSystemGetLong		(uaecptr);	// Inlined, defined below
		static Bool				CanSystemGetWord		(uaecptr);	// Inlined, defined below
		static Bool				CanSystemGetByte		(uaecptr);	// Inlined, defined below

		static Bool				CanSystemSetLong		(uaecptr);	// Inlined, defined below
		static Bool				CanSystemSetWord		(uaecptr);	// Inlined, defined below
		static Bool				CanSystemSetByte		(uaecptr);	// Inlined, defined below

		// Can the Memory Managers parts of the system access this memory location?
		// Forbidden locations are low-memory and uninitialized memory.

		static Bool				CanMemMgrGetLong		(uaecptr);	// Inlined, defined below
		static Bool				CanMemMgrGetWord		(uaecptr);	// Inlined, defined below
		static Bool				CanMemMgrGetByte		(uaecptr);	// Inlined, defined below

		static Bool				CanMemMgrSetLong		(uaecptr);	// Inlined, defined below
		static Bool				CanMemMgrSetWord		(uaecptr);	// Inlined, defined below
		static Bool				CanMemMgrSetByte		(uaecptr);	// Inlined, defined below

		static Errors::EAccessType
								GetWhatHappened			(uaecptr iAddress, long size, Bool forRead);
		static Errors::EAccessType
								AllowForBugs			(uaecptr iAddress, long size, Bool forRead, Errors::EAccessType);

		// Accessors for getting ranges of memory.

		static uaecptr			GetLowMemoryBegin		(void);
		static uaecptr			GetLowMemoryEnd			(void);
		static uaecptr			GetSysGlobalsBegin		(void);
		static uaecptr			GetSysGlobalsEnd		(void);
		static uaecptr			GetScreenBegin			(void);
		static uaecptr			GetScreenEnd			(void);
		static uaecptr			GetHeapHdrBegin			(UInt heapID);
		static uaecptr			GetHeapHdrEnd			(UInt heapID);

		// Predicates to help determine what kind of access error
		// just occurred.

		static Bool				IsLowMemory				(uaecptr testAddress, uae_u32 size);
		static Bool				IsSystemGlobal			(uaecptr testAddress, uae_u32 size);
		static Bool				IsScreenBuffer			(uaecptr testAddress, uae_u32 size);
		static Bool				IsMemMgrData			(uaecptr testAddress, uae_u32 size);
		static Bool				IsUnlockedChunk			(uaecptr testAddress, uae_u32 size);
		static Bool				IsFreeChunk				(uaecptr testAddress, uae_u32 size);
		static Bool				IsUninitialized			(uaecptr testAddress, uae_u32 size);
		static Bool				IsStack					(uaecptr testAddress, uae_u32 size);
		static Bool				IsLowStack				(uaecptr testAddress, uae_u32 size);
		static Bool				IsAllocatedChunk		(uaecptr testAddress, uae_u32 size);

		static Bool				IsScreenBuffer8			(uaecptr testAddress);
		static Bool				IsScreenBuffer16		(uaecptr testAddress);
		static Bool				IsScreenBuffer32		(uaecptr testAddress);

		// Hack function.  MetaMemory knows when the Palm OS heaps are initialized
		// when SyncWithHeap is called with fullSync == true.  We'll keep track of
		// that for functions that rely on knowing when the heaps are initialized.

		static Bool				HeapsInitialized		(void) { return fgHeapsAreInitialized; }

	private:
		struct ChunkCheck
		{
			uaecptr	testAddress;
			Bool	result;
		};

		struct WhatHappenedData
		{
			Errors::EAccessType	result;
			uaecptr				address;
			uae_u32				size;
			Bool				forRead;
		};

		static void				MarkRange				(uaecptr start, uaecptr end, uae_u8 v);
		static void				UnmarkRange				(uaecptr start, uaecptr end, uae_u8 v);
		static void				MarkUnmarkRange			(uaecptr start, uaecptr end,
														 uae_u8 andValue, uae_u8 orValue);

		static DWord			GetChunkSize			(VoidPtr	p);
		static DWord			GetChunkLockCount		(VoidPtr	p);

		static PalmHeap::EWalkerProcResult
								SyncOneBlock			(const PalmHeap::HeapInfo& heapInfo,
														 const PalmHeap::ChunkInfo& chunkInfo,
														 void* userData);
		static PalmHeap::EWalkerProcResult
								WhatHappenedCallback	(const PalmHeap::HeapInfo& heapInfo,
														 const PalmHeap::ChunkInfo& chunkInfo,
														 void* userData);
		static PalmHeap::EWalkerProcResult
								CheckForChunkHeader		(const PalmHeap::HeapInfo& heapInfo,
														 const PalmHeap::ChunkInfo& chunkInfo,
														 void* userData);
		static PalmHeap::EWalkerProcResult
								CheckForUnlockedChunk	(const PalmHeap::HeapInfo& heapInfo,
														 const PalmHeap::ChunkInfo& chunkInfo,
														 void* userData);
		static PalmHeap::EWalkerProcResult
								CheckForFreeChunk		(const PalmHeap::HeapInfo& heapInfo,
														 const PalmHeap::ChunkInfo& chunkInfo,
														 void* userData);
		static PalmHeap::EWalkerProcResult
								CheckForAllocatedChunk	(const PalmHeap::HeapInfo& heapInfo,
														 const PalmHeap::ChunkInfo& chunkInfo,
														 void* userData);

		static Bool				ValidChunk				(VoidPtr);
		static Bool				ValidHandle				(VoidHand);


		/*
			Memory is laid out as follows:

				+-------------------+
				| Low Memory		|	Off Limits to everyone except IRQ handlers, SysSleep, and SysDoze
				+-------------------+
				|					|
				| System Globals	|	Off Limits to applications
				|					|
				+-------------------+
				| Dynamic Heap		|
				|+-----------------+|
				|| Heap Header     ||	Off Limits to applications and non-MemMgr system
				|+-----------------+|
				|| Chunks	       ||	Off Limits to apps and non-MemMgr system if unlocked
				|+-----------------+|
				|| Free chunk      ||	Off Limits to applications and non-MemMgr system
				|+-----------------+|
				|| Chunks	       ||
				|+-----------------+|
				|| Chunks	       ||
				|+-----------------+|
				|| Chunks	       ||
				|+-----------------+|
				|| Stack Chunk     ||	Off Limits to everyone if below A7
				|+-----------------+|
				|| Screen Chunk    ||	Off Limits to applications
				|+-----------------+|
				+-------------------+

			Any memory location can be marked as "uninitialized", which means that
			it can be written to but not read from.  An exception to this would be
			the parts of the memory manager that moves around blocks (which may
			contain uninitialized sections).
		*/

		enum
		{
			kNoAppAccess		= 0x01,
			kNoSystemAccess		= 0x02,
			kNoMemMgrAccess		= 0x04,

			kScreenBuffer		= 0x20,	// Screen buffer; update host screen if these bytes are changed.
#if FOR_LATER
			kUninitialized		= 0x40,	// Bytes have not been written to.
#endif
			kPatchResource		= 0x80,	// Bytes belong to a system update resource (stored on even bytes.)
			kCodeCoverage		= 0x80,	// Bytes have been executed during code coverage test (stored on odd bytes.)

			kLowMemoryBits		= kNoAppAccess | kNoSystemAccess | kNoMemMgrAccess,
			kGlobalsBits		= kNoAppAccess,
			kMPTBits			= kNoAppAccess,
			kMemStructBits		= kNoAppAccess | kNoSystemAccess,
			kLowStackBits		= kNoAppAccess | kNoSystemAccess | kNoMemMgrAccess,
			kFreeChunkBits		= kNoAppAccess | kNoSystemAccess,	//kInaccessibleChunk,
			kUnlockedChunkBits	= kNoAppAccess | kNoSystemAccess,	//kInaccessibleChunk,
			kScreenBits			= kNoAppAccess | kScreenBuffer,
			kFreeAccessBits		= 0
		};

		static uae_u8*			fgMetaMemory;
		static uae_u32			fgMetaMemorySize;

		static Bool				fgHeapsAreInitialized;
};




// Macros to take a byte of bits and produce an
// 8-bit, 16-bit, or 32-bit version of it.

#define META_BITS_32(bits)			\
	(((uae_u32) (bits)) << 24) |	\
	(((uae_u32) (bits)) << 16) |	\
	(((uae_u32) (bits)) <<  8) |	\
	(((uae_u32) (bits)))

#define META_BITS_16(bits)			\
	(((uae_u16) (bits)) <<  8) |	\
	(((uae_u16) (bits)))

#define META_BITS_8(bits)			\
	(((uae_u8) (bits)))

// Macros to fetch the appropriate 8-bit, 16-bit, or
// 32-bit value from meta-memory.

#define META_VALUE_32(p)			\
	(*(uae_u32*) &fgMetaMemory[p])

#define META_VALUE_16(p)			\
	(*(uae_u16*) &fgMetaMemory[p])

#define META_VALUE_8(p)				\
	(*(uae_u8*) &fgMetaMemory[p])

// Macros to define CanAppGetLong, CanAppSetLong,
// CanSystemGetLong, CanSystemSetLong, etc.

#define DEFINE_FUNCTIONS(kind, bits)				\
													\
inline Bool MetaMemory::Can##kind##Long (uaecptr p)	\
{													\
	const uae_u32	kMask = META_BITS_32 (bits);	\
													\
	return (META_VALUE_32 (p) & kMask) == 0;		\
}													\
													\
inline Bool MetaMemory::Can##kind##Word (uaecptr p)	\
{													\
	const uae_u16	kMask = META_BITS_16 (bits);	\
													\
	return (META_VALUE_16 (p) & kMask) == 0;		\
}													\
													\
inline Bool MetaMemory::Can##kind##Byte (uaecptr p)	\
{													\
	const uae_u8	kMask = META_BITS_8 (bits);		\
													\
	return (META_VALUE_8 (p) & kMask) == 0;			\
}


#if FOR_LATER
DEFINE_FUNCTIONS(AppGet, kNoAppAccess | kUninitialized)
DEFINE_FUNCTIONS(AppSet, kNoAppAccess)

DEFINE_FUNCTIONS(SystemGet, kNoSystemAccess | kUninitialized)
DEFINE_FUNCTIONS(SystemSet, kNoSystemAccess)

DEFINE_FUNCTIONS(MemMgrGet, kNoMemMgrAccess | kUninitialized)
DEFINE_FUNCTIONS(MemMgrSet, kNoMemMgrAccess)
#else
DEFINE_FUNCTIONS(AppGet, kNoAppAccess)
DEFINE_FUNCTIONS(AppSet, kNoAppAccess)

DEFINE_FUNCTIONS(SystemGet, kNoSystemAccess)
DEFINE_FUNCTIONS(SystemSet, kNoSystemAccess)

DEFINE_FUNCTIONS(MemMgrGet, kNoMemMgrAccess)
DEFINE_FUNCTIONS(MemMgrSet, kNoMemMgrAccess)
#endif


inline void MetaMemory::MarkLongInitialized (uaecptr p)
{
#if FOR_LATER
	const uae_u32	kMask	=	META_BITS_32 (kUninitialized);

	META_VALUE_32 (p) &= ~kMask;
#endif
}

inline void MetaMemory::MarkWordInitialized (uaecptr p)
{
#if FOR_LATER
	const uae_u16	kMask	=	META_BITS_16 (kUninitialized);

	META_VALUE_16 (p) &= ~kMask;
#endif
}

inline void MetaMemory::MarkByteInitialized (uaecptr p)
{
#if FOR_LATER
	const uae_u8	kMask	=	META_BITS_8 (kUninitialized);

	META_VALUE_8 (p) &= ~kMask;
#endif
}


inline Bool MetaMemory::IsScreenBuffer (uaecptr testAddress, uae_u32 size)
{
	if (size == 1)
	{
		const uae_u8 kMask = META_BITS_8 (kScreenBuffer);

		return ((*(uae_u8*) &fgMetaMemory[testAddress]) & kMask) != 0;
	}
	else if (size == 2)
	{
		const uae_u16 kMask = META_BITS_16 (kScreenBuffer);

		return ((*(uae_u16*) &fgMetaMemory[testAddress]) & kMask) != 0;
	}
	else if (size == 4)
	{
		const uae_u32 kMask = META_BITS_32 (kScreenBuffer);

		return ((*(uae_u32*) &fgMetaMemory[testAddress]) & kMask) != 0;
	}

	const uae_u8 kMask = META_BITS_8 (kScreenBuffer);

	for (uae_u32 ii = 0; ii < size; ++ii)
	{
		if (((*(uae_u8*) &fgMetaMemory[testAddress+ii]) & kMask) != 0)
		{
			return true;
		}
	}

	return false;
}

inline Bool MetaMemory::IsScreenBuffer8 (uaecptr testAddress)
{
	const uae_u8 kMask = META_BITS_8 (kScreenBuffer);

	return ((*(uae_u8*) &fgMetaMemory[testAddress]) & kMask) != 0;
}

inline Bool MetaMemory::IsScreenBuffer16 (uaecptr testAddress)
{
	const uae_u16 kMask = META_BITS_16 (kScreenBuffer);

	return ((*(uae_u16*) &fgMetaMemory[testAddress]) & kMask) != 0;
}

inline Bool MetaMemory::IsScreenBuffer32 (uaecptr testAddress)
{
	const uae_u32 kMask = META_BITS_32 (kScreenBuffer);

	return ((*(uae_u32*) &fgMetaMemory[testAddress]) & kMask) != 0;
}

#endif /* _METAMEMORY_H_ */
  
  
