/***********************************************************************
 *
 *	Copyright (c) Palm Computing 1996 -- All Rights Reserved
 *
 * PROJECT:  Pilot INet Library
 * FILE:     INetLib.h
 * AUTHOR:	 Ron Marianetti 6/2/97
 *
 * DESCRIPTION:
 *	  This header file contains equates for the Internet Library. 
 *
 **********************************************************************/
#ifndef 	__INETLIB_H__
#define	__INETLIB_H__



#pragma mark Constants

// Creator. Used for both the database that contains the INet Library and
//  it's features for the feature manager. 
#define	inetCreator				'inet'			// The Net Library creator

// INet Library features have this creator 
#define	inetLibFtrCreator		inetCreator		// creatorID of INet Lib features. 

// Name of the InetLib
#define	inetLibName				"INet.lib"		// pass in to SysLibFind()


// Feature inetCreator, #0 is index of the the version number feature.
// The Feature creator is inetLibFtrCreator.
// Encoding is: 0xMMmfsbbb, where MM is major version, m is minor version
// f is bug fix, s is stage: 3-release,2-beta,1-alpha,0-development,
// bbb is build number for non-releases 
// V1.12b3   would be: 0x01122003
// V2.00a2   would be: 0x02001002
// V1.01     would be: 0x01013000
#define  inetFtrNumVersion		0

#define	inetLibType				'libr'		// Our Net Code Resources Database type


// ID for proxy IP address in flash
#define	inetFlashProxyID		'IP'

//-----------------------------------------------------------------------------
// IP addresses of Elaine servers - used for default wireless proxies
//-----------------------------------------------------------------------------
#define netProxyIPManhattanHGA				0x0A0186A5			// Manhattan HGA = 10.1.134.165 or MAN 100005
#define	netProxyIPTuscanyHGA				0x0A0186A3 			// Tuscany HGA = 10.1.134.163 or MAN 100003
#define	netProxyIPRonsHGA					0x0A0186A4 			// Ron's HGA = 10.1.134.164 or MAN 100004
#define netProxyIPDefaultHGA				netProxyIPManhattanHGA
#define netProxyIPDefaultHGAStr			"10.1.134.165"		//Should correspond to above value

#pragma mark ErrorCodes
/********************************************************************
 * Error codes
 ********************************************************************/
#define 	inetErrNone							0
#define	inetErrTooManyClients			(inetErrorClass | 1)		// Too many clients already
#define	inetErrHandleInvalid				(inetErrorClass | 2)		// Invalid inetH or sockH
#define	inetErrParamsInvalid				(inetErrorClass | 3)		//  
#define	inetErrURLVersionInvalid		(inetErrorClass | 4)		// 
#define	inetErrURLBufTooSmall			(inetErrorClass | 5)		// 
#define	inetErrURLInvalid					(inetErrorClass | 6)		// 
#define	inetErrTooManySockets			(inetErrorClass | 7)		// 
#define	inetErrNoRequestCreated			(inetErrorClass | 8)
#define	inetErrNotConnected				(inetErrorClass | 9)
#define	inetErrInvalidRequest			(inetErrorClass | 10)
#define	inetErrNeedTime					(inetErrorClass | 11)
#define	inetErrHostnameInvalid			(inetErrorClass | 12)
#define	inetErrInvalidPort				(inetErrorClass | 13)
#define	inetErrInvalidHostAddr			(inetErrorClass | 14)
#define	inetErrNilBuffer					(inetErrorClass | 15)
#define	inetErrConnectTimeout			(inetErrorClass | 16)
#define	inetErrResolveTimeout			(inetErrorClass | 17)
#define	inetErrSendReqTimeout			(inetErrorClass | 18)
#define	inetErrReadTimeout				(inetErrorClass | 19)
#define	inetErrBufTooSmall				(inetErrorClass | 20)
#define	inetErrSchemeNotSupported		(inetErrorClass | 21)
#define	inetErrInvalidResponse			(inetErrorClass | 22)
#define	inetErrSettingTooLarge			(inetErrorClass | 25)
#define	inetErrSettingSizeInvalid		(inetErrorClass | 26)
#define	inetErrRequestTooLong			(inetErrorClass | 27)
#define	inetErrSettingNotImplemented	(inetErrorClass | 28)

// Configuration errors
#define	inetErrConfigNotFound			(inetErrorClass | 29)
#define	inetErrConfigCantDelete			(inetErrorClass | 30)
#define	inetErrConfigTooMany				(inetErrorClass | 31)
#define	inetErrConfigBadName				(inetErrorClass | 32)
#define	inetErrConfigNotAlias			(inetErrorClass | 33)
#define	inetErrConfigCantPointToAlias	(inetErrorClass | 34)
#define	inetErrConfigEmpty				(inetErrorClass | 35)
#define	inetErrConfigAliasErr			(inetErrorClass | 37)

#define	inetErrNoWirelessInterface		(inetErrorClass | 38)

// Encryption related errors
#define	inetErrEncryptionNotAvail		(inetErrorClass | 39)
	// Need to re-send transaction because server told us to reset our
	//  encryption sequence number
#define	inetErrNeedRetryEncSeqNum		(inetErrorClass | 40)
	// Need to re-send transaction because server sent us a new
	//  public key to use. 
#define	inetErrNeedRetryEncPublicKey	(inetErrorClass | 41)

#define	inetErrResponseTooShort			(inetErrorClass | 42)

// errors specific to handling Mobitex ILLEGAL responses
#define	inetErrMobitexIllegalOKHost	(inetErrorClass | 43)
#define	inetErrMobitexIllegalBadHost	(inetErrorClass | 44)
// see error 92 also

// HTTP errors
#define	inetErrHTTPBadRequest			(inetErrorClass | 45)
#define	inetErrHTTPUnauthorized			(inetErrorClass | 46)
#define	inetErrHTTPForbidden				(inetErrorClass | 47)
#define	inetErrHTTPNotFound				(inetErrorClass | 48)
#define	inetErrHTTPMethodNotAllowed	(inetErrorClass | 49)
#define	inetErrHTTPNotAcceptable		(inetErrorClass | 50)
#define	inetErrHTTPProxyAuthRequired	(inetErrorClass | 51)
#define	inetErrHTTPRequestTimeout		(inetErrorClass | 52)
#define	inetErrHTTPConflict				(inetErrorClass | 53)
#define	inetErrHTTPGone					(inetErrorClass | 54)
#define	inetErrHTTPLengthRequired		(inetErrorClass | 55)
#define	inetErrHTTPPreconditionFailed	(inetErrorClass | 56)
#define	inetErrHTTPRequestTooLarge		(inetErrorClass | 57)
#define	inetErrHTTPRequestURITooLong	(inetErrorClass | 58)
#define	inetErrHTTPUnsupportedType		(inetErrorClass | 59)
#define	inetErrHTTPServerError			(inetErrorClass | 60)

// CTP errors
#define	inetErrCTPServerError			(inetErrorClass | 61)

// Cache errors
#define	inetErrTypeNotCached				(inetErrorClass | 62)
#define 	inetErrCacheInvalid				(inetErrorClass | 63)

// Palm: and PalmCall: scheme errors
#define	inetErrURLDispatched				(inetErrorClass | 64)
#define 	inetErrDatabaseNotFound			(inetErrorClass | 65)

#define 	inetErrCTPMalformedRequest		(inetErrorClass | 66)
#define 	inetErrCTPUnknownCommand		(inetErrorClass | 67)
#define 	inetErrCTPTruncated				(inetErrorClass | 68)
#define 	inetErrCTPUnknownError			(inetErrorClass | 69)
#define 	inetErrCTPProxyError				(inetErrorClass | 70)
#define 	inetErrCTPSocketErr				(inetErrorClass | 71)
    
#define 	inetErrCTPInvalidURL				(inetErrorClass | 72)
#define 	inetErrCTPReferringPageOutOfDate (inetErrorClass | 73)
#define 	inetErrCTPBadRequest					(inetErrorClass | 74)
#define 	inetErrUNUSED					 	(inetErrorClass | 75)
#define 	inetErrCTPMailServerDown		(inetErrorClass | 76)
#define 	inetErrCTPHostNotFound			(inetErrorClass | 77)


// Content Conversion Errors
#define 	inetErrCTPContentInvalidTag       (inetErrorClass | 78)
#define 	inetErrCTPContentInternal         (inetErrorClass | 79)
#define 	inetErrCTPContentDataEnd          (inetErrorClass | 80)
#define 	inetErrCTPContentResourceTooBig   (inetErrorClass | 81)
#define 	inetErrCTPContentNoNoFrames       (inetErrorClass | 82)
#define 	inetErrCTPContentUnsupportedContent 	(inetErrorClass | 83)
#define 	inetErrCTPContentUnsupportedEncoding 	(inetErrorClass | 84)
#define 	inetErrCTPContentBadForm          		(inetErrorClass | 85)
#define 	inetErrCTPContentBadFormMissingAction 	(inetErrorClass | 86)
#define 	inetErrCTPContentBadFormMissingMethod 	(inetErrorClass | 87)
#define 	inetErrCTPContentNoSourceData     (inetErrorClass | 88)
#define 	inetErrCTPContentBadImage         (inetErrorClass | 89)
#define 	inetErrCTPContentImageTooLarge    (inetErrorClass | 90)

// Mobitex illegal handled error code.  This error is sent after
//INetLib handles inetErrMobitexIllegalOKHost or inetErrMobitexIllegalBadHost
//errors.  The application needs to know that something went wrong and it needs
//to change state.  This error does not need to be displayed to the user.
#define	inetErrMobitexErrorHandled		(inetErrorClass | 91)

// Proxy down, non-default host, show dialog asking to revert to default
#define	inetErrProxyDownBadHost			(inetErrorClass | 92)

// A second type of readtime.  This should occur only when some data is received
// and the connection is lost.
#define	inetErrHostConnectionLost		(inetErrorClass | 93)


//

#pragma mark Types

/********************************************************************
 * Input flags
 ********************************************************************/
//-----------------------------------------------------------------------------
// flag word definitions for INetLibURLOpen
//-----------------------------------------------------------------------------
#define	inetOpenURLFlagLookInCache	0x0001
#define  inetOpenURLFlagKeepInCache	0x0002
#define	inetOpenURLFlagForceEncOn  0x0008 // use encryption even if 
														 //  scheme does not desire it
#define	inetOpenURLFlagForceEncOff	0x0010 // no encryption even if 
														 //  scheme desires it

//-----------------------------------------------------------------------------
// flag word definitions for INetURLInfo. These flags bits are set in the
//   flags field of the INetURLINfoType structure by INetLibURLGetInfo()
//-----------------------------------------------------------------------------
#define	inetURLInfoFlagIsSecure		0x0001
#define  inetURLInfoFlagIsRemote		0x0002
#define  inetURLInfoFlagIsInCache	0x0004


/********************************************************************
 * Configuration Support
 ********************************************************************/
//-----------------------------------------------------------------------------
// Names of built-in configuration aliases available through the
//  INetLibConfigXXX calls
//-----------------------------------------------------------------------------
#define	inetCfgNameDefault		".Default"		// The default configuration
#define	inetCfgNameDefWireline	".DefWireline"	// The default wireline configuration
#define	inetCfgNameDefWireless	".DefWireless"	// The default wireless configuration
#define	inetCfgNameCTPDefault	".CTPDefault"	// Points to either .CTPWireline or .CTPWireless
#define	inetCfgNameCTPWireline	".CTPWireline"	// Wireline through the Jerry Proxy
#define	inetCfgNameCTPWireless	".CTPWireless"	// Wireless through the Jerry Proxy


//--------------------------------------------------------------------
// Structure of a configuration name. Used by INetLibConfigXXX calls
//---------------------------------------------------------------------
#define	inetConfigNameSize		32
typedef struct {
	Char			name[inetConfigNameSize];			  // name of configuration
	} INetConfigNameType, *INetConfigNamePtr;




/********************************************************************
 * Scheme Types
 ********************************************************************/
typedef enum {
	inetSchemeUnknown = -1,
	inetSchemeDefault = 0,
	
	inetSchemeHTTP,								// http:
	inetSchemeHTTPS,								// https:
	inetSchemeFTP,									// ftp:
	inetSchemeGopher,								// gopher:
	inetSchemeFile,								// file:
	inetSchemeNews,								// news:
	inetSchemeMailTo,								// mailto:
   inetSchemePalm,								// palm:
   inetSchemePalmCall,							// palmcall:
   
	inetSchemeMail,								// not applicable to URLS, but used
														//  for the INetLibSockOpen call when
														//  creating a socket for mail IO
	inetSchemeMac,									// mac: - Mac file system HTML													
	
	inetSchemeFirst = inetSchemeHTTP,		// first one
	inetSchemeLast	= inetSchemeMail			// last one
	} INetSchemeEnum;


/********************************************************************
 * Scheme Ports
 ********************************************************************/
#define	inetPortFTP		21
#define	inetPortHTTP	80
#define	inetPortGopher	70
#define	inetPortNews	119
#define	inetPortHTTPS	44



/********************************************************************
 * Structure of a cracked URL. 
 ********************************************************************/
typedef struct {
	Word		version;								// should be 0, for future compatibility
	
	BytePtr	schemeP;								// ptr to scheme portion
	Word		schemeLen;							// size of scheme portion
	Word		schemeEnum;							// INetSchemEnum
	
	BytePtr	usernameP;							// ptr to username portion
	Word		usernameLen;						// size of username
	
	BytePtr	passwordP;							// ptr to password portion
	Word		passwordLen;						// size of password
	
	BytePtr	hostnameP;							// ptr to host name portion
	Word		hostnameLen;						// size of host name
	
	Word		port;									// port number
	
	BytePtr	pathP;								// ptr to path portion
	Word		pathLen;								// size of path
	
	BytePtr	paramP;								// param (;param)
	Word		paramLen;							// size of param
	
	BytePtr	queryP;								// query (?query)
	Word		queryLen;							// size of query
	
	BytePtr	fragP;								// fragment (#frag)
	Word		fragLen;								// size of fragment
	} INetURLType;
	


/********************************************************************
 * Structure for INetURLInfo. This structure is filled in with info
 *  about a URL. 
 ********************************************************************/
typedef struct {
	Word		version;								// should be 0, for future compatibility
	
	Word		flags;								// flags word, one or ore of 
														//   inetURLInfoFlagXXX flags
	DWord		undefined;							// reserved for future use
	} INetURLInfoType;



/********************************************************************
 * Content and Compression Type Enums(from proxy server or PQF Builder)
 ********************************************************************/
typedef enum {
	inetContentTypeTextPlain = 0,
	inetContentTypeTextHTML,
	inetContentTypeImageGIF,
	inetContentTypeImageJPEG,
	inetContentTypeApplicationCML,
	inetContentTypeImagePalmOS
	} INetContentTypeEnum;
	
typedef enum {
	inetCompressionTypeNone = 0,
	inetCompressionTypeBitPacked,
	inetCompressionTypeLZ77
	} INetCompressionTypeEnum;

/********************************************************************
 * Proxy Types
 ********************************************************************/
typedef enum {
	inetProxyNone = 0,							// no proxy
	inetProxyCTP = 1								// CTP (Jerry) proxy
	} INetProxyEnum;



/********************************************************************
 * Settings for the INetLibSettingSet/Get call. 
 ********************************************************************/
typedef enum {
	inetSettingProxyType,						// (RW) DWord, INetProxyEnum
	
	inetSettingProxyName,						// (RW) Char[], name of proxy
	inetSettingProxyPort,						// (RW) DWord,  TCP port # of proxy
	
	inetSettingProxySocketType,				// (RW) DWord, which type of socket to use
														//  netSocketTypeXXX
	
	inetSettingCacheSize,						// (RW) DWord, max size of cache
	inetSettingCacheRef,							// (R) DmOpenRef, ref of cache DB
	
	inetSettingNetLibConfig,					// (RW) DWord, Which NetLib config to use.
		
	inetSettingRadioID,							// (R)  DWord[2], the 64-bit radio ID
	inetSettingBaseStationID,					// (R)  DWord, the radio base station ID
	
	inetSettingMaxRspSize,						// (W) DWord (in bytes)
	inetSettingConvAlgorithm,					// (W) DWord (CTPConvEnum)
	inetSettingContentWidth,					// (W) DWord (in pixels)

	inetSettingLast
	} INetSettingEnum;


/********************************************************************
 * Settings for the INetLibSockSettingSet/Get call. 
 ********************************************************************/
typedef enum {
	inetSockSettingScheme,						// (R) DWord, INetSchemeEnum
	inetSockSettingSockContext,				// (RW) DWord,

	inetSockSettingCompressionType,			// (R) Char[]
	inetSockSettingCompressionTypeID,		// (R) DWord (INetCompressionTypeEnum)
	inetSockSettingContentType,				// (R) Char[]
	inetSockSettingContentTypeID,				// (R) DWord (INetContentTypeEnum)
	inetSockSettingData,							// (R) DWord, pointer to data
	inetSockSettingDataHandle,					// (R) DWord, Handle to data
	inetSockSettingDataOffset,					// (R) DWord, offset to data from handle

	inetSockSettingTitle,						// (RW) Char[]
	inetSockSettingURL,							// (R) Char[]
	inetSockSettingIndexURL,					// (RW) Char[]
	
	inetSockSettingFlags,						// (W) Word, one or more of
														//   inetOpenURLFlagXXX flags
														
	inetSockSettingReadTimeout,				// (RW) DWord. Read timeout in ticks

	inetSockSettingLast
	} INetSockSettingEnum;




/********************************************************************
 * Possible socket status values that can be returned from INetLibSockStatus
 ********************************************************************/
typedef enum {
	inetStatusNew,									// just opened
	inetStatusResolvingName,					// looking up host address
	inetStatusNameResolved,						// found host address
	inetStatusConnecting,						// connecting to host
	inetStatusConnected,							// connected to host
	inetStatusSendingRequest,					// sending request
	inetStatusWaitingForResponse,				// waiting for response
	inetStatusReceivingResponse,				// receiving response
	inetStatusResponseReceived,				// response received
	inetStatusClosingConnection,				// closing connection
	inetStatusClosed,								// closed
	inetStatusAcquiringNetwork,				// network temporarily
													// unreachable; socket on hold
	inetStatusPrvInvalid = 30					// internal value, not
													// returned by INetMgr. Should
													// be last.
	} INetStatusEnum;



/********************************************************************
 * HTTP Attributes which can be set/get using the
 *  INetLibHTTPAttrSet/Get calls.
 *
 * Generally, attributes are only set BEFORE calling 
 *		INetLibSockHTTPReqSend
 * and attributes are only gotten AFTER the complete response
 *		 has been received.
 *
 * Attributes marked with the following flags:
 *		(R)	- read only
 *		(W)	- write only
 *		(RW)	- read/write
 *		(-) 	- not implemented yet
 ********************************************************************/
typedef enum {

	// local error trying to communicate with server, if any
	inetHTTPAttrCommErr,							// (R) DWord, read-only
	
	// object attributes, defined at creation
	inetHTTPAttrEntityURL,						// (-) Char[], which resource was requested
	
	
	//-----------------------------------------------------------
	// Request only attributes
	//-----------------------------------------------------------
	inetHTTPAttrReqAuthorization,				// (-) Char[]		
	inetHTTPAttrReqFrom,							// (-) Char[]
	inetHTTPAttrReqIfModifiedSince,			// (-) DWord
	inetHTTPAttrReqReferer,						// (-) Char[]

	// The following are ignored unless going through a CTP proxy
	inetHTTPAttrWhichPart,						// (W) DWord (0 -> N)
	inetHTTPAttrIncHTTP,							// (W) DWord (Boolean) only applicable
														//   when inetHTTPAttrConvAlgorithm set to
														//   ctpConvNone
	inetHTTPAttrCheckMailHi,					// (W) DWord
	inetHTTPAttrCheckMailLo,					// (W) DWord
	inetHTTPAttrReqContentVersion,			// (W) DWord Desired content version. Represented
														//  as 2 low bytes. Lowest byte is minor version,
														//  next higher byte is major version. 
	
	

	//--------------------------------------------------------------
	// Response only attributes
	//--------------------------------------------------------------
	// Server response info
	inetHTTPAttrRspAll,							// (-) Char[] - entire HTTP response including
														//   data
	inetHTTPAttrRspSize,							// (R) DWord - entire HTTP Response size including
														//   header and data
	inetHTTPAttrRspVersion,						// (-) Char[]
	inetHTTPAttrResult,							// (R) DWord (ctpErrXXX when using CTP Proxy)
	inetHTTPAttrErrDetail,						// (R) DWord (server/proxy err code when 
														//      using CTP Proxy)
	inetHTTPAttrReason,							// (R) Char[]
	inetHTTPAttrDate,								// (-) DWord
	inetHTTPAttrNoCache,							// (-) DWord
	inetHTTPAttrPragma,							// (-) Char[]
	inetHTTPAttrServer,							// (-) Char[]
	inetHTTPAttrWWWAuthentication,			// (-) Char[]
	
	
	// Returned entity attributes
	inetHTTPAttrContentAllow,					// (-) Char[]
	inetHTTPAttrContentLength,					// (R) DWord
	inetHTTPAttrContentLengthUncompressed,	// (R) DWord (in bytes)
	inetHTTPAttrContentPtr,						// (-) CharPtr
	inetHTTPAttrContentExpires,				// (-) DWord
	inetHTTPAttrContentLastModified,			// (-) DWord
	inetHTTPAttrContentLocation,				// (-) Char[]
	inetHTTPAttrContentLengthUntruncated,	// (R) DWord
	inetHTTPAttrContentVersion,					// (R) DWord, actual content version. Represented
														//  as 2 low bytes. Lowest byte is minor version,
														//  next higher byte is major version. 
	inetHTTPAttrContentCacheID					// (R) DWord, cacheID for this item
	
	
	
	} INetHTTPAttrEnum;




/********************************************************************
 * Structure of our Internet events. This structure is a superset of
 *  the regular event type. Note that we use the first 2 user events
 *  for the Internet Library so any app that uses this library must be
 *  to use user event IDs greater than inetLastEvent.
 *
 * DOLATER.... define a new system event for libraries and store the
 *  library refNum in it....
 ********************************************************************/
#define		inetSockReadyEvent				firstINetLibEvent
#define		inetSockStatusChangeEvent		firstINetLibEvent+1
#define		inetLastEvent						firstINetLibEvent+1

typedef struct {
	enum events    eType;
	Boolean        penDown;
	Byte				reserved;
	SWord          screenX;
	SWord          screenY;
	union {
		struct  {
			Word data1;
			Word data2;
			Word data3;
			Word data4;
			Word data5;
			Word data6;
			Word data7;
			Word data8;
			} generic;

		struct  {
			Handle	sockH;								// socket handle
			DWord		context;								// application defined
			Boolean	inputReady;							// true if ready for reads
			Boolean	outputReady;						// true if ready for writes
			} inetSockReady;

		struct  {
			Handle	sockH;								// socket handle
			DWord		context;								// application defined
			Word		status;								// new status
			Err		sockErr;								// socket err, if any
			}inetSockStatusChange;
		} data;
	} INetEventType;


/********************************************************************
 * INet Library functions. 
 ********************************************************************/
#pragma mark Traps
typedef enum {
	inetLibTrapSettingGet = sysLibTrapCustom,
	inetLibTrapSettingSet,
	
	inetLibTrapGetEvent,
	
	inetLibTrapURLOpen,
	
	inetLibTrapSockRead,
	inetLibTrapSockWrite,
	
	inetLibTrapSockOpen,
	inetLibTrapSockClose,
	inetLibTrapSockStatus,
	inetLibTrapSockSettingGet,
	inetLibTrapSockSettingSet,
	inetLibTrapSockConnect,
	
	// Utilities
	inetLibTrapURLCrack,
	inetLibTrapURLsAdd,
	inetLibTrapURLsCompare,
	inetLibTrapURLGetInfo,
	
	// HTTP calls
	inetLibTrapSockHTTPReqCreate,
	inetLibTrapSockHTTPAttrSet,
	inetLibTrapSockHTTPReqSend,
	inetLibTrapSockHTTPAttrGet,
	
	// Mail traps
	inetLibTrapSockMailReqCreate,
	inetLibTrapSockMailAttrSet,
	inetLibTrapSockMailReqAdd,
	inetLibTrapSockMailReqSend,
	inetLibTrapSockMailAttrGet,
	inetLibTrapSockMailQueryProgress,
	
	// Cache calls
	inetLibTrapCacheList,
	inetLibTrapCacheGetObject,
	
	// Config calls
	inetLibConfigMakeActive,
	inetLibConfigList,
	inetLibConfigIndexFromName,
	inetLibConfigDelete,
	inetLibConfigSaveAs,
	inetLibConfigRename,
	inetLibConfigAliasSet,
	inetLibConfigAliasGet,


	inetLibTrapLast
	} INetLibTrapNumberEnum;


/********************************************************************
 * Structure of cache entry
 *	Used as a parameter to INetLibCacheList. If urlP or titleP are NULL,
 *	the corresponding length fields will be updated with the desired lengths
 ********************************************************************/
typedef struct {
	BytePtr	urlP;
	Word		urlLen;
	
	BytePtr	titleP;
	Word		titleLen;
	
	DWord		lastViewed;							// seconds since 1/1/1904 
	DWord		firstViewed;						// seconds since 1/1/1904 
} INetCacheEntryType, *INetCacheEntryP;

/********************************************************************
 * Structure for INetLibCacheGetObject. This structure is filled in with info
 *  about a cache entry. 
 ********************************************************************/
typedef struct {
	Handle						recordH;
	INetContentTypeEnum		contentType;
	INetCompressionTypeEnum	encodingType;
	DWord							uncompressedDataSize;
	Byte							flags;
	Byte							reserved;
	Word							dataOffset;					// offset to content
	Word							dataLength;					// size of content
	Word							urlOffset;					// offset to URL
	ULong							viewTime;					// time last viewed
	ULong							createTime;					// time entry was created
	Word							murlOffset;					// offset to master URL
	} INetCacheInfoType, *INetCacheInfoPtr;

/********************************************************************
 * Mail Specific Equates
 ********************************************************************/
#include <INetMgrMail.h>


#pragma mark Functions
//--------------------------------------------------
// Library initialization, shutdown, sleep and wake
//--------------------------------------------------
Err				INetLibOpen (Word libRefnum, Word config, DWord flags,
							DmOpenRef cacheRef, DWord cacheSize, 
							Handle* inetHP)
						SYS_TRAP(sysLibTrapOpen);
					
Err				INetLibClose (Word libRefnum, Handle inetH)
						SYS_TRAP(sysLibTrapClose);
					
Err				INetLibSleep (Word libRefnum)
						SYS_TRAP(sysLibTrapSleep);
					
Err				INetLibWake (Word libRefnum)
						SYS_TRAP(sysLibTrapWake);
					
//--------------------------------------------------
// Settings
//--------------------------------------------------
Err				INetLibSettingGet(Word libRefnum, Handle inetH, 
							Word /*INetSettingEnum */ setting,
							VoidPtr bufP, WordPtr bufLenP)
						SYS_TRAP(inetLibTrapSettingGet);

Err				INetLibSettingSet(Word libRefnum, Handle inetH, 
							Word /*INetSettingEnum*/ setting,
							VoidPtr bufP, Word bufLen)
						SYS_TRAP(inetLibTrapSettingSet);


//--------------------------------------------------
// Event Management
//--------------------------------------------------

void			INetLibGetEvent(Word libRefnum, Handle inetH, 
									INetEventType* eventP, SDWord timeout)
						SYS_TRAP(inetLibTrapGetEvent);

						

//--------------------------------------------------
// High level calls
//--------------------------------------------------

Err				INetLibURLOpen(Word libRefnum, Handle inetH, BytePtr urlP, 
							BytePtr cacheIndexURLP, Handle* sockHP, SDWord timeout, Word flags)
						SYS_TRAP(inetLibTrapURLOpen);
						

Err				INetLibSockClose(Word libRefnum, Handle socketH)
						SYS_TRAP(inetLibTrapSockClose);
						

//--------------------------------------------------
// Read/Write
//--------------------------------------------------

Err				INetLibSockRead(Word libRefnum, Handle sockH, VoidPtr bufP,
							DWord reqBytes, DWordPtr actBytesP, SDWord timeout)
						SYS_TRAP(inetLibTrapSockRead);
						
Err				INetLibSockWrite(Word libRefnum, Handle sockH, VoidPtr bufP,
							DWord reqBytes, DWordPtr actBytesP, SDWord timeout)
						SYS_TRAP(inetLibTrapSockWrite);
						

//--------------------------------------------------
// Low level Socket calls
//--------------------------------------------------

Err				INetLibSockOpen(Word libRefnum, Handle inetH,
							Word /*INetSchemEnum*/ scheme,  Handle* sockHP)
						SYS_TRAP(inetLibTrapSockOpen);
						
Err				INetLibSockStatus(Word libRefnum, Handle socketH, WordPtr statusP,
							Err* sockErrP, Boolean* inputReadyP, Boolean* outputReadyP)
						SYS_TRAP(inetLibTrapSockStatus);
						

Err				INetLibSockSettingGet(Word libRefnum, Handle socketH, 
							Word /*INetSockSettingEnum*/ setting,
							VoidPtr bufP, WordPtr bufLenP)
						SYS_TRAP(inetLibTrapSockSettingGet);

Err				INetLibSockSettingSet(Word libRefnum, Handle socketH, 
							Word /*INetSockSettingEnum*/ setting,
							VoidPtr bufP, Word bufLen)
						SYS_TRAP(inetLibTrapSockSettingSet);


Err				INetLibSockConnect(Word libRefnum, Handle sockH, BytePtr hostnameP,
							Word port, SDWord timeou)
						SYS_TRAP(inetLibTrapSockConnect);

//--------------------------------------------------
// HTTP specific calls
//--------------------------------------------------

Err				INetLibSockHTTPReqCreate(Word libRefnum, Handle sockH,
							BytePtr verbP, BytePtr resNameP, BytePtr refererP)
						SYS_TRAP(inetLibTrapSockHTTPReqCreate);
						
Err				INetLibSockHTTPAttrSet(Word libRefnum, Handle sockH, 
							Word /*inetHTTPAttrEnum*/ attr, Word attrIndex, 
							BytePtr bufP,  Word bufLen, Word flags)
						SYS_TRAP(inetLibTrapSockHTTPAttrSet);
						
Err				INetLibSockHTTPReqSend(Word libRefnum, Handle sockH, 
							VoidPtr writeP, DWord writeLen, 
							SDWord timeout)
						SYS_TRAP(inetLibTrapSockHTTPReqSend);
						
Err				INetLibSockHTTPAttrGet(Word libRefnum, Handle sockH, 
							Word /*inetHTTPAttrEnum*/ attr, Word attrIndex,
							VoidPtr bufP, DWordPtr bufLenP)
						SYS_TRAP(inetLibTrapSockHTTPAttrGet);
						


//--------------------------------------------------
// Utilities
//--------------------------------------------------

Err				INetLibURLCrack(Word libRefnum, BytePtr urlTextP, INetURLType* urlP)
						SYS_TRAP(inetLibTrapURLCrack);

Err 				INetLibURLsAdd(Word libRefnum, CharPtr baseURLStr, CharPtr embeddedURLStr, 
						CharPtr resultURLStr, WordPtr resultLenP)
						SYS_TRAP(inetLibTrapURLsAdd);

Int 				INetLibURLsCompare(Word libRefnum, CharPtr URLStr1, CharPtr URLStr2)
						SYS_TRAP(inetLibTrapURLsCompare);

Err 				INetLibURLGetInfo(Word libRefnum, Handle inetH,
						BytePtr urlTextP, INetURLInfoType* urlInfoP)
						SYS_TRAP(inetLibTrapURLGetInfo);

//--------------------------------------------------
// Cache interface
//--------------------------------------------------

Err				INetLibCacheList(Word libRefnum, Handle inetH,
							BytePtr cacheIndexURLP, WordPtr indexP, ULongPtr uidP,
							INetCacheEntryP cacheP)
						SYS_TRAP(inetLibTrapCacheList);

Err 				INetLibCacheGetObject(Word libRefnum, Handle clientParamH, BytePtr urlTextP,
							ULong uniqueID, INetCacheInfoPtr cacheInfoP )
						SYS_TRAP(inetLibTrapCacheGetObject);

//--------------------------------------------------
// Configuration Calls
//--------------------------------------------------
Err				INetLibConfigMakeActive( Word refNum, Handle inetH, 
							Word configIndex)
						SYS_TRAP(inetLibConfigMakeActive);
						
Err				INetLibConfigList( Word refNum, INetConfigNameType nameArray[],
							WordPtr arrayEntriesP)
						SYS_TRAP(inetLibConfigList);
						
Err				INetLibConfigIndexFromName( Word refNum, INetConfigNamePtr nameP,
							WordPtr indexP)
						SYS_TRAP(inetLibConfigIndexFromName);
						
Err				INetLibConfigDelete( Word refNum, Word index)
						SYS_TRAP(inetLibConfigDelete);
						
Err				INetLibConfigSaveAs( Word refNum, Handle inetH, 
							INetConfigNamePtr nameP)
						SYS_TRAP(inetLibConfigSaveAs);
						
Err				INetLibConfigRename( Word refNum, Word index,
							INetConfigNamePtr newNameP)
						SYS_TRAP(inetLibConfigRename);

Err				INetLibConfigAliasSet( Word refNum, Word configIndex,
							Word aliasToIndex)
						SYS_TRAP(inetLibConfigAliasSet);

Err				INetLibConfigAliasGet( Word refNum, Word aliasIndex,
							WordPtr indexP, BooleanPtr isAnotherAliasP)
						SYS_TRAP(inetLibConfigAliasGet);



#endif 	//__INETLIB_H__
