/***********************************************************************
 *
 * Copyright (c) Palm Computing 1998 -- All Rights Reserved
 *
 * PROJECT:  Pilot Encryption Library
 * FILE:     SecurityMgr.h
 * AUTHOR:   Ron Marianetti 5/14/98
 *
 * DESCRIPTION:
 *   This header file contains equates for the Security Library. 
 *
 **********************************************************************/
#ifndef     __SECURITY_MGR_H__
#define     __SECURITY_MGR_H__


#include <Common.h>


//********************************************************************************
// Constants
//********************************************************************************
#define secLibType      'libr'      // PalmOS Library database type and creator
#define secLibCreator   'secl'

// Name of the SecurityLib
#define	secLibName		"Security.lib"		// pass in to SysLibFind()
        
// This is how much working space is need to decrypt in place over and
// above the # of bytes of ciphertext
// NOTE: This is a hack for the first encryption scheme until SecDecryptStart
//   is fixed....
#define secDecWorkingSpace  16


// Buffer size for calling SecLibGetSignedPublicKey()
#define secSignedPublicKeyBufSize	256


// Encryption param size (derived from totalEncParamSize in SecurityPrv.h)
#define secEncParamSize				7



//********************************************************************************
// Enums
//********************************************************************************

/* Enum for SecEncryptBegin */
typedef enum {
    secUserServer = 0,
    secUserClient,
    secUserClientNoSeq
    } SecUserEnum;


//********************************************************************************
// Error Codes
//********************************************************************************
#if PLATFORM_TYPE == PLATFORM_SERVER
  #define secErrorClass    0xC1040000  // Value of msgErrSecErrorStart in Messages.h
  #ifndef SYS_TRAP 
    #define SYS_TRAP(trapNum)
  #endif  
#endif

#define   secErrSuccess         0
#define   secErrGeneral         (secErrorClass | 1)
#define   secErrCerticom        (secErrorClass | 2)
#define   secErrUsage           (secErrorClass | 3)
#define   secErrMicMismatch     (secErrorClass | 4)
#define   secErrSeqNum          (secErrorClass | 5)
#define   secErrKeyNotFound     (secErrorClass | 6)
#define   secErrKeyVerMismatch  (secErrorClass | 7)
#define   secErrBfrTooSmall     (secErrorClass | 8)
#define   secErrDataSize        (secErrorClass | 9)   // Error in call to sb_dataSize
#define   secErrDataTooLarge    (secErrorClass | 10)  // datasize too large
#define   secErrInitialize      (secErrorClass | 11)  // Error in call to sb_initialize
#define   secErrMemory          (secErrorClass | 12)  // Cannot allocate needed memory.
#define   secErrAlreadyOpen     (secErrorClass | 13)  // Warning: library already open
#define   secErrNotOpen         (secErrorClass | 14)  // LIbrary not open
#define   secErrDBInit          (secErrorClass | 15)  // Could not init needed database.
#define	  secErrLibClosed		(secErrorClass | 16)  // JB 9/24/98 if the Mobitex stack needs
													  // to use the library, it will close it when
													  // it's done.  This is to notify Activate
													  // in case it had it open.


//********************************************************************************
// Security Library Traps
//********************************************************************************
#if PLATFORM_TYPE == PLATFORM_VIEWER
typedef enum {
  secLibTrapGetContextSize  = sysLibTrapCustom,
  secLibTrapResetSeqNum,
  secLibTrapGetSignedPublicKey,
  secLibTrapProcessSignedKey,
  secLibTrapEncryptBegin,
  secLibTrapEncrypt,
  secLibTrapEncryptEnd,
  secLibTrapDecryptBegin,
  secLibTrapDecrypt,
  secLibTrapDecryptEnd,
  secLibTrapDebug,
  secLibTrapRamEncrypt,
  secLibTrapRamDecrypt,
  secLibTrapSetKeyPath
  } secLibTrapNumberEnum;
#endif


//********************************************************************************
// Macros
//********************************************************************************

// This macro can be used to locate the refNum of the Security library
#if PLATFORM_TYPE == PLATFORM_VIEWER
  #if EMULATION_LEVEL == EMULATION_NONE                               
    #define SecLibGetRefnum(err, ref) {                         \
      if (SysLibFind("Security.lib", &ref))                     \
        err = SysLibLoad(secLibType, secLibCreator, &ref);      \
      else err = 0;                                             \
    }   

  #else //EMULATION_LEVEL == EMULATION_NONE   
    #define SecLibGetRefnum(err, ref) {                         \
      if (SysLibFind("Security.lib", &ref))                     \
        err = SysLibInstall(SecLibPrvInstallDispatcher, &ref);  \
      else err = 0;                                             \
    }           
#endif
    
#else
  #define SecLibGetRefnum(err,ref)    {ref = 0; err = 0;}
#endif


//********************************************************************************
// Function Prototypes
//********************************************************************************
#if PLATFORM_TYPE == PLATFORM_SERVER
  #ifdef _WIN32
    // This is the Elaine NT Server case
    #ifdef _CTPSEC10_EXPORT_
      #define SEC_EXPORT   _declspec(dllexport)
    #else
      #define SEC_EXPORT   _declspec(dllimport)
    #endif

    extern SEC_EXPORT int SecVerbose;    /* 0: Disabled, Non-zero: Enabled */
  #else
    // And this is the George Solaris server case
    #define SEC_EXPORT
    extern int SecVerbose;
  #endif
#else
  // And this is the Pilot PalmOS case
  #define SEC_EXPORT
#endif

#ifdef __cplusplus
extern "C" {
#endif

SEC_EXPORT  DWord   SecLibOpen (Word libRefNum)
                    SYS_TRAP(sysLibTrapOpen);
                    
SEC_EXPORT  DWord   SecLibClose (Word libRefNum)
                    SYS_TRAP(sysLibTrapClose);
                    
SEC_EXPORT  DWord   SecLibSleep (Word libRefNum)
                    SYS_TRAP(sysLibTrapSleep);
                    
SEC_EXPORT  DWord   SecLibWake (Word libRefNum)
                    SYS_TRAP(sysLibTrapWake);

SEC_EXPORT  DWord SecLibDebug(Word libRefNum, Boolean secDebugEnable)
                    SYS_TRAP(secLibTrapDebug);

SEC_EXPORT  DWord   SecLibGetContextSize (Word libRefNum, DWord *contextSizeP, 
                    Word encScheme)
                    SYS_TRAP (secLibTrapGetContextSize);

SEC_EXPORT  DWord   SecLibResetSeqNum (Word libRefNum, DWord serverID, DWord newSeqNum)
                    SYS_TRAP (secLibTrapResetSeqNum);

SEC_EXPORT  DWord   SecLibGetSignedPublicKey (Word libRefNum, Word *sigCountP, 
                    Byte *sigBfrP, Word *sigVersionP)
                    SYS_TRAP (secLibTrapGetSignedPublicKey);

SEC_EXPORT  DWord   SecLibProcessSignedKey (Word libRefNum, Byte *secContextP,  
                    DWord serverID, Word sigCount,  Byte *sigBfrP)
                    SYS_TRAP (secLibTrapProcessSignedKey);

SEC_EXPORT  DWord   SecLibEncryptBegin (Word libRefNum, Byte *secContextP, 
                    DWord serverID, Word encScheme,  Word /*SecUserEnum*/ user, 
                    DWord *ciphertextCountP,  Byte *ciphertextBfrP)
                    SYS_TRAP (secLibTrapEncryptBegin);

SEC_EXPORT  DWord   SecLibEncrypt (Word libRefNum, Byte *secContextP, 
                    DWord plaintextCount,  Byte *plaintextBfrP, 
                    DWord *ciphertextCountP,  Byte *ciphertextBfrP)
                    SYS_TRAP (secLibTrapEncrypt);

SEC_EXPORT  DWord   SecLibEncryptEnd (Word libRefNum, Byte *secContextP, 
                    DWord *ciphertextCountP, Byte *ciphertextBfrP)
                    SYS_TRAP (secLibTrapEncryptEnd);

SEC_EXPORT  DWord   SecLibDecryptBegin (Word libRefNum, Byte *secContextP, 
                    Word encScheme,
                    DWord ciphertextCount, Byte *ciphertextBfrP,
                    DWord *plaintextCountP, Byte *plaintextBfrP,
                    DWord* sequenceNumberP, Word sequenceTolerance)
                    SYS_TRAP (secLibTrapDecryptBegin);

SEC_EXPORT  DWord   SecLibDecrypt (Word libRefNum, Byte *secContextP, 
                    DWord ciphertextCount, 
                    Byte *ciphertextBfrP, DWord *plaintextCountP, 
                    Byte *plaintextBfrP)
                    SYS_TRAP (secLibTrapDecrypt);

SEC_EXPORT  DWord   SecLibDecryptEnd (Word libRefNum, Byte *secContextP, 
                    DWord *plaintextCountP, Byte *plaintextBfrP)
                    SYS_TRAP (secLibTrapDecryptEnd);


// Prototype needed under Simulator for installing the Security Library
#if PLATFORM_TYPE == PLATFORM_VIEWER
Err               SecLibPrvInstallDispatcher(UInt refNum, SysLibTblEntryPtr entryP);
#endif

/* BellSouth functions for DES 56-bit encryption */
int SecRamEncrypt(Word RefLibNum, void *plaintextBfrP, Word plaintextCount,
                  void *ciphertextBfrP, Word *ciphertextCountP, void *keyBfrP,
                  Word keyCount)
    SYS_TRAP (secLibTrapRamEncrypt);

int SecRamDecrypt (Word RefLibNum, void *ciphertextBfrP, Word ciphertextCount,
                   void *plaintextBfrP, Word *plaintextCountP, void *keyBfrP,
                   Word keyCount)
    SYS_TRAP (secLibTrapRamDecrypt);

// Special function to set path to public key files (Elaine and George)
SEC_EXPORT  DWord   SecLibSetKeyPath(Word libRefNum, Byte *secKeyPath,
                    Word secClearCache)
                    SYS_TRAP (secLibTrapSetKeyPath);

#ifdef __cplusplus
}
#endif

#endif  //__ENCRYPT_MGR_H__

