/***********************************************************************
 *
 *	Copyright (c) Palm Computing 1996 -- All Rights Reserved
 *
 * PROJECT:  Pilot TCP/IP Library
 * FILE:     NetPlugIn.h
 * AUTHOR:	 Ron Marianetti 2/14/96
 *
 * DESCRIPTION:
 *	  This header file is supplied for inclusion by Net Library Plug-ins. 
 *
 **********************************************************************/
#ifndef 	__NETPLUGIN_H__
#define	__NETPLUGIN_H__


// Net Library plug-ins register the socket domain and protocols they support
// by creating features of the following creator ID. The feature number contains
//  the domain in the upper 4 bits, socket type in the next 4 bits, and the
//  protocol in the lower byte. The feature value contains the library refnum 
//  of the plug-in 
#define		netPlugInFtrCreator	'netp'		// creatorID of plug-in features. 




/********************************************************************
 * Net Library Plug-in functions. 
 ********************************************************************/
#pragma mark PlugInFunctions

typedef enum {
	npiLibTrapInfo = sysLibTrapCustom,
	
	npiLibTrapSocketOpen,
	npiLibTrapSocketClose,
	
	npiLibTrapSocketOptionSet,
	npiLibTrapSocketOptionGet,
	npiLibTrapSocketBind,
	npiLibTrapSocketConnect,
	npiLibTrapSocketListen,
	npiLibTrapSocketAccept,
	npiLibTrapSocketShutdown,
	npiLibTrapSocketAddr,
	
	npiLibTrapSocketNativeRefnum,
	npiLibTrapSocketReady,
	
	npiLibTrapSocketReceivePB,
	npiLibTrapSocketSendPB,

	npiLibTrapLast
	} NPILibTrapNumberEnum;


//--------------------------------------------------
// Library initialization, shutdown, sleep and wake
//--------------------------------------------------
Err				NPILibOpen (Word libRefnum)
						SYS_TRAP(sysLibTrapOpen);
					
Err				NPILibClose (Word libRefnum)
						SYS_TRAP(sysLibTrapClose);
					
Err				NPILibSleep (Word libRefnum)
						SYS_TRAP(sysLibTrapSleep);
					
Err				NPILibWake (Word libRefnum)
						SYS_TRAP(sysLibTrapWake);
					
//--------------------------------------------------
// Plug-in Info
// This call enables the NetLib to check the implementaton level of the plug-in.
// The current implementation level is 1. If new calls are added to the plug-in
//  interface later, the implementation level will increase. 
//--------------------------------------------------
Err				NPILibInfo(Word libRefnum, Char verString[16],
							WordPtr implementationLevelP)
						SYS_TRAP(npiLibTrapInfo);


//--------------------------------------------------
// Socket creation and option setting
//--------------------------------------------------


NetSocketRef	NPILibSocketOpen(Word libRefnum, NetSocketAddrEnum domain, 
							NetSocketTypeEnum type, SWord protocol, SDWord timeout, 
							Err* errP, NetSocketRef socket)
						SYS_TRAP(npiLibTrapSocketOpen);

SWord				NPILibSocketClose(Word libRefnum, NetSocketRef socket, SDWord timeout, 
							Err* errP)
						SYS_TRAP(npiLibTrapSocketClose);
						
SWord				NPILibSocketOptionSet(Word libRefnum, NetSocketRef socket,
							Word /*NetSocketOptLevelEnum*/ level, Word /*NetSocketOptEnum*/ option, 
							VoidPtr optValueP, Word optValueLen,
							SDWord timeout, Err* errP)
						SYS_TRAP(npiLibTrapSocketOptionSet);
						
SWord				NPILibSocketOptionGet(Word libRefnum, NetSocketRef socket,
							Word /*NetSocketOptLevelEnum*/ level, Word /*NetSocketOptEnum*/ option,
							VoidPtr optValueP, WordPtr optValueLenP,
							SDWord timeout, Err* errP)
						SYS_TRAP(npiLibTrapSocketOptionGet);
						

//--------------------------------------------------
// Socket Control
//--------------------------------------------------

SWord				NPILibSocketBind(Word libRefnum, NetSocketRef socket,
							NetSocketAddrType* sockAddrP, SWord addrLen, SDWord timeout, 
							Err* errP)
						SYS_TRAP(npiLibTrapSocketBind);
						
SWord				NPILibSocketConnect(Word libRefnum, NetSocketRef socket,
							NetSocketAddrType* sockAddrP, SWord addrLen, SDWord timeout, 
							Err* errP)
						SYS_TRAP(npiLibTrapSocketConnect);
						
SWord				NPILibSocketListen(Word libRefnum, NetSocketRef socket,
							Word	queueLen, SDWord timeout, Err* errP)
						SYS_TRAP(npiLibTrapSocketListen);
						
SWord				NPILibSocketAccept(Word libRefnum, NetSocketRef socket,
							NetSocketAddrType* sockAddrP, SWord* addrLenP, SDWord timeout,
							Err* errP)
						SYS_TRAP(npiLibTrapSocketAccept);

SWord				NPILibSocketShutdown(Word libRefnum, NetSocketRef socket, 
							SWord /*NetSocketDirEnum*/ direction, SDWord timeout, Err* errP)
						SYS_TRAP(npiLibTrapSocketShutdown);

SWord				NPILibSocketAddr(Word libRefnum, NetSocketRef socketRef,
							NetSocketAddrType* locAddrP, SWord* locAddrLenP, 
							NetSocketAddrType* remAddrP, SWord* remAddrLenP, 
							SDWord timeout, Err* errP)
						SYS_TRAP(npiLibTrapSocketAddr);
						
						

//--------------------------------------------------
// Select Support.
// This call is used by NetLib during select(). It should return non-zero
//  in *readyP if the socket is ready for I/O. The 'write' parameter is true
//  when checking for readiness for writing and false when checking for readiness
//  for reading. If *readyP is not true on exit, the NetLib will block on the
//  the native socket specified in *nativeSocketP and then call this routine
//  again when *nativeSocketP has data to determine if 'socket' is ready also. 
//--------------------------------------------------
NetSocketRef	NPILibSocketNativeRefnum(Word libRefNum, NetSocketRef socketRef)
						SYS_TRAP(npiLibTrapSocketNativeRefnum);

Boolean			NPILibSocketReady(Word libRefNum, NetSocketRef socket,
							Word write)
						SYS_TRAP(npiLibTrapSocketReady);

//--------------------------------------------------
// Sending and Receiving
//--------------------------------------------------
SWord				NPILibSocketReceivePB(Word libRefNum, NetSocketRef socket, VoidPtr recordP,
							NetIOParamType* pbP, Word flags, SDWord timeout, Err* errP)
						SYS_TRAP(npiLibTrapSocketReceivePB);


SWord				NPILibSocketSendPB(Word libRefNum, NetSocketRef socket,
							NetIOParamType* pbP, Word flags, SDWord timeout, Err* errP)
						SYS_TRAP(npiLibTrapSocketSendPB);

#endif 	//__NETPLUGIN_H__
