/*********************************************************************
 *
 * Copyright (c) 1998
 * 3Com/Palm Computing Division.  All rights reserved.
 *
 * Portions copyright (c) 1996-1997
 * Greg Hewgill.  All rights reserved.
 *
 * Portions copyright (c) 1995-1997
 * Illume Software, Inc.  All rights reserved.
 *
 * Portions copyright (c) 1995-1997
 * Bernd Schmidt.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *
 *		This product includes software developed by 3Com and its
 *		contributors.
 *
 * 4. Neither 3Com nor the names of its contributors may be used to
 *    endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE 3COM AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 3COM OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 *******************************************************************/

#include "ByteSwapping.h"

#if defined(_WIN32)	// Windows build of PilotDebugger

unsigned long	flipBits[] =
{
	/* 0000 -> 0000 */	0x00,
	/* 0001 -> 1000 */	0x08,
	/* 0010 -> 0100 */	0x04,
	/* 0011 -> 1100 */	0x0C,
	/* 0100 -> 0010 */	0x02,
	/* 0101 -> 1010 */	0x0A,
	/* 0110 -> 0110 */	0x06,
	/* 0111 -> 1110 */	0x0E,
	/* 1000 -> 0001 */	0x01,
	/* 1001 -> 1001 */	0x09,
	/* 1010 -> 0101 */	0x05,
	/* 1011 -> 1101 */	0x0D,
	/* 1100 -> 0011 */	0x03,
	/* 1101 -> 1011 */	0x0B,
	/* 1110 -> 0111 */	0x07,
	/* 1111 -> 1111 */	0x0F
};

void Canonical (BreakpointType& bp)
{
	Canonical (bp.addr);

	Canonical (bp.enabled);
	Canonical (bp.installed);
}


void Canonical (CardHeaderType& cardHdr)
{
	Canonical (cardHdr.initStack);
	Canonical (cardHdr.resetVector);
	Canonical (cardHdr.signature);
	Canonical (cardHdr.hdrVersion);
	Canonical (cardHdr.flags);
	Canonical (cardHdr.version);
	Canonical (cardHdr.creationDate);
	Canonical (cardHdr.numRAMBlocks);
	Canonical (cardHdr.blockListOffset);
	Canonical (cardHdr.readWriteParmsOffset);
	Canonical (cardHdr.readWriteParmsSize);
	Canonical (cardHdr.readOnlyParmsOffset);
	Canonical (cardHdr.bigROMOffset);
	Canonical (cardHdr.checksumBytes);
	Canonical (cardHdr.checksumValue);
	Canonical (cardHdr.version);
}


void Canonical (DatabaseHdrType& db)
{
	Canonical (db.attributes);
	Canonical (db.version);
	Canonical (db.creationDate);
	Canonical (db.modificationDate);
	Canonical (db.lastBackupDate);
	Canonical (db.modificationNumber);
	Canonical (db.appInfoID);
	Canonical (db.sortInfoID);
	Canonical (db.type);
	Canonical (db.creator);
	Canonical (db.uniqueIDSeed);
	Canonical (db.recordList.nextRecordListID);
	Canonical (db.recordList.numRecords);
}


void Canonical (EventType& event)
{
	// Byte swapping of this enum is only supported when using the
	//  Emulator's own include files - not the standard PalmOS ones
#ifdef BUILDING_EMULATOR
	Canonical (event.eType);
#endif
	Canonical (event.penDown);
	Canonical (event.screenX);
	Canonical (event.screenY);

	switch(event.eType)
	{
		case nilEvent:
			break;

		case penDownEvent:
			Canonical (event.data.penUp.start);
			Canonical (event.data.penUp.end);
			break;

		case penUpEvent:
			Canonical (event.data.penUp.start);
			Canonical (event.data.penUp.end);
			break;

		case penMoveEvent:
			Canonical (event.data.penUp.start);
			Canonical (event.data.penUp.end);
			break;

		case keyDownEvent:
			Canonical (event.data.keyDown.chr);
			Canonical (event.data.keyDown.keyCode);
			Canonical (event.data.keyDown.modifiers);
			break;

		case winEnterEvent:
			Canonical (event.data.winEnter.enterWindow);
			Canonical (event.data.winEnter.exitWindow);
			break;

		case winExitEvent:
			Canonical (event.data.winExit.enterWindow);
			Canonical (event.data.winExit.exitWindow);
			break;

		case ctlEnterEvent:
			Canonical (event.data.ctlEnter.controlID);
			Canonical (event.data.ctlEnter.pControl);
			break;

		case ctlExitEvent:
			break;

		case ctlSelectEvent:
			Canonical (event.data.ctlSelect.controlID);
			Canonical (event.data.ctlSelect.pControl);
			Canonical (event.data.ctlSelect.on);
			break;

		case ctlRepeatEvent:
			Canonical (event.data.ctlRepeat.controlID);
			Canonical (event.data.ctlRepeat.pControl);
			Canonical (event.data.ctlRepeat.time);
			break;

		case lstEnterEvent:
			Canonical (event.data.lstEnter.listID);
			Canonical (event.data.lstEnter.pList);
			Canonical (event.data.lstEnter.selection);
			break;

		case lstSelectEvent:
			Canonical (event.data.lstSelect.listID);
			Canonical (event.data.lstSelect.pList);
			Canonical (event.data.lstSelect.selection);
			break;

		case lstExitEvent:
			Canonical (event.data.lstExit.listID);
			Canonical (event.data.lstExit.pList);
			break;

		case popSelectEvent:
			Canonical (event.data.popSelect.controlID);
			Canonical (event.data.popSelect.controlP);
			Canonical (event.data.popSelect.listID);
			Canonical (event.data.popSelect.listP);
			Canonical (event.data.popSelect.selection);
			Canonical (event.data.popSelect.priorSelection);
			break;

		case fldEnterEvent:
			Canonical (event.data.fldEnter.fieldID);
			Canonical (event.data.fldEnter.pField);
			break;

		case fldHeightChangedEvent:
			Canonical (event.data.fldHeightChanged.fieldID);
			Canonical (event.data.fldHeightChanged.pField);
			Canonical (event.data.fldHeightChanged.newHeight);
			Canonical (event.data.fldHeightChanged.currentPos);
			break;

		case fldChangedEvent:
			Canonical (event.data.fldChanged.fieldID);
			Canonical (event.data.fldChanged.pField);
			break;

		case tblEnterEvent:
			Canonical (event.data.tblEnter.tableID);
			Canonical (event.data.tblEnter.pTable);
			Canonical (event.data.tblEnter.row);
			Canonical (event.data.tblEnter.column);
			break;

		case tblSelectEvent:
			Canonical (event.data.tblEnter.tableID);
			Canonical (event.data.tblEnter.pTable);
			Canonical (event.data.tblEnter.row);
			Canonical (event.data.tblEnter.column);
			break;

		case daySelectEvent:
			Canonical (event.data.daySelect.pSelector);
			Canonical (event.data.daySelect.selection);
			Canonical (event.data.daySelect.useThisDate);
			break;

		case menuEvent:
			Canonical (event.data.menu.itemID);
			break;

		case appStopEvent:
			break;

		case frmLoadEvent:
			Canonical (event.data.frmLoad.formID);
			break;

		case frmOpenEvent:
			Canonical (event.data.frmOpen.formID);
			break;

		case frmGotoEvent:
			Canonical (event.data.frmGoto.formID);
			Canonical (event.data.frmGoto.recordNum);
			Canonical (event.data.frmGoto.matchPos);
			Canonical (event.data.frmGoto.matchLen);
			Canonical (event.data.frmGoto.matchFieldNum);
			Canonical (event.data.frmGoto.matchCustom);
			break;

		case frmUpdateEvent:
			Canonical (event.data.frmUpdate.formID);
			Canonical (event.data.frmUpdate.updateCode);
			break;

		case frmSaveEvent:
			break;

		case frmCloseEvent:
			Canonical (event.data.frmClose.formID);
			break;

		case frmTitleEnterEvent:
			Canonical (event.data.frmTitleEnter.formID);
			break;

		case frmTitleSelectEvent:
			Canonical (event.data.frmTitleSelect.formID);
			break;

		case tblExitEvent:
			Canonical (event.data.tblExit.tableID);
			Canonical (event.data.tblExit.pTable);
			Canonical (event.data.tblExit.row);
			Canonical (event.data.tblExit.column);
			break;

		case sclEnterEvent:
			Canonical (event.data.sclEnter.scrollBarID);
			Canonical (event.data.sclEnter.pScrollBar);
			break;

		case sclExitEvent:
			Canonical (event.data.sclExit.scrollBarID);
			Canonical (event.data.sclExit.pScrollBar);
			Canonical (event.data.sclExit.value);
			Canonical (event.data.sclExit.newValue);
			break;

		case sclRepeatEvent:
			Canonical (event.data.sclRepeat.scrollBarID);
			Canonical (event.data.sclRepeat.pScrollBar);
			Canonical (event.data.sclRepeat.value);
			Canonical (event.data.sclRepeat.newValue);
			Canonical (event.data.sclRepeat.time);
			break;
	}
}


void Canonical (FieldAttrType& attr)
{
	Canonical (*(unsigned short*) &attr);

		unsigned short	v = *(unsigned short*) &attr;

		/*
			======== BITFIELD LAYOUT CHEAT-SHEET ========

			typedef struct {
				Word usable			:1;
				Word visible		:1;
				Word editable		:1;
				Word singleLine		:1;
				Word hasFocus       :1;
				Word dynamicSize	:1;
				Word insPtVisible	:1;
				Word dirty			:1;
				Word underlined		:2;
				Word justification	:2;
				Word autoShift		:1;
				Word hasScrollBar	:1;
				Word numeric		:1;
			} FieldAttrType;

			// On the Mac:

			|---------- high byte ----------|---------- low byte -----------|

			 15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0  
			  7   6   5   4   3   2   1   0   7   6   5   4   3   2   1   0
			+---+---+---+---+---+---+---+---+-------+-------+---+---+---+---+
			| u | v | e | s | h | d | i | d |   u   |   j   | a | h | n | * |
			+---+---+---+---+---+---+---+---+-------+-------+---+---+---+---+


			// On Windows (in-register representation):

			|---------- high byte ----------|---------- low byte -----------|

			 15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0  
			  7   6   5   4   3   2   1   0   7   6   5   4   3   2   1   0
			+---+---+---+---+-------+-------+---+---+---+---+---+---+---+---+
			| * | n | h | a |   j   |   u   | d | i | d | h | s | e | v | u |
			+---+---+---+---+-------+-------+---+---+---+---+---+---+---+---+
		*/

		v = (unsigned short)
			((flipBits[ (v >>  0) & 0x0F ] << 12) |
			 (flipBits[ (v >>  4) & 0x0F ] <<  8) |
			 (flipBits[ (v >>  8) & 0x0F ] <<  4) |
			 (flipBits[ (v >> 12) & 0x0F ] <<  0));

		*(unsigned short*) &attr = v;
}



void Canonical (M68KRegsType& debuggerRegs)
{
	Canonical (debuggerRegs.d[0]);
	Canonical (debuggerRegs.d[1]);
	Canonical (debuggerRegs.d[2]);
	Canonical (debuggerRegs.d[3]);
	Canonical (debuggerRegs.d[4]);
	Canonical (debuggerRegs.d[5]);
	Canonical (debuggerRegs.d[6]);
	Canonical (debuggerRegs.d[7]);

	Canonical (debuggerRegs.a[0]);
	Canonical (debuggerRegs.a[1]);
	Canonical (debuggerRegs.a[2]);
	Canonical (debuggerRegs.a[3]);
	Canonical (debuggerRegs.a[4]);
	Canonical (debuggerRegs.a[5]);
	Canonical (debuggerRegs.a[6]);
//	Canonical (debuggerRegs.a[7]);

	Canonical (debuggerRegs.usp);
	Canonical (debuggerRegs.ssp);
	Canonical (debuggerRegs.sr);
	Canonical (debuggerRegs.pc);
}



void Canonical (PointType& p)
{
	Canonical (p.x);
	Canonical (p.y);
}


void Canonical (RecordEntryType& entry)
{
	Canonical (entry.localChunkID);
	Canonical (entry.attributes);
//	Canonical (entry.uniqueID);	// array of 3 bytes
}


void Canonical (RectangleType& r)
{
	Canonical (r.topLeft);
	Canonical (r.extent);
}


void Canonical (RsrcEntryType& entry)
{
	Canonical (entry.type);
	Canonical (entry.id);
	Canonical (entry.localChunkID);
}


void Canonical (SlkPktHeaderType& hdr)
{
	Canonical (hdr.signature1);
	Canonical (hdr.signature2);
	Canonical (hdr.dest);
	Canonical (hdr.src);
	Canonical (hdr.type);
	Canonical (hdr.bodySize);
	Canonical (hdr.transId);
	Canonical (hdr.checksum);
}

void Canonical (SlkPktFooterType& ftr)
{
	Canonical (ftr.crc16);
}

void Canonical (SysNVParamsType& params)
{
	Canonical (params.rtcHours);
	Canonical (params.rtcHourMinSecCopy);
}


void Canonical (SysPktBodyType& packet)
{
	Canonical (packet.command);

	switch (packet.command)
	{
		case sysPktStateCmd:
			Canonical ((SysPktStateCmdType&) packet);
			break;

		case sysPktStateRsp:
			Canonical ((SysPktStateRspType&) packet);
			break;

		case sysPktReadMemCmd:
			Canonical ((SysPktReadMemCmdType&) packet);
			break;

		case sysPktWriteMemCmd:
			Canonical ((SysPktWriteMemCmdType&) packet);
			break;

		case sysPktSingleStepCmd:
			// I don't think there's anything to do here.  I think that
			// single-stepping is performed by the debugger setting the
			// trace bit in the SR and sending a sysPktContinueCmd.
			break;

		case sysPktGetRtnNameCmd:
			Canonical ((SysPktRtnNameCmdType&) packet);
			break;

		case sysPktReadRegsCmd:
			Canonical ((SysPktReadRegsCmdType&) packet);
			break;

		case sysPktWriteRegsCmd:
			Canonical ((SysPktWriteRegsCmdType&) packet);
			break;

		case sysPktContinueCmd:
			Canonical ((SysPktContinueCmdType&) packet);
			break;

		case sysPktRPCCmd:
			Canonical ((SysPktRPCType&) packet);
			break;

		case sysPktGetBreakpointsCmd:
			Canonical ((SysPktGetBreakpointsCmdType&) packet);
			break;

		case sysPktSetBreakpointsCmd:
#ifdef BUILDING_EMULATOR
			Canonical ((SysPktSetBreakpointsCmdType2&) packet);
#endif
			break;

//		case sysPktRemoteUIUpdCmd:
//			// Sent TO debugger; never received FROM debugger.
//			break;

//		case sysPktRemoteEvtCmd:
//			// Sent TO debugger; never received FROM debugger.
//			break;

		case sysPktDbgBreakToggleCmd:
			Canonical ((SysPktDbgBreakToggleCmdType&) packet);
			break;

		case sysPktFlashCmd:
			// Not supported in this release.
			break;

		case sysPktCommCmd:
			// Not supported in this release.
			break;

		case sysPktGetTrapBreaksCmd:
			Canonical ((SysPktGetTrapBreaksCmdType&) packet);
			break;

		case sysPktSetTrapBreaksCmd:
#ifdef BUILDING_EMULATOR
			Canonical ((SysPktSetTrapBreaksCmdType2&) packet);
#endif
			break;

		case sysPktGremlinsCmd:
			// !!! TBD.
			break;

		case sysPktFindCmd:
			Canonical ((SysPktFindCmdType&) packet);
			break;

		case sysPktGetTrapConditionsCmd:
			Canonical ((SysPktGetTrapConditionsCmdType&) packet);
			break;

		case sysPktSetTrapConditionsCmd:
#ifdef BUILDING_EMULATOR
			Canonical ((SysPktSetTrapConditionsCmdType2&) packet);
#endif
			break;

		case sysPktChecksumCmd:
			// Not supported in this release.
			break;

		case sysPktExecFlashCmd:
			// Not supported in this release.
			break;

		case sysPktRemoteMsgCmd:
			// Nothing for us to do.  With a serial link,
			// these are sent to clear the input port.
			break;

		default:
			break;
	}
}


void Canonical (SysPktContinueCmdType& packet)
{
	Canonical (packet.command);

	Canonical (packet.regs);
	Canonical (packet.stepSpy);
	Canonical (packet.ssAddr);
	Canonical (packet.ssCount);
	Canonical (packet.ssCheckSum);
}


void Canonical (SysPktDbgBreakToggleCmdType& packet)
{
	Canonical (packet.command);
}


void Canonical (SysPktFindCmdType& packet)
{
	Canonical (packet.command);

	Canonical (packet.firstAddr);
	Canonical (packet.lastAddr);
	Canonical (packet.numBytes);
	Canonical (packet.caseInsensitive);
}


void Canonical (SysPktFindRspType& packet)
{
	Canonical (packet.command);

	Canonical (packet.addr);
	Canonical (packet.found);
}


void Canonical (SysPktGetBreakpointsCmdType& packet)
{
	Canonical (packet.command);
}


void Canonical (SysPktGetTrapBreaksCmdType& packet)
{
	Canonical (packet.command);
}



void Canonical (SysPktGetTrapConditionsCmdType& packet)
{
	Canonical (packet.command);
}



void Canonical (SysPktReadMemCmdType& packet)
{
	Canonical (packet.command);

	Canonical (packet.address);
	Canonical (packet.numBytes);
}


void Canonical (SysPktReadRegsCmdType& packet)
{
	Canonical (packet.command);
}


void Canonical (SysPktRPCType& packet)
{
	Canonical (packet.command);

	Canonical (packet.trapWord);
	Canonical (packet.resultD0);
	Canonical (packet.resultA0);
	Canonical (packet.numParams);

	// Nothing really to do with the "param" field.  "byRef" and "size"
	// are bytes, so nothing to swap there, and the "data" field is
	// of indeterminate size, so we can't bytewap it either.  That means
	// that we must take care to always treat it as a big-endian value.	
}


void Canonical (SysPktRtnNameCmdType& packet)
{
	Canonical (packet.command);

	Canonical (packet.address);
}


void Canonical (SysPktRtnNameRspType& packet)
{
	Canonical (packet.command);

	Canonical (packet.address);
	Canonical (packet.startAddr);
	Canonical (packet.endAddr);
}



void Canonical (SysPktStateCmdType& packet)
{
	Canonical (packet.command);
}


void Canonical (SysPktStateRspType& packet)
{
	Canonical (packet.command);

	Canonical (packet.resetted);
	Canonical (packet.exceptionId);
	Canonical (packet.reg);

	int ii;
	for (ii = 0; ii < dbgTotalBreakpoints; ++ii)
		Canonical (packet.bp[ii]);

	for (ii = 0; ii < sysPktStateRspInstWords; ++ii)
		Canonical (packet.inst[ii]);

	Canonical (packet.startAddr);
	Canonical (packet.endAddr);
	Canonical (packet.trapTableRev);	
}


void Canonical (SysPktWriteRegsCmdType& packet)
{
	Canonical (packet.command);

	Canonical (packet.reg);
}


void Canonical (SysPktWriteMemCmdType& packet)
{
	Canonical (packet.command);

	Canonical (packet.address);
	Canonical (packet.numBytes);
}

void Canonical (SysPktWriteMemRspType& packet)
{
		Canonical (packet.command);

}

void Canonical (SysPktRemoteEvtCmdType& packet)
{
	Canonical (packet.command);
	Canonical (packet.penDown);				// true if pen down
  	Canonical (packet.penX);				// X location of pen  
  	Canonical (packet.penY);				// Y location of pen  
  	
  	Canonical (packet.keyPress);			// true if key event follows
  	Canonical (packet.keyModifiers);		// keyboard modifiers
  	Canonical (packet.keyAscii);			// key ascii code
  	Canonical (packet.keyCode);				// key virtual code

}

void Canonical (DmOpenInfoType& p)
{
	Canonical(p.next);
	Canonical(p.openCount);
	Canonical(p.ownerTaskID);
//	Canonical(p.exclusive);
//	Canonical(p.writeAccess);
//	Canonical(p.resDB);
	Canonical(p.hdrID);
	Canonical((Ptr&) p.hdrH);
	Canonical((Ptr&)p.hdrP);
	Canonical(p.cardNo);
	Canonical(p.numRecords);
	Canonical(p.hdrMasterP);
	Canonical((Ptr&)p.handleTableP);
}

void Canonical(SysAppInfoType& p)
{
	Canonical(p.cmd);
	Canonical(p.cmdPBP);
	Canonical(p.launchFlags);
	Canonical(p.taskID);
	Canonical((Ptr&)p.codeH);
	Canonical(p.dbP);
	Canonical(p.stackP);
	Canonical(p.globalsChunkP);
	Canonical(p.memOwnerID);
	Canonical(p.dmAccessP);
	Canonical(p.dmLastErr);
	Canonical(p.errExceptionP);
	Canonical(p.a5Ptr);
	Canonical(p.stackEndP);
	Canonical(p.globalEndP);
	Canonical((Ptr&)p.rootP);
	Canonical(p.extraP);

}

// Now up to version 3...
void Canonical(MemHeapHeaderType& p)
{
	Canonical(p.flags);					// heap flags;
	Canonical(p.size);						// size of heap
	Canonical(p.firstFreeChunkOffset);
	Canonical(p.mstrPtrTbl.numEntries);
	Canonical(p.mstrPtrTbl.nextTblOffset);
}
void Canonical(Mem2HeapHeaderType& p)
{
	Canonical(p.flags);					// heap flags;
	Canonical(p.size);						// size of heap
	Canonical(p.mstrPtrTbl.numEntries);
	Canonical(p.mstrPtrTbl.nextTblOffset);
}
void Canonical(Mem1HeapHeaderType& p)
{
	Canonical(p.flags);					// heap flags;
	Canonical(p.size);						// size of heap
	Canonical(p.mstrPtrTbl.numEntries);
	Canonical(p.mstrPtrTbl.nextTblOffset);
}

//<chg 24-Nov-98 RM>
void Canonical(MemChunkHeaderType& p)
{
	DWord size = p.size << 8;
	Canonical(size);
	p.size = size;

	DWord hOffset = p.hOffset << 8;
	Canonical(hOffset);			// offset/2 to handle for movables
	p.hOffset = hOffset;
}
void Canonical(Mem1ChunkHeaderType& p)
{
	Canonical(p.size);				// size of block in bytes, including header
	Canonical(p.lockOwner);			// lock count in high nibble, owner in low
	Canonical(p.flags);				// flags, size adj in low nibble
	Canonical(p.hOffset);			// offset/2 to handle for movables
}

void Canonical(MemMstrPtrTableType& p)
{
	Canonical(p.numEntries);				// number of master pointer entries
	Canonical(p.nextTblOffset);			// offset from start of heap to next table
}
void Canonical(Mem1MstrPtrTableType& p)
{
	Canonical(p.numEntries);				// number of master pointer entries
	Canonical(p.nextTblOffset);			// offset from start of heap to next table
}

void Canonical(DmAccessType& p)
{
	Canonical((Ptr&)p.openP);					// ptr to DmOpenInfoType for database
	Canonical((Ptr&)p.next);						// ptr to next access structure
	Canonical(p.mode);						// mode it was opened in
	Canonical(p.savedModNum);			// saved modification number of database

}

void Canonical(DatabaseDirType& p)
{
	Canonical(p.nextDatabaseListID);		// continuation list, if any
	Canonical(p.numDatabases);				// # of database ID's in this list
	for (int i=0;i<p.numDatabases;i++)
		Canonical(p.databaseID[i]);	// ChunkID of each Database
}

void Canonical(RecordListType& p)
{
	Canonical(p.nextRecordListID);			// local chunkID of next list
	Canonical(p.numRecords);					// number of records in this list
	//Canonical(p.firstEntry);					// array of Record/Rsrc entries 
}

void Canonical(DmSearchStateType& p)
{
	for (int i=0;i<8;i++)
		Canonical(p.info[i]);
}

void Canonical(char**& p)
{
	Canonical((Ptr&) p);
}

void Canonical (SysPktFlashWriteType& p)
{
	Canonical(p.command);								
//	Canonical(p._filler);
	Canonical(p.numBytes);							// -> # of bytes in sector
	Canonical(p.flashBase);							// -> base address of FLASH, or 0 for default
	Canonical(p.srcAddr);								// -> source address of new data
	Canonical(p.dstAddr);								// -> dest address in flash memory
	Canonical(p.manuf);								// <- manufacturer
	Canonical(p.device);								// <- device
	Canonical(p.width);								// <- Flash memory width 
	Canonical(p.filler);
	Canonical(p.ramSpaceP);							// <-  available space in RAM for sector

}

void Canonical (sysPktExecFlashType& p)
{
	Canonical(p.command);								
//	Canonical(p._filler);
	Canonical(p.numBytes);							//  -> # of bytes in sector
	Canonical(p.flashBase);							//  -> base address of FLASH
	Canonical(p.srcAddr);								//  -> source address of new data
	Canonical(p.dstAddr);								//  -> dest address in flash memory
	Canonical(p.codeAddr);							// <-> address of flash code
	Canonical(p.flashResID);							// <-  resource ID of flash write code
	Canonical(p.manuf);								// <-  manufacturer
	Canonical(p.device);								// <-  device
	Canonical(p.devInfo);								// <-  device info
	Canonical(p.remoteCmd);							//  -> remote command (e.g.: flash/erase/both)
	Canonical(p.response);							// <-  response from flash/ident code
	Canonical(p.spareByte);							// <-> filler (byte for word alignment)
	Canonical(p.spareDWord);							// <-> spare
}

void Canonical (SysPktChecksumType& p)
{
	Canonical(p.command);								
//	Canonical(p._filler);
	Canonical(p.firstAddr);									// -> first address to checksum
	Canonical(p.numBytes);									// -> number of bytes to checksum
	Canonical(p.seed);											// -> initial checksum value
	Canonical(p.checksum);									// <- checksum result

}

void Canonical (SysPktCommCmdType& p)
{
	Canonical(p.command);								
//	Canonical(p._filler);
	Canonical(p.set);					// true to change parameters
	Canonical(p.baudRate);			// new baud rate
	Canonical(p.flags);				// new flags

}

void Canonical (SysPktCommRspType& p)
{
	Canonical(p.command);								
//	Canonical(p._filler);
	Canonical(p.baudRate);			// current baud rate
	Canonical(p.flags);				// current flags

}

void Canonical (SysPktRemoteMsgCmdType& p)
{
	Canonical(p.command);								
//	Canonical(p._filler);
//Byte		text;									// variable length text goes here

}

void Canonical (SysPktWriteRegsRspType& p)
{
	Canonical(p.command);								
//	Canonical(p._filler);

}

void Canonical (SysPktSetBreakpointsCmdType& p)
{
	Canonical(p.command);								
//	Canonical(p._filler);
//	for (int i=0 i<dbgTotalBreakpoints`
	//BreakpointType	bp[dbgTotalBreakpoints];

}

void Canonical (SysPktSetBreakpointsRspType& p)
{
	Canonical(p.command);								
//	Canonical(p._filler);
//	for (int i=0 i<dbgTotalBreakpoints`
	//BreakpointType	bp[dbgTotalBreakpoints];

}

void Canonical (SysPktSetTrapBreaksCmdType& p)
{
	Canonical(p.command);								
//	Canonical(p._filler);
//Word		trapBP[dbgTotalTrapBreaks];

}
void Canonical (SysPktSetTrapBreaksRspType& p)
{
	Canonical(p.command);								
//	Canonical(p._filler);
//Word		trapBP[dbgTotalTrapBreaks];

}

void Canonical (SysPktSetTrapConditionsCmdType& p)
{
	Canonical(p.command);								
//	Canonical(p._filler);
	//Word		trapParam[dbgTotalTrapBreaks];
}
void Canonical (SysPktSetTrapConditionsRspType& p)
{
	Canonical(p.command);								
//	Canonical(p._filler);
	//Word		trapParam[dbgTotalTrapBreaks];
}

void Canonical(SysPktDbgBreakToggleRspType& p)
{
	Canonical(p.command);								
//	Canonical(p._filler);
	Canonical(p.newState);

}


//######################################################################
// The following byteswaps can not be used in the PalmOS Emulator because
//  it doesn't include all the PalmOS Equates necessary
//#########################################################################
#ifndef BUILDING_EMULATOR
void Canonical(AlmEntryType& p)
{
	Canonical(p.ref);
	Canonical(p.alarmSeconds);	
	Canonical(p.dbID);
	Canonical(p.cardNo);
//	Canonical((Word&)p.quiet);
}

void Canonical(SysKernelInfoType& p)
{
//	Canonical(p.selector);
	Canonical(p.id);
	switch (p.selector) {
		case sysKernelInfoSelCurTaskInfo:
		case sysKernelInfoSelTaskInfo:
			Canonical(p.param.task.id);							// ID of task
			Canonical(p.param.task.nextID);						// ID of next task
			
			Canonical(p.param.task.tag);							// task tag
			Canonical(p.param.task.status);						// task status
			Canonical(p.param.task.timer);						// # ticks left if task's timeout
			Canonical(p.param.task.timeSlice);					// task's time slice
			Canonical(p.param.task.priority);					// task priority
			Canonical(p.param.task.attributes);					// task attributes
			Canonical(p.param.task.pendingCalls);				// # of pending calls
			Canonical(p.param.task.senderTaskID);				// task ID of message sender
			Canonical(p.param.task.msgExchangeID);				// task's message exchange ID, if any
			
			Canonical(p.param.task.tcbP);							// pointer to TCB
			Canonical(p.param.task.stackP);						// stack pointer of task
			Canonical(p.param.task.stackStart);					// top of task stack (where it started).
			Canonical(p.param.task.stackSize);					// size of stack
		break;
		case sysKernelInfoSelSemaphoreInfo:
			Canonical(p.param.semaphore.id);
			Canonical(p.param.semaphore.nextID);						// ID of next semaphore

			Canonical(p.param.semaphore.tag);							// semaphore tag
			Canonical(p.param.semaphore.initValue);					// semaphore initial value 
														//  >= 0 for counting, -1 for resource
			Canonical(p.param.semaphore.curValue);					// semaphore current value
														//  >0 available
														//   0 not available
														//  -n not available, n tasks waiting
			Canonical(p.param.semaphore.nestLevel);					// 0 for counting, >= 0 for resource
			Canonical(p.param.semaphore.ownerID);						// owner ID
														//   0 for counting or resource free
		break;
		case sysKernelInfoSelTimerInfo:
			Canonical(p.param.timer.id);
			Canonical(p.param.timer.nextID);						// ID of next timer
			
			Canonical(p.param.timer.tag);							// timer tag
			Canonical(p.param.timer.ticksLeft);					// ticks left till timer runs
			Canonical(p.param.timer.period);						// timer period, if periodic
			Canonical(p.param.timer.proc);							// timer procedure
		break;

	}
}
#endif //#ifndef BUILDING_EMULATOR



//######################################################################
// The following byteswaps are only used by the PalmOS Emulator
//#########################################################################
#ifdef BUILDING_EMULATOR
void Canonical (HwrDBallEZType& regs)
{
	Canonical (regs.swID);

	Canonical (regs.csAGroupBase);
	Canonical (regs.csBGroupBase);
	Canonical (regs.csCGroupBase);
	Canonical (regs.csDGroupBase);

	Canonical (regs.csASelect);
	Canonical (regs.csBSelect);
	Canonical (regs.csCSelect);
	Canonical (regs.csDSelect);

	Canonical (regs.emuCS);

	Canonical (regs.pllControl);
	Canonical (regs.pllFreqSel);
	Canonical (regs.pllTest);

	Canonical (regs.intControl);
	Canonical (regs.intMaskHi);
	Canonical (regs.intMaskLo);
	Canonical (regs.intStatusHi);
	Canonical (regs.intStatusLo);
	Canonical (regs.intPendingHi);
	Canonical (regs.intPendingLo);

	Canonical (regs.pwmControl);

	Canonical (regs.tmr1Control);
	Canonical (regs.tmr1Prescaler);
	Canonical (regs.tmr1Compare);
	Canonical (regs.tmr1Capture);
	Canonical (regs.tmr1Counter);
	Canonical (regs.tmr1Status);

	Canonical (regs.wdControl);
	Canonical (regs.wdReference);
	Canonical (regs.wdCounter);

	Canonical (regs.spiMasterData);
	Canonical (regs.spiMasterControl);

	Canonical (regs.uControl);
	Canonical (regs.uBaud);
	Canonical (regs.uReceive);
	Canonical (regs.uTransmit);
	Canonical (regs.uMisc);
	Canonical (regs.uNonIntPresc);

	Canonical (regs.lcdStartAddr);
	Canonical (regs.lcdScreenWidth);
	Canonical (regs.lcdScreenHeight);
	Canonical (regs.lcdCursorXPos);
	Canonical (regs.lcdCursorYPos);
	Canonical (regs.lcdCursorWidthHeight);
	Canonical (regs.lcdContrastControlPWM);

	Canonical (regs.rtcHourMinSec);
	Canonical (regs.rtcAlarm);
	Canonical (regs.rtcWatchDog);
	Canonical (regs.rtcControl);
	Canonical (regs.rtcIntStatus);
	Canonical (regs.rtcIntEnable);
	Canonical (regs.stopWatch);
	Canonical (regs.rtcDay);
	Canonical (regs.rtcDayAlarm);

	Canonical (regs.dramConfig);
	Canonical (regs.dramControl);

	Canonical (regs.emuAddrCompare);   
	Canonical (regs.emuAddrMask);
	Canonical (regs.emuControlCompare);
	Canonical (regs.emuControlMask);
	Canonical (regs.emuControl);
	Canonical (regs.emuStatus);
}

void Canonical (HwrDBallType& regs)
{
	Canonical (regs.csAGroupBase);
	Canonical (regs.csBGroupBase);
	Canonical (regs.csCGroupBase);
	Canonical (regs.csDGroupBase);

	Canonical (regs.csAGroupMask);
	Canonical (regs.csBGroupMask);
	Canonical (regs.csCGroupMask);
	Canonical (regs.csDGroupMask);

	Canonical (regs.csASelect0);
	Canonical (regs.csASelect1);
	Canonical (regs.csASelect2);
	Canonical (regs.csASelect3);

	Canonical (regs.csBSelect0);
	Canonical (regs.csBSelect1);
	Canonical (regs.csBSelect2);
	Canonical (regs.csBSelect3);

	Canonical (regs.csCSelect0);
	Canonical (regs.csCSelect1);
	Canonical (regs.csCSelect2);
	Canonical (regs.csCSelect3);

	Canonical (regs.csDSelect0);
	Canonical (regs.csDSelect1);
	Canonical (regs.csDSelect2);
	Canonical (regs.csDSelect3);

	Canonical (regs.csDebug);

	Canonical (regs.pllControl);
	Canonical (regs.pllFreqSel);
	Canonical (regs.pllTest);

	Canonical (regs.intControl);
	Canonical (regs.intMaskHi);
	Canonical (regs.intMaskLo);
	Canonical (regs.intWakeupEnHi);
	Canonical (regs.intWakeupEnLo);
	Canonical (regs.intStatusHi);
	Canonical (regs.intStatusLo);
	Canonical (regs.intPendingHi);
	Canonical (regs.intPendingLo);

	Canonical (regs.pwmControl);
	Canonical (regs.pwmPeriod);
	Canonical (regs.pwmWidth);
	Canonical (regs.pwmCounter);

	Canonical (regs.tmr1Control);
	Canonical (regs.tmr1Prescaler);
	Canonical (regs.tmr1Compare);
	Canonical (regs.tmr1Capture);
	Canonical (regs.tmr1Counter);
	Canonical (regs.tmr1Status);
	
	Canonical (regs.tmr2Control);
	Canonical (regs.tmr2Prescaler);
	Canonical (regs.tmr2Compare);
	Canonical (regs.tmr2Capture);
	Canonical (regs.tmr2Counter);
	Canonical (regs.tmr2Status);
	
	Canonical (regs.wdControl);
	Canonical (regs.wdReference);
	Canonical (regs.wdCounter);

	Canonical (regs.spiSlave);

	Canonical (regs.spiMasterData);
	Canonical (regs.spiMasterControl);

	Canonical (regs.uControl);
	Canonical (regs.uBaud);
	Canonical (regs.uReceive);
	Canonical (regs.uTransmit);
	Canonical (regs.uMisc);

	Canonical (regs.lcdStartAddr);
	Canonical (regs.lcdScreenWidth);
	Canonical (regs.lcdScreenHeight);
	Canonical (regs.lcdCursorXPos);
	Canonical (regs.lcdCursorYPos);
	Canonical (regs.lcdCursorWidthHeight);
	Canonical (regs.lcdGrayPalette);

	Canonical (regs.rtcHourMinSec);
	Canonical (regs.rtcAlarm);
	Canonical (regs.rtcReserved);
	Canonical (regs.rtcControl);
	Canonical (regs.rtcIntStatus);
	Canonical (regs.rtcIntEnable);
	Canonical (regs.stopWatch);
}

void Canonical (PenBtnInfoType& p)
{
	Canonical (p.boundsR);
	Canonical (p.asciiCode);
	Canonical (p.keyCode);
	Canonical (p.modifiers);
}

void Canonical (Point& p)
{
	Canonical (p.h);
	Canonical (p.v);
}

void Canonical (SysPktGetBreakpointsRspType2& packet)
{
	Canonical (packet.command);

	for (int ii = 0; ii < dbgTotalBreakpoints; ++ii)
		Canonical (packet.bp[ii]);
}


void Canonical (SysPktGetTrapBreaksRspType2& packet)
{
	Canonical (packet.command);

	for (int ii = 0; ii < dbgTotalTrapBreaks; ++ii)
		Canonical (packet.trapBP[ii]);
}

void Canonical (SysPktGetTrapConditionsRspType2& packet)
{
	Canonical (packet.command);

	for (int ii = 0; ii < dbgTotalTrapBreaks; ++ii)
		Canonical (packet.trapParam[ii]);
}

void Canonical (SysPktReadRegsRspType2& packet)
{
	Canonical (packet.command);

	Canonical (packet.reg);
}

void Canonical (SysPktSetBreakpointsCmdType2& packet)
{
	Canonical (packet.command);

	for (int ii = 0; ii < dbgTotalBreakpoints; ++ii)
		Canonical (packet.bp[ii]);
}

void Canonical (SysPktSetTrapBreaksCmdType2& packet)
{
	Canonical (packet.command);

	for (int ii = 0; ii < dbgTotalTrapBreaks; ++ii)
		Canonical (packet.trapBP[ii]);
}


void Canonical (SysPktSetTrapConditionsCmdType2& packet)
{
	Canonical (packet.command);

	for (int ii = 0; ii < dbgTotalTrapBreaks; ++ii)
		Canonical (packet.trapParam[ii]);
}
#endif // #ifdef BUILDING_EMULATOR



#endif	// _WIN32






