/*******************************************************************
 * 							 Pilot Software
 *
 *		 Copyright(c) 1998, Palm Computing Inc., All Rights Reserved
 *
 *-------------------------------------------------------------------
 * FileName:
 *		Transport.cp
 *
 * Description:
 *		Implementation for routines to stream data.
 *
 * History:
 *   	10/30/98	Newt oday.  Modelled after SerialMgr.cp.
 *
 *******************************************************************/

#include <Common.h>

// PalmOS Headers
#include <SerialMgr.h>

// Include Transport equates
#include "SerTransport.h"

// Various Transports
#include "SerLibSockets.h"
#include "SerLibPPC.h"


static int	gType;


// Serial buffer to use
static char		PrvSerBuffer[30*1024];


// -------------------------------------------------------------------
// PPC (Process to Process Communcation) is only available on the Macintosh
// -------------------------------------------------------------------
#if EMULATION_LEVEL != EMULATION_MAC
  Err PPPCOpen(void) {return -1;}
  Err PPPCClose(void) {return -1;}
  Err PPCClearErr(void) {return -1;}
  Err PPCSendWait(Long timeout) {return -1;}
  Err PPCReceiveWait(ULong bytes, Long timeout) {return -1;}
  void PPCReceiveFlush(Long timeout) {}
  Err PPCSetReceiveBuffer(VoidPtr bufP, UInt bufSize) {return -1;}
  ULong PPCSend(VoidPtr bufP, ULong count, Err* errP) {return 0;}
  ULong PPCReceive(VoidPtr bufP, ULong count, Long timeout, Err* errP) {return 0;}
#endif


/***************************************************************
 *	Function:	PrvTranslateTptErr	 
 *
 *	Summary:	Translates a return error code from one of the
 *					SerXXX transport providers into a SerialMgr
 *					error.
 *
 *				The SerXX transport provides don't include SerialMgr.h
 *				because they are written to not be required to include
 *				any PalmOS equates. 
 *
 *				Since they can't return SerialMgr error codes directly.
 *				they instead return one our our generic transport
 *				error codes defined in SerTransport.h (which they can
 *				include. 
 *	  
 *	Parameters:
 *				err		IN	  SerSocket() error from 
 *
 *	Returns:	
 *				appropriate SerialMgr.h error code
 *	
 *	Called By: 
 *				All entry point routines in this module
 *
 *	Notes:
 *	
 *	History:
 *
 ****************************************************************/
static Err	PrvTranslateErr	(int err)
{
	switch (err)
	  {
		case	tptErrNone:			return errNone;
		case	tptErrBadParam:		return serErrBadParam;
		case	tptErrBadPort:		return serErrBadPort;
		case	tptErrNoMem:		return serErrNoMem;
		case	tptErrBadConnID:	return serErrBadConnID;
		case	tptErrTimeOut:		return serErrTimeOut;
		case	tptErrLineErr:		return serErrLineErr;
		case	tptErrAlreadyOpen:	return serErrAlreadyOpen;
		case	tptErrStillOpen:	return serErrStillOpen;
		case	tptErrNotOpen:		return serErrNotOpen;
		case	tptErrNotSupported:	return serErrNotSupported;
		case	tptErrInternal:		return serErrNotSupported;
	  }
  return serErrNotSupported;
}





/***************************************************************
 *	Function:	 
 *
 *	Summary:	  
 *	  
 *	Parameters:
 *
 *	Returns:
 *	
 *	Called By: 
 *
 *	Notes:
 *	
 *	History:
 *
 ****************************************************************/
void	TptSetType			(int type)
{
	gType = type;
}



/***************************************************************
 *	Function:	 
 *
 *	Summary:	  
 *	  
 *	Parameters:
 *
 *	Returns:
 *	
 *	Called By: 
 *
 *	Notes:
 *	
 *	History:
 *
 ****************************************************************/
Err	  TptOpen			(UInt refNum, Word port, DWord baud)
{
	Err	err;
	
	if (gType == tptSerial)
	{
		SerSetMapPort (0, port);
		err = SerOpen (refNum, port, baud);
		if (!err)	
			SerSetReceiveBuffer( refNum, PrvSerBuffer, sizeof(PrvSerBuffer) );

	}
	else if (gType == tptSockets)
	{
		PrvTranslateErr (SerSockStart (true /*connecting*/));
	}
	else if (gType == tptPPC)
	{
		err = PPPCOpen ();
		if (!err)
			PPCSetReceiveBuffer (PrvSerBuffer, sizeof (PrvSerBuffer));
	}

	return 0;
}

/***************************************************************
 *	Function:	 
 *
 *	Summary:	  
 *	  
 *	Parameters:
 *
 *	Returns:
 *	
 *	Called By: 
 *
 *	Notes:
 *	
 *	History:
 *
 ****************************************************************/
Err	  TptClose		(UInt refNum)
{
	if (gType == tptSerial)
	{
		SerSetReceiveBuffer( refNum, 0, 0);
		return SerClose (refNum);
	}
	else if (gType == tptSockets)
	{
		PrvTranslateErr (SerSockStop ());
	}
	else if (gType == tptPPC)
	{
		PPCSetReceiveBuffer (0, 0);
		return PPPCClose();
	}

	return 0;
}



/***************************************************************
 *	Function:	 
 *
 *	Summary:	  
 *	  
 *	Parameters:
 *
 *	Returns:
 *	
 *	Called By: 
 *
 *	Notes:
 *	
 *	History:
 *
 ****************************************************************/
ULong	TptSend				(UInt refNum, VoidPtr bufP, ULong count, Err* errP)
{
	if (gType == tptSerial)
	{
		return SerSend (refNum, bufP, count, errP);
	}
	else if (gType == tptSockets)
	{ 
		long	result;

		result =  SerSockSend (bufP, count, errP);
		*errP = PrvTranslateErr (*errP);
		return result;
	}
	else if (gType == tptPPC)
	{
		return PPCSend(bufP, count, errP);
	}

	return 0;
}


/***************************************************************
 *	Function:	 
 *
 *	Summary:	  
 *	  
 *	Parameters:
 *
 *	Returns:
 *	
 *	Called By: 
 *
 *	Notes:
 *	
 *	History:
 *
 ****************************************************************/
Err		TptSendWait			(UInt refNum, Long timeout)
{
	if (gType == tptSerial)
	{
		return SerSendWait (refNum, timeout);
	}
	else if (gType == tptSockets)
	{
	  // All the data is considered sent when using sockets
	  return errNone;
	}
	else if (gType == tptPPC)
	{
		return PPCSendWait (timeout);
	}

	return errNone;
}



/***************************************************************
 *	Function:	 
 *
 *	Summary:	  
 *	  
 *	Parameters:
 *
 *	Returns:
 *	
 *	Called By: 
 *
 *	Notes:
 *	
 *	History:
 *
 ****************************************************************/
ULong	TptReceive			(UInt refNum, VoidPtr rcvBufP, ULong count, 
							  Long timeout, Err* errP)
{
	if (gType == tptSerial)
	{
		return SerReceive (refNum, rcvBufP, count, timeout, errP);
	}
	else if (gType == tptSockets)
	{
		ULong result;	

		result = SerSockReceive (rcvBufP, count, timeout, errP);
		*errP = PrvTranslateErr (*errP);
		return result;
	}
	else if (gType == tptPPC)
	{
		return PPCReceive (rcvBufP, count, timeout, errP);
	}

	return 0;
}



/***************************************************************
 *	Function:	 
 *
 *	Summary:	  
 *	  
 *	Parameters:
 *
 *	Returns:
 *	
 *	Called By: 
 *
 *	Notes:
 *	
 *	History:
 *
 ****************************************************************/
Err		TptReceiveWait		(UInt refNum, ULong bytes, Long timeout)
{
	if (gType == tptSerial)
	{
		return SerReceiveWait (refNum, bytes, timeout);
	}
	else if (gType == tptSockets)
	{
		return  PrvTranslateErr (SerSockReceiveWait (bytes, timeout));
	}
	else if (gType == tptPPC)
	{
		return PPCReceiveWait (bytes, timeout);
	}

	return errNone;
}



/***************************************************************
 *	Function:	 
 *
 *	Summary:	  
 *	  
 *	Parameters:
 *
 *	Returns:
 *	
 *	Called By: 
 *
 *	Notes:
 *	
 *	History:
 *
 ****************************************************************/
void	TptReceiveFlush		(UInt refNum, Long timeout)
{
	if (gType == tptSerial)
	{
		SerReceiveFlush (refNum, timeout);
	}
	else if (gType == tptSockets)
	{
		SerSockReceiveFlush (timeout);
	}
	else if (gType == tptPPC)
	{
		PPCReceiveFlush (timeout);
	}
}


/***************************************************************
 *	Function:	 
 *
 *	Summary:	  
 *	  
 *	Parameters:
 *
 *	Returns:
 *	
 *	Called By: 
 *
 *	Notes:
 *	
 *	History:
 *
 ****************************************************************/
Err		TptClearErr			(UInt refNum)
{
	if (gType == tptSerial)
	{
		return SerClearErr (refNum);
	}
	else if (gType == tptSockets)
	{
		return 0;
	}
	else if (gType == tptPPC)
	{
		return PPCClearErr ();
	}

	return errNone;
}


