/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#ifndef _PALMHEAP_H_
#define _PALMHEAP_H_


namespace PalmHeap
{
	enum EWalkerProcResult
	{
		kKeepIterating,
		kStopIterating
	};

	enum EPalmHeapResult
	{
			// The following are returned by WalkHeap.

		kWalkHeapFinished,
		kWalkHeapAborted,

		kChunkOK,

			// Handy "first error" enumerated value.

		kWalkHeapError,

			// The following are returned by ValidateFoo functions
			// as well as WalkHeap (implicitly, as it calls the
			// ValidateFoo functions).

		kCorruptedHeap_ChunkNotInHeap,
		kCorruptedHeap_ChunkTooLarge,
		kCorruptedHeap_InvalidFlags,
		kCorruptedHeap_HOffsetNotInMPT,
		kCorruptedHeap_HOffsetNotBackPointing,
		kCorruptedHeap_InvalidLockCount,

		kNULLPointerOrHandle
	};

	struct HeapInfo
	{
		int			version;			// 1 = original
										// 2 = >64K
										// 3 = has free list

		uaecptr		heapHdrStart;
		uae_u32		heapHdrSize;		// Up to the master pointer table struct
		uaecptr		firstChunk;			// Just after the end of the master pointer table

		Word		flags;
		DWord		size;
		DWord		firstFree;
		UInt		chunkHdrSize;
	};

	struct MPTInfo
	{
		int			version;			// 1 = original
										// 2 = >64K

		uaecptr		mptHdrStart;
		uae_u32		mptHdrSize;			// Up to the array of master pointers

		Word		numEntries;
		DWord		nextTblOffset;
	};

	struct ChunkInfo
	{
		int			version;			// 1 = original
										// 2 = >64K
		uaecptr		chunkHdrStart;
		uae_u32		chunkHdrSize;

		Bool		free;				// set if free chunk
		Bool		moved;				// used by MemHeapScramble
		Bool		unused2;			// unused
		Bool		unused3;			// unused
		char		sizeAdj;			// size adjustment
		DWord		size;				// actual size of chunk

		char		lockCount;			// lock count
		char		owner;				// owner ID
		SDWord		hOffset;			// signed handle offset/2
										// used in free chunks to point to next free chunk
	};


	// Standard sub-system functions.

	void						Initialize			(void);
	void						Reset				(void);
	void						Dispose				(void);


	// Client API -- getting heap info.

	void						GetHeapInfo			(UInt		heapID,
													 HeapInfo*	info);
	void						GetHeapInfo			(uaecptr	heapPtr,
													 HeapInfo*	info);
	void						GetMPTInfo			(uaecptr	mptStart,
													 int		heapVersion,
													 MPTInfo*	info);
	void						GetChunkInfo		(uaecptr	hdrP,
													 int		heapVersion,
													 ChunkInfo*	info);

	uaecptr						GetChunkHeader		(VoidPtr	p, const HeapInfo& heapInfo);
	uaecptr						GetChunkHeader		(VoidHand	h, const HeapInfo& heapInfo);


	// Client API -- validating heap chunks.

	PalmHeap::EPalmHeapResult	ValidateChunk		(const HeapInfo&	heapInfo,
													 const ChunkInfo&	chunkInfo,
													 Bool				validateHOffset);
	PalmHeap::EPalmHeapResult	ValidatePointer		(VoidPtr			p,
													 const HeapInfo&	heapInfo);
	PalmHeap::EPalmHeapResult	ValidateHandle		(VoidHand			h,
													 const HeapInfo&	heapInfo);

	void						ValidateHeaps		(void);


	// Client API -- walking the heap.

	typedef EWalkerProcResult	(*WalkerProc)		(const HeapInfo&	heapInfo,
													 const ChunkInfo&	chunkInfo,
													 void*				userData);

	EPalmHeapResult				WalkHeap			(UInt heapID, WalkerProc fn, void* userData);


	// Client API - other queries.

	HeapInfo					GetHeapContaining	(uaecptr);
	ChunkInfo					GetChunkContaining	(uaecptr);


	// Internal utilities.

	UInt						GetNumHeaps			(UInt		cardNo);
	uaecptr						GetHeapPtr			(UInt		heapID);
	int							GetHeapVersion		(uaecptr	heapPtr);
	VoidPtr						GetPointer			(VoidHand	h);

	struct GetChunkData
	{
		uaecptr		testAddress;
		ChunkInfo	chunkInfo;
	};

	EWalkerProcResult			GetChunkCallback	(const HeapInfo& heapInfo,
													 const ChunkInfo& chunkInfo,
													 void* userData);
	EWalkerProcResult			ValidateChunkCallback(const PalmHeap::HeapInfo& heapInfo,
													 const PalmHeap::ChunkInfo& chunkInfo,
													 void* userData);
}

#endif /* _PALMHEAP_H_ */

