/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#ifndef _PLATFORM_FILES_H_
#define _PLATFORM_FILES_H_

#include "ErrorHandling.h"		// ErrCode


class FileReference
{
	public:
								FileReference	(void);
								FileReference	(const FileReference&);
								FileReference	(const char*);
								FileReference	(const string&);
								FileReference	(const unsigned char*);
								~FileReference	(void);

		FileReference&			operator=		(const FileReference&);

		Bool					IsSpecified		(void) const;
		Bool					Exists			(void) const;

		string					GetFileName		(void) const;

		bool					operator==		(const FileReference&) const;
		bool					operator!=		(const FileReference&) const;
		bool					operator<		(const FileReference&) const;
		bool					operator>		(const FileReference&) const;

#if defined (__MACOS__)
								FileReference	(const FSSpec&);
								FileReference	(AliasHandle);

		AliasHandle				GetAlias		(void) const;
		FSSpec					GetFSSpec		(void) const;

	private:
		void					UpdateAlias		(void);
		void					UpdateSpec		(void);

		Bool					EqualAlias		(AliasHandle, AliasHandle) const;
		Bool					EqualFSSpec		(const FSSpec&, const FSSpec&) const;

		AliasHandle				fFileAlias;
		FSSpec					fFileSpec;
#endif


#if defined (_WINDOWS) || defined (UNIX)
		const char*				GetFilePath		(void);

	private:
		string					fFilePath;
#endif
};


enum
{
	kCreateNew,			// Creates a new file; fails if it already exists.
	kCreateAlways,		// Always creates a new file, destroying any existing file.
	kOpenExisting,		// Opens the file it it exists, fails otherwise.
	kOpenAlways,		// Opens the file if it exists, creates it otherwise.
	kTruncateExisting,	// Opens and clears existing file, fails otherwise.

	kOpenTypeMask = 0x0F,

	kOpenRead	= 16,
	kOpenWrite	= 32,
	kOpenReadWrite = kOpenRead | kOpenWrite
};

enum SeekMode
{
	kSeekSet,
	kSeekCur,
	kSeekEnd
};

class FileHandle
{
	public:
							FileHandle		(FileReference&,
											 long openMode,
											 uae_u32 creator = 0,
											 uae_u32 fileType = 0);
							~FileHandle		(void);

		FileReference		GetReference	(void);
		Bool				IsOpen			(void);

		void				SetPos			(long offset, SeekMode);
		uae_u32				GetPos			(void);

		void				SetLength		(uae_u32);
		uae_u32				GetLength		(void);

		void				Read			(uae_u32 length, void* buffer);
		void				Write			(uae_u32 length, const void* buffer);

	protected:
		void				Open			(FileReference&,
											 long openMode,
											 uae_u32 creator,
											 uae_u32 fileType);

#if defined (__MACOS__)
		void				CreateNew		(const FSSpec&,
											 long openMode,
											 uae_u32 creator,
											 uae_u32 fileType);
		void				CreateAlways	(const FSSpec&,
											 long openMode,
											 uae_u32 creator,
											 uae_u32 fileType);
		void				OpenExisting	(const FSSpec&,
											 long openMode,
											 uae_u32 creator,
											 uae_u32 fileType);
		void				OpenAlways		(const FSSpec&,
											 long openMode,
											 uae_u32 creator,
											 uae_u32 fileType);
		void				TruncateExisting(const FSSpec&,
											 long openMode,
											 uae_u32 creator,
											 uae_u32 fileType);
		char				GetPermission	(long openMode);
#endif

		void				Close			(void);

		void				Throw			(ErrCode);
		void				SetFileNameParameter	(void);

	private:
			// Protect the copy constructor so that we don't
			// accidentally make copies (what does it mean to
			// copy a reference to an open file?).

							FileHandle		(const FileHandle&);

		FileReference		fFileRef;

#if defined (__MACOS__)
		short				fRefNum;
#endif

#if defined (_WINDOWS)
		void*				fHandle;	// == HANDLE
#endif

#if defined (UNIX)
		FILE *				fStream;
#endif
};

#endif /* _PLATFORM_FILES_H_ */

