/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#ifndef _PROFILING_H_
#define _PROFILING_H_


// wait states vary with hardware

#define PILOT		1
#define	PALM_PILOT	2
#define PALM_III	3
#define PALM_V		5
#define PALM_VII	7


#define PALM_HARDWARE_TYPE PALM_III



#if PALM_HARDWARE_TYPE == PILOT

#define WAITSTATES_ROM			2
#define WAITSTATES_SRAM			4	
#define WAITSTATES_DRAM			4	
#define WAITSTATES_REGISTERS	0
#define WAITSTATES_DUMMYBANK	99	// hopefully won't be used!
#define PROCESSOR_SPEED 16


#elif PALM_HARDWARE_TYPE == PALM_PILOT

#define WAITSTATES_ROM			2
#define WAITSTATES_SRAM			4	
#define WAITSTATES_DRAM			4	
#define WAITSTATES_REGISTERS	0
#define WAITSTATES_DUMMYBANK	99	// hopefully won't be used!
#define PROCESSOR_SPEED 16


#elif PALM_HARDWARE_TYPE == PALM_III

#define WAITSTATES_ROM			2
#define WAITSTATES_SRAM			4	// according to Wayne Harrington
#define WAITSTATES_DRAM			4	// according to Wayne Harrington
#define WAITSTATES_REGISTERS	0
#define WAITSTATES_DUMMYBANK	99	// hopefully won't be used!
#define PROCESSOR_SPEED 16

#elif PALM_HARDWARE_TYPE == PALM_V

#define WAITSTATES_ROM			1
#define WAITSTATES_SRAM			1	
#define WAITSTATES_DRAM			1	
#define WAITSTATES_REGISTERS	0
#define WAITSTATES_DUMMYBANK	99	// hopefully won't be used!
#define PROCESSOR_SPEED 16


#elif PALM_HARDWARE_TYPE == PALM_VII

#define WAITSTATES_ROM			1
#define WAITSTATES_SRAM			1	
#define WAITSTATES_DRAM			1	
#define WAITSTATES_REGISTERS	0
#define WAITSTATES_DUMMYBANK	99	// hopefully won't be used!
#define PROCESSOR_SPEED 16

#endif	// PALM_HARDWARE_TYPE






// 16MHz, more or less
#define kCyclesPerSecond 16580608
#define kCyclesPerMilliSecond 16580




	// Declare these before including UAE.h, as UAE.h refers
	// to them (well, gProfilingEnabled at least).

#ifdef __cplusplus
extern "C" {
#endif

// Public (HostControl) functions
extern void ProfileInit(int maxCalls, int maxDepth);
extern void ProfileStart();
extern void ProfileStop();

// note, ProfileDump destructively modifies the profile data,
// you must call ProfileCleanup after dumping, do not attempt
// to gather more profiling data.
extern void ProfileDump(const char* dumpFileName);

extern void ProfilePrint(const char* textFileName);
extern void ProfileCleanup();



// Private (Emulator Only) stuff

	// Indicates that the profiling system has been initialized
	// by calling ProfileInit.  Set back to false after ProfileCleanup.
	// If true, adds calls to ProfileIncrementClock throught the 68K
	// emuilator.  Also controls calls to ProfileFnEnter, ProfileFnExit,
	// ProfileInterruptEnter and ProfileInterruptExit, (but that seems
	// redundant with gProfilingOn).  Finally, enables cycle counting for
	// memory accesses (but only if gProfilingCounted is also true).
extern int gProfilingEnabled;

	// Set to false in ProfileInit and ProfileStop, set to true in ProfileStart.
	// If false, ProfileFnEnter, ProfileFnExit, ProfileInterruptEnter and
	// ProfileInterruptExit merely return.
extern int gProfilingOn;

	// Set to false in ProfileInit, set to true at the start of the loop in
	// Emulator::Execute and false at the end of the loop.
	// It controls whether or not memory access cycles are counted.  It
	// is always checked in conjunction with gProfilingEnabled; both have
	// to be true in order for memory access cycles to be counted.
extern int gProfilingCounted;


#ifdef __cplusplus

class StDisableAllProfiling
{
	public:
		StDisableAllProfiling () :
		  fOld (gProfilingEnabled)
		{
		  gProfilingEnabled = false;
		}

		~StDisableAllProfiling ()
		{
			gProfilingEnabled = fOld;
		}

	private:
		int	fOld;
};

class StDisableFunctionTracking
{
	public:
		StDisableFunctionTracking () :
		  fOld (gProfilingOn)
		{
		  gProfilingOn = false;
		}

		~StDisableFunctionTracking ()
		{
			gProfilingOn = fOld;
		}

	private:
		int	fOld;
};

class StDisableMemoryCycleCounting
{
	public:
		StDisableMemoryCycleCounting () :
		  fOld (gProfilingCounted)
		{
		  gProfilingCounted = false;
		}

		~StDisableMemoryCycleCounting ()
		{
			gProfilingCounted = fOld;
		}

	private:
		int	fOld;
};

#endif

#if defined(__MWERKS__)
#include <Math64.h>
#endif

#if defined(_MSC_VER)
typedef __int64 SInt64;
typedef __int64 SInt64;
typedef unsigned __int64 UInt64;
#endif

#if defined(__GNUC__)
#include <sys/types.h>
typedef int64_t SInt64;
typedef int64_t SInt64;
typedef u_int64_t UInt64;
#endif

#if defined(_MSC_VER) || defined(__GNUC__)
#define S64Max() 9223372036854775807
#define S64Min() (-S64Max() - 1)
#define S64Add(x, y) ((x) + (y))
#define S64Subtract(x, y) ((x) - (y))
#define S64Negate(x) (-(x))
#define S64Absolute(x) absll((x))
#define S64Multiply(x, y) ((x) * (y))
#define S64Div(x, y) ((x) / (y))
#define S64Mod(x, y) ((x) % (y))
#define S64Set(x) ((SInt64) (x))
#define S64SetU(x) ((SInt64) (x))
#define S32Set(x) ((SInt32) (x))
#define S64Compare(x, y) ((int)((x) - (y)))
#define S64And(x, y) ((Boolean)((x) && (y)))
#define S64Or(x, y) ((Boolean)((x) || (y)))
#define S64Eor(x, y) ((Boolean)((x) ^ (y)))
#define S64Not(x) ((Boolean)(!(x)))
#define S64BitwiseAnd(x, y) ((x) & (y))
#define S64BitwiseOr(x, y) ((x) | (y))
#define S64BitwiseEor(x, y) (((x) & (~(y))) | ((~(x)) & (y)))
#define S64BitwiseNot(x) (~(x))
#define S64ShiftRight(x, y) ((x) >> (y))
#define S64ShiftLeft(x, y) ((x) << (y))
#define SInt64ToLongDouble(x) ((long double)(x))
#define LongDoubleToSInt64(x) ((SInt64)(x))
#endif

#define DEFAULT_RESULTS_FILENAME	"Profile Results.mwp"
#define DEFAULT_RESULTS_TEXT_FILENAME	"Profile Results.txt"

#define HAS_PROFILING_DEBUG		0	// turn on code to help debug profiler


//  normally passed in to ProfileInit, but here for CEmulatorApp.cpp
#define MAXFNCALLS	20000
#define MAXUNIQUEFNS 2500
#define MAXDEPTH 	200



	// Total running clock cycles.  Includes opcode execution
	// time and memory fetch time.
extern SInt64 gClockCycles;

	// Portion of gClockCycles spent reading memory.
extern SInt64 gReadCycles;

	// Portion of gClockCycles spent writing memory.
extern SInt64 gWriteCycles;

static __inline__ void ProfileIncrementClock(unsigned long by)
{
	gClockCycles = S64Add(gClockCycles, S64SetU(by));
}

static __inline__ void ProfileIncrementRead(int reads, int waitstates)
{
	gClockCycles = S64Add(gClockCycles, S64Set(reads * (4+waitstates)));
	gReadCycles = S64Add(gReadCycles, S64Set(reads));
}

static __inline__ void ProfileIncrementWrite(int writes, int waitstates)
{
	gClockCycles = S64Add(gClockCycles, S64Set(writes * (4+waitstates)));
	gWriteCycles = S64Add(gWriteCycles, S64Set(writes));
}


extern void ProfileFnEnter(uaecptr destAddress, uaecptr returnAddress);
extern void ProfileFnExit(uaecptr returnAddress, uaecptr oldAddress);

extern void ProfileInterruptEnter(uae_s32 iException, uaecptr returnAddress);
extern void ProfileInterruptExit(uaecptr returnAddress);



// debugging stuff

extern long gReadMismatch;
extern long gWriteMismatch;

extern void ProfileTest();


#ifdef __cplusplus
}
#endif


#endif /* _PROFILING_H_ */
  
  
