/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "TrapPatches.h"

#include "AppPreferences.h"		// gPrefs
#include "Byteswapping.h"		// Canonical
#include "CGremlinsStubs.h"		// StubAppEnqueueKey
#include "CPU_REG.h"			// LowMem::GetEvtMgrIdle, LowMem::TrapExists, gKeyQueue, DateToDays
#include "DebugMgr.h"			// Debug::SendMessage
#include "ErrorHandling.h"		// Errors::SysFatalAlert
#include "Gremlins.h"			// Gremlins_IsOn, Gremlins_PostFakeEvent, Gremlins_CanSwitchToApp
#include "HostControlPrv.h"		// HandleHostControlCall
#include "Logging.h"			// LogEvtAddEventToQueue, etc.
#include "MetaMemory.h"			// MetaMemory mark functions
#include "Miscellaneous.h"		// SetHotSyncUserName, DateToDays
#include "Platform.h"			// Platform::GetDate
#include "Profiling.h"			// StDisableAllProfiling
#include "RAM_ROM.h"			// CEnableFullAccess
#include "Strings.r.h"			// kStr_ values
#include "UAE_Utils.h"			// uae_memmove, uae_memset

#if defined (_WINDOWS)
#include "SocketMessaging.h"
#endif




// ======================================================================
// Patches for system functions.
// ======================================================================

namespace SysHeadpatch
{
	CallROMType		RecordTrapNumber		(void);	// EvtGetEvent & EvtGetPen

	CallROMType		DbgMessage				(void);
	CallROMType		DmInit					(void);
	CallROMType		ErrDisplayFileLineMsg	(void);
	CallROMType		EvtAddEventToQueue		(void);
	CallROMType		EvtAddUniqueEventToQueue(void);
	CallROMType		EvtEnqueueKey			(void);
	CallROMType		EvtEnqueuePenPoint		(void);
	CallROMType		HostControl				(void);
	CallROMType		HwrBatteryLevel			(void);
	CallROMType		HwrGetROMToken			(void);
	CallROMType		HwrSleep				(void);
	CallROMType		KeyCurrentState			(void);
	CallROMType		MemMove					(void);
	CallROMType		MemSet					(void);
	CallROMType		PenOpen					(void);
	CallROMType		RctPtInRectangle		(void);
	CallROMType		SysAppExit				(void);
	CallROMType		SysBinarySearch			(void);
	CallROMType		SysEvGroupWait			(void);
	CallROMType		SysFatalAlert			(void);
	CallROMType		SysLaunchConsole		(void);
	CallROMType		SysSemaphoreWait		(void);
	CallROMType		SysUIAppSwitch			(void);
	CallROMType		SysUIBusy				(void);
	CallROMType		TimInit					(void);
}


namespace SysTailpatch
{
	void		EvtGetEvent				(void);
	void		EvtGetPen				(void);
	void		EvtGetSysEvent			(void);
	void		FtrInit					(void);
	void		HwrMemReadable			(void);
	void		HwrSleep				(void);
	void		SysAppStartup			(void);
	void		SysBinarySearch			(void);
	void		SysLibFind				(void);
	void		TimInit					(void);
	void		UIInitialize			(void);
}


// ======================================================================
//	Proto patch table for the system functions.  This array will be used
//	to create a sparse array at runtime.
// ======================================================================

ProtoPatchTableEntry	gProtoSysPatchTable[] =
{
	sysTrapDbgMessage,				SysHeadpatch::DbgMessage,				NULL,
	sysTrapDmInit,					SysHeadpatch::DmInit,					NULL,
	sysTrapErrDisplayFileLineMsg,	SysHeadpatch::ErrDisplayFileLineMsg,	NULL,
	sysTrapEvtAddEventToQueue,		SysHeadpatch::EvtAddEventToQueue,		NULL,
	sysTrapEvtAddUniqueEventToQueue,SysHeadpatch::EvtAddUniqueEventToQueue,	NULL,
	sysTrapEvtEnqueueKey,			SysHeadpatch::EvtEnqueueKey,			NULL,
	sysTrapEvtEnqueuePenPoint,		SysHeadpatch::EvtEnqueuePenPoint,		NULL,
	sysTrapEvtGetEvent,				SysHeadpatch::RecordTrapNumber,			SysTailpatch::EvtGetEvent,
	sysTrapEvtGetPen,				SysHeadpatch::RecordTrapNumber,			SysTailpatch::EvtGetPen,
	sysTrapEvtGetSysEvent,			NULL,									SysTailpatch::EvtGetSysEvent,
	sysTrapFtrInit,					NULL,									SysTailpatch::FtrInit,
	sysTrapHostControl,				SysHeadpatch::HostControl,				NULL,
	sysTrapHwrBatteryLevel,			SysHeadpatch::HwrBatteryLevel,			NULL,
	sysTrapHwrGetROMToken,			SysHeadpatch::HwrGetROMToken,			NULL,
	sysTrapHwrMemReadable,			NULL,									SysTailpatch::HwrMemReadable,
	sysTrapHwrSleep,				SysHeadpatch::HwrSleep,					SysTailpatch::HwrSleep,
	sysTrapKeyCurrentState,			SysHeadpatch::KeyCurrentState,			NULL,
	sysTrapMemMove,					SysHeadpatch::MemMove,					NULL,
	sysTrapMemSet,					SysHeadpatch::MemSet,					NULL,
	sysTrapPenOpen,					SysHeadpatch::PenOpen,					NULL,
	sysTrapRctPtInRectangle,		SysHeadpatch::RctPtInRectangle,			NULL,
	sysTrapSysAppExit,				SysHeadpatch::SysAppExit,				NULL,
	sysTrapSysAppStartup,			NULL,									SysTailpatch::SysAppStartup,
	sysTrapSysBinarySearch,			SysHeadpatch::SysBinarySearch,			SysTailpatch::SysBinarySearch,
	sysTrapSysEvGroupWait,			SysHeadpatch::SysEvGroupWait,			NULL,
	sysTrapSysFatalAlert,			SysHeadpatch::SysFatalAlert,			NULL,
	sysTrapSysLaunchConsole,		SysHeadpatch::SysLaunchConsole,			NULL,
	sysTrapSysLibFind,				NULL,									SysTailpatch::SysLibFind,
	sysTrapSysSemaphoreWait,		SysHeadpatch::SysSemaphoreWait,			NULL,
	sysTrapSysUIAppSwitch,			SysHeadpatch::SysUIAppSwitch,			NULL,
	sysTrapSysUIBusy,				SysHeadpatch::SysUIBusy,				NULL,
	sysTrapTimInit,					SysHeadpatch::TimInit,					SysTailpatch::TimInit,
	sysTrapUIInitialize,			NULL,									SysTailpatch::UIInitialize,

	0,								NULL,									NULL
};


// ======================================================================
//	Patches for HtalLib functions
// ======================================================================

namespace HtalLibHeadpatch
{
	CallROMType		HtalLibSendReply			(void);
}


#pragma mark -

// ===========================================================================
//		 ModulePatchTable
// ===========================================================================
//	A simple class for managing patches on a module.  A "module" is defined
//	a the main set of system functions (dispatch numbers 0xA000 - 0xA7FFF) or
//	a library (dispatch numbers 0xA800 - 0xAFFF, with a unique refnum to
//	select which library).

class ModulePatchTable
{
	public:
		void Clear ();

		void AddProtoPatchTable (ProtoPatchTableEntry protoPatchTable[]);

		// B.S. operators for VC++ so that we can put objects of
		// this class into STL collections.

		bool operator < (const ModulePatchTable& other) const
		{
			return this < &other;
		}

		bool operator == (const ModulePatchTable& other) const
		{
			return this == &other;
		}

		// Return the patch function for the given module function.  The given
		// module function *must* be given as a zero-based index.  If there is
		// no patch function for the modeule function, return NULL.

		HeadpatchProc GetHeadpatch (uae_u16 index)
		{
			if (index < fHeadpatches.size ())
			{
				return fHeadpatches[index];
			}

			return NULL;
		}

		TailpatchProc GetTailpatch (uae_u16 index)
		{
			if (index < fTailpatches.size ())
			{
				return fTailpatches[index];
			}

			return NULL;
		}

	private:
		vector<HeadpatchProc>	fHeadpatches;
		vector<TailpatchProc>	fTailpatches;
};


void ModulePatchTable::Clear (void)
{
	fHeadpatches.clear ();
	fTailpatches.clear ();
}


void ModulePatchTable::AddProtoPatchTable (ProtoPatchTableEntry protoPatchTable[])
{
	// Create a fast dispatch table for the managed module.  A "fast
	// dispatch table" is a table with a headpatch and tailpatch entry
	// for each possible function in the module.  If the function is
	// not head or tailpatched, the corresponding entry is NULL.  When
	// a patch function is needed, the trap dispatch number is used as
	// an index into the table in order to get the right patch function.
	//
	// For simplicity, "fast patch tables" are created from "proto patch
	// tables".  A proto patch table is a compact table containing the
	// information needed to create a fast patch table.  Each entry in
	// the proto patch table is a trap-number/headpatch/tailpatch tupple.
	// Each tuple is examined in turn.  If there is a head or tail patch
	// function for the indicated module function, that patch function
	// is entered in the fast dispatch table, using the trap number as
	// the index.

	for (long ii = 0; protoPatchTable[ii].fTrapWord; ++ii)
	{
		// If there is a headpatch function...

		if (protoPatchTable[ii].fHeadpatch)
		{
			// Get the trap number.

			uae_u16	index = SysTrapIndex (protoPatchTable[ii].fTrapWord);

			// If the trap number is 0xA800-based, make it zero based.

			if (IsLibraryTrap (index))
				index -= SysTrapIndex (sysLibTrapBase);

			// Resize the fast patch table, if necessary.

			if (index >= fHeadpatches.size ())
			{
				fHeadpatches.resize (index + 1);
			}

			// Add the headpatch function.

			fHeadpatches[index] = protoPatchTable[ii].fHeadpatch;
		}

		// If there is a tailpatch function...

		if (protoPatchTable[ii].fTailpatch)
		{
			// Get the trap number.

			uae_u16	index = SysTrapIndex (protoPatchTable[ii].fTrapWord);

			// If the trap number is 0xA800-based, make it zero based.

			if (IsLibraryTrap (index))
				index -= SysTrapIndex (sysLibTrapBase);

			// Resize the fast patch table, if necessary.

			if (index >= fTailpatches.size ())
			{
				fTailpatches.resize (index + 1);
			}

			// Add the tailpatch function.

			fTailpatches[index] = protoPatchTable[ii].fTailpatch;
		}
	}
}


// ===========================================================================
//		 TailpatchType
// ===========================================================================
// Structure used to hold tail-patch information.

struct TailpatchType
{
	uaecptr			fPC;
	uae_u16			fOldOpcode;
	uae_s32			fCount;
	TailpatchProc	fTailpatch;

	// I hate VC++...really I do...
	bool operator< (const TailpatchType&) const {return false;}
	bool operator> (const TailpatchType&) const {return false;}
	bool operator== (const TailpatchType&) const {return false;}
	bool operator!= (const TailpatchType&) const {return false;}
};


// ===========================================================================
//		 Patches
// ===========================================================================

// ======================================================================
//	Globals and constants
// ======================================================================
const uae_u16		kOpcode_Tailpatch	= m68kTrapInstr + kATrapTailpatchTrapNum;

const UInt			kMagicRefNum		= 0x666;	// See comments in HtalLibSendReply.

EmuCurAppInfo		Patches::fgCurAppInfo;
Bool				Patches::fgUIInitialized;
long				Patches::fgSysBinarySearchCount;

UInt				Patches::fgNextAppCardNo;
LocalID				Patches::fgNextAppDbID;
Bool				Patches::fgQuitOnAppExit;
QuitStage			Patches::fgQuitStage;


ModulePatchTable	gSysPatchTable;
ModulePatchTable	gNetLibPatchTable;

typedef vector<ModulePatchTable*>	PatchTableType;
PatchTableType		gLibPatches;

typedef vector<TailpatchType>		TailpatchTableType;
TailpatchTableType	gInstalledTailpatches;

uae_u16				gLastEvtTrap;
AbsRectType			gScreen_UpdateRect;
DWord				gOSVersion;

const DWord			kOSUndeterminedVersion = ~0;


#if defined (__MACOS__)
extern Bool			gOKToAllocateMemory;
#endif


// ======================================================================
//	Private functions
// ======================================================================

static string		PrvToString (uaecptr s);
static void			PrvAutoload (void);


// ========================================================================
// Macros for extracting parameters from the emulated stack
// Note that use of these has been superceded by the PARAM_FOO
// macros;  we should eventually move over to those macros.
// ========================================================================

#define PARAMETER_SIZE(x)	\
	(sizeof (((StackFrame*) 0)->x))

#define PARAMETER_OFFSET(x)	\
	(m68k_areg (regs, 7) + offsetof (StackFrame, x))

#define GET_PARAMETER(x)		\
	((PARAMETER_SIZE(x) == sizeof (char)) ? get_byte (PARAMETER_OFFSET(x)) :	\
	 (PARAMETER_SIZE(x) == sizeof (short)) ? get_word (PARAMETER_OFFSET(x)) :	\
											get_long (PARAMETER_OFFSET(x)))

#define SET_PARAMETER(x, v)		\
	((PARAMETER_SIZE(x) == sizeof (char)) ? put_byte (PARAMETER_OFFSET(x), v) :	\
	 (PARAMETER_SIZE(x) == sizeof (short)) ? put_word (PARAMETER_OFFSET(x), v) :	\
											put_long (PARAMETER_OFFSET(x), v))


// ========================================================================
// The following functions define a bunch of StackFrame structs.
// These structs need to mirror the format of parameters pushed
// onto the stack by the emulated code, and so need to be packed
// to 2-byte boundaries.
//
// The pragmas are reversed at the end of the file.
// ========================================================================

#include "PalmPack.h"


/***********************************************************************
 *
 * FUNCTION:	Patches::Initialize
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/
void Patches::Initialize (void)
{
	gSysPatchTable.Clear ();
	gSysPatchTable.AddProtoPatchTable (gProtoSysPatchTable);
	gSysPatchTable.AddProtoPatchTable (gProtoMemMgrPatchTable);

	gNetLibPatchTable.Clear ();
	gNetLibPatchTable.AddProtoPatchTable (gProtoNetLibPatchTable);

	// Oh, man...this one's very important.  Make sure we do this here
	// BEFORE calling Reset, or Reset will try to remove any installed
	// tailpatches.  If Patches::Initialize is being called for a second
	// time because we're re-initializing the CPU, then Patches::Reset
	// will try to restore opcodes from the PREVIOUS ROM image.  To
	// prevent that from happening, zap the table NOW!

	gInstalledTailpatches.clear ();
}


/***********************************************************************
 *
 * FUNCTION:	Patches::Reset
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Patches::Reset (void)
{
	gLastEvtTrap = 0;
	gOSVersion = kOSUndeterminedVersion;

	Patches::RemoveAllTailpatches ();

	gInstalledTailpatches.clear ();

	fgUIInitialized				= false;
	fgSysBinarySearchCount		= 0;
	fgMemMgrCount				= 0;
	fgMemSemaphoreCount			= 0;
	fgMemSemaphoreReserveTime	= 0;

	Patches::SetSwitchApp (0, 0);
	Patches::QuitOnAppExit (false);

	gLibPatches.clear ();

	// Clear out everything we know about the current application.

	memset (&fgCurAppInfo, 0, sizeof (fgCurAppInfo));
}


/***********************************************************************
 *
 * FUNCTION:	Patches::InstallLibPatch
 *
 * DESCRIPTION:	Called from a patch to SysLibFind to see if we want to
 *				patch the requested library.  Determine if we want to
 *				patch the library by looking at the library name (a
 *				case-sensitive test should be OK).  If we want to patch
 *				the library, return a pointer to the object that
 *				contains the patches.  This object address will be
 *				entered into a table associated with the library's
 *				reference number.  When something later tries to call a
 *				library function with that reference number, the patch
 *				object will be queried as to whether or not the specific
 *				function should be patched.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Patches::InstallLibPatch (const string& libName, UInt /*refNum*/, ModulePatchTable*& entry)
{
	if (libName == "Net.lib")
	{
		entry = &gNetLibPatchTable;
	}
}


/***********************************************************************
 *
 * FUNCTION:	Patches::HandleSystemCall
 *
 * DESCRIPTION:	If this is a trap we could possibly have head- or
 *				tail-patched, handle those cases.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType Patches::HandleSystemCall (uae_u16 trapWord, uaecptr startPC)
{
	HeadpatchProc	hp;
	TailpatchProc	tp;
	Patches::GetPatches (hp, tp, trapWord);

	CallROMType	handled = Patches::HandlePatches (hp, tp, startPC, trapWord);

	return handled;
}


/***********************************************************************
 *
 * FUNCTION:	Patches::GetPatches
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Patches::GetPatches (HeadpatchProc& hp, TailpatchProc& tp, uae_u16 trapWord)
{
	ModulePatchTable* patchTable = NULL;

	// Turn the trap word (the $Axxx value) into a zero-based index.

	uae_u16	trapIndex = SysTrapIndex (trapWord);

	// If this is in the system function range, check our table of
	// system function patches.

	if (IsSystemTrap (trapWord))
	{
		patchTable = &gSysPatchTable;
	}

	// Otherwise, see if this is a call to a library function

	else
	{
		trapIndex = LibTrapIndex (trapWord);

		// See if we've installed a patch table for this library.

		UInt	refNum = (UInt) get_word (m68k_areg (regs, 7));

		if (refNum < gLibPatches.size ())
		{
			patchTable = gLibPatches[refNum];
		}
		else if (refNum == kMagicRefNum)	// See comments in HtalLibSendReply.
		{
			hp = HtalLibHeadpatch::HtalLibSendReply;
			tp = NULL;
			return;
		}
	}

	// Now that we've got the right patch table for this module, see if
	// that patch table has head- or tailpatches for this function.

	if (patchTable)
	{
		hp = patchTable->GetHeadpatch (trapIndex);
		tp = patchTable->GetTailpatch (trapIndex);
	}
	else
	{
		hp = NULL;
		tp = NULL;
	}
}


/***********************************************************************
 *
 * FUNCTION:	Patches::HandlePatches
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType Patches::HandlePatches (HeadpatchProc hp, TailpatchProc tp,
									uaecptr startPC, uae_u16 trapWord)
{
	CallROMType	handled = kExecuteROM;

	// First, see if we have a SysHeadpatch for this function. If so, call
	// it. If it returns true, then that means that the head patch
	// completely handled the function.

	// !!! May have to mess with PC here in case patches do something
	// to enter the debugger.

	if (hp)
	{
		handled = CallHeadpatch (hp);
	}

	// Next, see if there's a SysTailpatch function for this trap. If
	// so, install the TRAP that will cause us to regain control
	// after the trap function has executed.

	if (tp)
	{
		if (handled == kExecuteROM)
		{
			SetupForTailpatch (tp, startPC);
		}
		else
		{
			CallTailpatch (tp);
		}
	}

	return handled;
}


/***********************************************************************
 *
 * FUNCTION:	Patches::HandleTrap13
 *
 * DESCRIPTION:	Handle a tail patch.  Traps that we want to tail patch
 *				will have a TRAP $D stored after them.  When we hit
 *				that TRAP $D, we end up here so that we can replace the
 *				opcode, back up the PC, and then call the tailpatch
 *				function.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Patches::HandleTrap13 (uaecptr startPC)
{
	// Restore the PC with what it should be (the saved return address)

	TailpatchProc	tp = RecoverFromTailpatch (startPC);

	// We want to call the tailpatch handler for the trap that just returned.
	// "startPC" should now be pointing to the TRAP $D, which is just after
	// the TRAP $F / $Axxx codes that invoked the trap, so use it to look up
	// the $Axxx word.

	CallTailpatch (tp);

	// !!! Need to adjust the profiling counters.  They will have counted
	// the time it takes to execute a TRAP $D, and will have recorded an
	// interrupt function entry.
}


/***********************************************************************
 *
 * FUNCTION:	Patches::SetupForTailpatch
 *
 * DESCRIPTION:	Set up the pending TRAP $F call to be tailpatched.  A
 *				system function is tailpatched by storing a TRAP $D
 *				after the TRAP $F / $Axxx sequence. If there is already
 *				a TRAP $D there, we increment a use-count.  When the
 *				TRAP $D is encountered, we decrement the use-count and
 *				restore the old opcode when it reaches zero.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Patches::SetupForTailpatch (TailpatchProc tp, uaecptr startPC)
{
	// Determine the memory location to store the TRAP $D.  If we got here with
	// a TRAP $F, then we need to bump the PC by 4 in order to skip past it and
	// the $Axxx selector.  Otherwise, we probably got here via a JSR(A1), in
	// which case we need to bump the PC by 2.

	uaecptr	patchPC = startPC;
	uae_u8*	realMem = get_real_address (startPC);

	if (do_get_mem_word (realMem) == (m68kTrapInstr + sysDispatchTrapNum))
		patchPC += 4;
	else
		patchPC += 2;

	realMem = get_real_address (patchPC);
	uae_u16	oldOpcode = do_get_mem_word (realMem);

	// See if this function is already tailpatched.  If so, merely increment
	// the use-count field.

	if (oldOpcode == kOpcode_Tailpatch)
	{
		// It is.  Find the corresponding entry.

		TailpatchTableType::iterator	iter = gInstalledTailpatches.begin ();

		while (iter != gInstalledTailpatches.end ())
		{
			if ((*iter).fPC == patchPC)
			{
				++((*iter).fCount);

//				PRINTF ("Incremented tailpatch for %s at PC = 0x%08X.\n",
//							::GetTrapName (get_word (pc - 2)),
//							fgInstalledTailpatches[ii].fPC);
				return;
			}

			++iter;
		}

		assert (!"Not enough tail patches");
		abort ();
	}

	// This function is not already tailpatched, so add a new entry
	// for the the PC/opcode we want to save.

	TailpatchType	newTailpatch;

	newTailpatch.fPC = patchPC;
	newTailpatch.fOldOpcode = oldOpcode;
	newTailpatch.fCount = 1;
	newTailpatch.fTailpatch = tp;

	gInstalledTailpatches.push_back (newTailpatch);

#if defined (__MACOS__)
	// For the Mac, make sure there's always at least 10 spaces available.
	// We use these spaces in times when we can't resize the array, as
	// when we're handling debugger packets.

	if (gOKToAllocateMemory)
	{
		if (gInstalledTailpatches.size () + 10 > gInstalledTailpatches.capacity ())
		{
			gInstalledTailpatches.reserve (gInstalledTailpatches.capacity () + 10);
		}
	}
#endif

//	CEnableFullAccess	munge;	// Remove blocks on memory access.

	do_put_mem_word (realMem, kOpcode_Tailpatch);

//	PRINTF ("Installed tailpatch for %s at PC = 0x%08X.\n",
//				::GetTrapName (get_word (pc - 2)),
//				fgInstalledTailpatches[ii].fPC);

}


/***********************************************************************
 *
 * FUNCTION:	Patches::RecoverFromTailpatch
 *
 * DESCRIPTION:	A TRAP $D has just been executed.  Update our tailpatch
 *				table by either restoring the replace opcode, or by
 *				decrementing the use-count field and providing the
 *				opcode to the CPU opcode-fetcher as the next opcode to
 *				execute.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

TailpatchProc Patches::RecoverFromTailpatch (uaecptr startPC)
{
	// Get the current PC so that we can find the record for this tailpatch.

	uaecptr	patchPC = startPC;
	uae_u8*	realMem = get_real_address (patchPC);

	// Find the PC.

	TailpatchTableType::iterator	iter = gInstalledTailpatches.begin ();

	while (iter != gInstalledTailpatches.end ())
	{
		if ((*iter).fPC == patchPC)
		{
			TailpatchProc	result = (*iter).fTailpatch;

			// Decrement the use-count.  If it reaches zero, remove the
			// TRAP $D.  Otherwise, leave the TRAP $D there, but give the
			// real opcode the the CPU opcode-fetcher.

			if (--((*iter).fCount) == 0)
			{
				CEnableFullAccess	munge;	// Remove blocks on memory access.

				do_put_mem_word (realMem, (*iter).fOldOpcode);

				gInstalledTailpatches.erase (iter);

//				PRINTF ("Restored tailpatch %s at PC = 0x%08X.\n",
//						::GetTrapName (get_word (pc - 2)), pc);
			}
			else
			{
				Emulator::SetNextOpcode ((*iter).fOldOpcode);

//				PRINTF ("Decremented tailpatch %s at PC = 0x%08X.\n",
//						::GetTrapName (get_word (pc - 2)), pc);
			}

			return result;
		}

		++iter;
	}

//	PRINTF ("Could not find tailpatch entry for PC = 0x%08X.\n", pc);

	assert (!"Not enough tail patches");
	abort ();

	return NULL;
}


/***********************************************************************
 *
 * FUNCTION:	Patches::RemoveAllTailpatches
 *
 * DESCRIPTION:	Grovel over our "tail patch" array and remove all
 *				tailpatches by replacing the TRAP we've installed with
 *				the opcode that used to be there.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Patches::RemoveAllTailpatches (void)
{
//	CEnableFullAccess	munge;	// Remove blocks on memory access.

	TailpatchTableType::iterator	iter = gInstalledTailpatches.begin ();

	while (iter != gInstalledTailpatches.end ())
	{
		uae_u8*	realMem = get_real_address ((*iter).fPC);
		do_put_mem_word (realMem, (*iter).fOldOpcode);
		++iter;
	}
}


/***********************************************************************
 *
 * FUNCTION:	Patches::InstallAllTailpatches
 *
 * DESCRIPTION:	Grovel over our "tail patch" array and install all
 *				tailpatches by storing a special TRAP instruction.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Patches::InstallAllTailpatches (void)
{
//	CEnableFullAccess	munge;	// Remove blocks on memory access.

	TailpatchTableType::iterator	iter = gInstalledTailpatches.begin ();

	while (iter != gInstalledTailpatches.end ())
	{
		uae_u8*	realMem = get_real_address ((*iter).fPC);
		do_put_mem_word (realMem, kOpcode_Tailpatch);
		++iter;
	}
}


/***********************************************************************
 *
 * FUNCTION:	Patches::CallHeadpatch
 *
 * DESCRIPTION:	If the given system function is head patched, then call
 *				the headpatch.  Return "handled" (which means whether
 *				or not to call the ROM function after this one).
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType Patches::CallHeadpatch (HeadpatchProc hp)
{
	CallROMType	handled = kExecuteROM;

	if (hp)
	{
		if (hp != &SysHeadpatch::HostControl)
		{
			// Stop all profiling activities. Stop cycle counting and stop the
			// recording of function entries and exits.  We want our trap patches
			// to be as transparent as possible.

			StDisableAllProfiling	stopper;

			handled = hp ();
		}
		else
		{
			handled = hp ();
		}
	}

	return handled;
}


/***********************************************************************
 *
 * FUNCTION:	Patches::CallTailpatch
 *
 * DESCRIPTION:	If the given function is tail patched, then call the
 *				tailpatch.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Patches::CallTailpatch (TailpatchProc tp)
{
	// Stop all profiling activities. Stop cycle counting and stop the
	// recording of function entries and exits.  We want our trap patches
	// to be as transparent as possible.

	StDisableAllProfiling	stopper;

	if (tp)
	{
		tp ();
	}
}


/***********************************************************************
 *
 * FUNCTION:	Patches::OSVersion
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

DWord Patches::OSVersion (void)
{
	assert (gOSVersion != kOSUndeterminedVersion);

	return gOSVersion;
}


/***********************************************************************
 *
 * FUNCTION:	Patches::OSMajorMinorVersion
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

DWord Patches::OSMajorMinorVersion (void)
{
	return OSMajorVersion () * 10 + OSMinorVersion ();
}


/***********************************************************************
 *
 * FUNCTION:	Patches::OSMajorVersion
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

DWord Patches::OSMajorVersion (void)
{
	assert (gOSVersion != kOSUndeterminedVersion);

	return sysGetROMVerMajor (gOSVersion);
}


/***********************************************************************
 *
 * FUNCTION:	Patches::OSMinorVersion
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

DWord Patches::OSMinorVersion (void)
{
	assert (gOSVersion != kOSUndeterminedVersion);

	return sysGetROMVerMinor (gOSVersion);
}


/***********************************************************************
 *
 * FUNCTION:	SetSwitchApp
 *
 * DESCRIPTION:	Sets an application or launchable document to switch to
 *				the next time the system can manage it.
 *
 * PARAMETERS:	cardNo - the card number of the app to switch to.
 *
 *				dbID - the database id of the app to switch to.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Patches::SetSwitchApp (UInt cardNo, LocalID dbID)
{
	fgNextAppCardNo = cardNo;
	fgNextAppDbID = dbID;
}


/***********************************************************************
 *
 * FUNCTION:	Patches::SwitchToApp
 *
 * DESCRIPTION:	Switches to the given application or launchable document.
 *
 * PARAMETERS:	cardNo - the card number of the app to switch to.
 *
 *				dbID - the database id of the app to switch to.
 *
 * RETURNED:	Err number of any errors that occur
 *
 ***********************************************************************/

Err Patches::SwitchToApp (UInt cardNo, LocalID dbID)
{
	UInt	dbAttrs;
	ULong	type, creator;

	Err	err = ::DmDatabaseInfo (
				cardNo,
				dbID,
				NULL,				/*name*/
				&dbAttrs,
				NULL,				/*version*/
				NULL,				/*create date*/
				NULL,				/*modDate*/
				NULL,				/*backup date*/
				NULL,				/*modNum*/
				NULL,				/*appInfoID*/
				NULL,				/*sortInfoID*/
				&type,
				&creator);

	if (err)
		return err;

	//---------------------------------------------------------------------
	// If this is an executable, call SysUIAppSwitch
	//---------------------------------------------------------------------
	if (::IsExecutable (type, creator, dbAttrs))
	{
		err = ::SysUIAppSwitch (cardNo, dbID,
						sysAppLaunchCmdNormalLaunch, NULL);

		if (err)
			return err;
	}

	//---------------------------------------------------------------------
	// else, this must be a launchable data database. Find it's owner app
	//  and launch it with a pointer to the data database name.
	//---------------------------------------------------------------------
	else
	{
		DmSearchStateType	searchState;
		UInt				appCardNo;
		LocalID				appDbID;

		err = ::DmGetNextDatabaseByTypeCreator (true, &searchState,
						sysFileTApplication, creator,
						true, &appCardNo, &appDbID);
		if (err)
			return err;

		// Create the param block

		uaecptr	cmdPBP = (uaecptr) ::MemPtrNew (sizeof (SysAppLaunchCmdOpenDBType));
		if (cmdPBP == UAE_NULL)
			return memErrNotEnoughSpace;

		// Fill it in

		::MemPtrSetOwner ((VoidPtr) cmdPBP, 0);
		put_word (cmdPBP + offsetof (SysAppLaunchCmdOpenDBType, cardNo), cardNo);
		put_long (cmdPBP + offsetof (SysAppLaunchCmdOpenDBType, dbID), dbID);

		// Switch now

		err = ::SysUIAppSwitch (appCardNo, appDbID, sysAppLaunchCmdOpenDB, (Ptr) cmdPBP);
		if (err)
			return err;
	}

	if (fgQuitStage == kWaitingForSysUIAppSwitch)
	{
		fgQuitStage = kWaitingForSysAppStartup;
	}

	return errNone;
}


/***********************************************************************
 *
 * FUNCTION:	Patches::QuitOnAppExit
 *
 * DESCRIPTION:	Tells out patching mechanisms that the emulator should
 *				quit once the current application quits.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Patches::QuitOnAppExit (Bool val)
{
	if (val)
		fgQuitStage = kWaitingForSysUIAppSwitch;
	else
		fgQuitStage = kNoQuit;
}


/***********************************************************************
 *
 * FUNCTION:	Patches::PuppetString
 *
 * DESCRIPTION:	Puppet stringing function for inserting events into
 *				the system.  We want to insert events when:
 *
 *				- Gremlins is running
 *				- The user types characters
 *				- The user clicks with the mouse
 *				- We need to trigger a switch another application
 *
 *				This function is called from headpatches to
 *				SysEvGroupWait and SysSemaphoreWait.  When the Palm OS
 *				needs an event, it calls EvtGetEvent.  EvtGetEvent
 *				looks in all the usual places for events to return. If
 *				it doesn't find any, it puts the system to sleep by
 *				calling SysEvGroupWait (or SysSemaphoreWait on 1.0
 *				systems).  SysEvGroupWait will wake up and return when
 *				an event is posted via something like EvtEnqueuePenPoint,
 *				EvtEnqueueKey, or KeyHandleInterrupt.
 *
 *				To puppet-string Palm OS, we therefore patch those
 *				functions and post events, preventing them from actually
 *				going to sleep.
 *
 * PARAMETERS:	callROM - return here whether or not the original ROM
 *					function still needs to be called.  Normally, the
 *					answer is "yes".
 *
 *				clearTimeout - set to true if the "timeout" parameter
 *					of the function we've patched needs to be prevented
 *					from being "infinite".
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Patches::PuppetString (CallROMType& callROM, Bool& clearTimeout)
{
	callROM = kExecuteROM;
	clearTimeout = false;

	// Set the return value (Err) to zero in case we return
	// "true" (saying that we handled the trap).

	m68k_dreg (regs, 0) = 0;

	// If the low-memory global "idle" is true, then we're being
	// called from EvtGetEvent or EvtGetPen, in which case we
	// need to check if we need to post some events.

	if (LowMem::GetEvtMgrIdle ())
	{
		if (Gremlins_IsOn ())
		{
			if (gLastEvtTrap == sysTrapEvtGetEvent)
			{
				if (!Gremlins_PostFakeEvent ())
				{
					// No event was posted.  What we'd like right now
					// is to force EvtGetEvent to return a nil event.
					// We can do that by returning a non-zero result
					// code from SysEvGroupWait.  EvtGetEvent doesn't
					// look too closely at the result, but let's try
					// to be as close to reality as possible.
					// SysEvGroupWait currently returns "4" (CJ_WRTMOUT)
					// to indicate a timeout condition.  It should
					// probably get turned into sysErrTimeout
					// somewhere along the way, but that translation
					// doesn't seem to occur.

					m68k_dreg(regs, 0) = 4;

					callROM = kSkipROM;
					return;
				}
			}
			else if (gLastEvtTrap == sysTrapEvtGetPen)
			{
				Gremlins_PostFakePenEvent ();
			}

#if 0
			// Ensure that there is an event posted.  If there isn't
			// a timeout of zero (= forever) would kill us.

			assert (get_word (objID + 0x1E) == 1);
#endif

			// Never let the timeout be infinite.  If the above event-posting
			// attempts failed (which could happen, for instance, if we attempted
			// to post a pen event with the same coordinates as the previous
			// pen event), we'd end up waiting forever.

			clearTimeout = true;
		}

		// Gremlins aren't on; let's see if the user has typed some
		// keys that we need to pass on to the Palm device.

		else if (gKeyQueue.GetUsed () > 0)
		{
			StubAppEnqueueKey (gKeyQueue.Get (), 0, 0);
		}

		// No key events, let's see if there are pen events.

		else if (Hardware::fgHavePenEvent)
		{
			Point	pen = { -1, -1 };
			if (Hardware::fgPenIsDown)
			{
				pen = Hardware::fgPenLocation;
			}

			Hardware::fgHavePenEvent = false;

			StubAppEnqueuePt (&pen);
		}

		// E. None of the above.  Let's see if there's an app
		//    we're itching to switch to.

		else if (fgNextAppDbID != 0)
		{
			Err	err = SwitchToApp (fgNextAppCardNo, fgNextAppDbID);

			fgNextAppCardNo = 0;
			fgNextAppDbID = 0;

			clearTimeout = true;
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	Patches::HasWellBehavedMemSemaphoreUsage
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

Bool Patches::HasWellBehavedMemSemaphoreUsage (void)
{
	// Palm OS 3.0 and later should not be holding the memory manager
	// semaphore for longer than 1 minute.  I imagine that older ROMs
	// don't hold the semaphore for longer than this, but Roger still
	// suggested testing for 3.0.

	return gOSVersion != kOSUndeterminedVersion && OSMajorMinorVersion () >= 30;
}


/***********************************************************************
 *
 * FUNCTION:	Patches::EnterMemMgr
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Patches::EnterMemMgr (const char* fnName)
{
//	if (fgMemMgrCount == 0)
//	{
//		for (int ii = 0; ii < fgMemMgrCount; ++ii)
//			fputc ('\t', LogFile ());
//		fprintf (LogFile (), "Entering %s.\n", fnName);
//	}

	++fgMemMgrCount;
	assert (fgMemMgrCount < 10);

//	MetaMemory::ValidateHeaps ();
}


/***********************************************************************
 *
 * FUNCTION:	Patches::ExitMemMgr
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Patches::ExitMemMgr (const char* fnName)
{
//	MetaMemory::ValidateHeaps ();

	--fgMemMgrCount;
	assert (fgMemMgrCount >= 0);

//	if (fgMemMgrCount == 0)
//	{
//		for (int ii = 0; ii < fgMemMgrCount; ++ii)
//			fputc ('\t', LogFile ());
//		fprintf (LogFile (), "Exiting %s.\n", fnName);
//	}
}


/***********************************************************************
 *
 * FUNCTION:	Patches::CollectCurrentAppInfo
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void Patches::CollectCurrentAppInfo (void)
{
	// Clear out everything we know about the current application.

	memset (&Patches::fgCurAppInfo, 0, sizeof (Patches::fgCurAppInfo));

	// Get the AppInfo block for the current app

	uaecptr			appInfoP;
	SysAppInfoPtr	unusedAppInfoP;

	if (LowMem::TrapExists (sysTrapSysGetAppInfo))
		appInfoP = (uaecptr) SysGetAppInfo (&unusedAppInfoP, &unusedAppInfoP);
	else
		appInfoP = (uaecptr) SysCurAppInfoPV20 ();

	// Scarf some information out of it.

	Patches::fgCurAppInfo.fDB = (DmOpenRef) get_long (appInfoP + offsetof (SysAppInfoType, dbP));
	Patches::fgCurAppInfo.fStackP = get_long (appInfoP + offsetof (SysAppInfoType, stackP));
	Patches::fgCurAppInfo.fMemOwnerID = get_word (appInfoP + offsetof (SysAppInfoType, memOwnerID));

	// Determine the current stack range.  Under Palm OS 3.0 and later, this information
	// is in the AppInfo block.  Under earlier OSes, we only get the low-end of the stack
	// (that is, the address that was returned by MemPtrNew).  To get the high-end of
	// the stack, assume that stackP pointed to a chunk of memory allocated by MemPtrNew
	// and call MemPtrSize.

	if (AppInfoHasStackInfo ())
	{
		if (Patches::fgCurAppInfo.fStackP)
		{
			ULong	stackSize = ::MemPtrSize ((VoidPtr) Patches::fgCurAppInfo.fStackP);
			if (stackSize)
			{
				Patches::fgCurAppInfo.fStackEndP = Patches::fgCurAppInfo.fStackP + stackSize;
			}
		}
	}
	else
	{
		Patches::fgCurAppInfo.fStackEndP = get_long (appInfoP + offsetof (SysAppInfoType, stackEndP));
	}

	Patches::fgCurAppInfo.fStackSize = Patches::fgCurAppInfo.fStackEndP - Patches::fgCurAppInfo.fStackP;

	// Remember the current application name and version information.  We use
	// this information when telling users that something has gone haywire.
	//
	// If the database has a 'tAIN' resource, get the name from there.
	// Otherwise, use the database name.
	//
	// (Write the name into a temporary local variable.  The local variable is
	// on the stack, which will get "mapped" into the emulated address space so
	// that the emulated DmDatabaseInfo can get to it.)

	UInt	cardNo;
	LocalID	dbID;

	Err	err = DmOpenDatabaseInfo (Patches::fgCurAppInfo.fDB, &dbID, NULL, NULL, &cardNo, NULL);
	if (err)
		return;

	Patches::fgCurAppInfo.fCardNo = cardNo;
	Patches::fgCurAppInfo.fDBID = dbID;

	char	appName[dmDBNameLength] = {0};
	char	appVersion[256] = {0};	// <gulp> I hope this is big enough...

//	DmOpenRef	dbP = DmOpenDatabase (cardNo, dbID, dmModeReadOnly);

//	if (dbP)
	{
		VoidHand	strH;

		// Get the app name from the 'tAIN' resource.

		strH = ::DmGet1Resource (ainRsc, ainID);
		if (strH)
		{
			uaecptr	strP = (uaecptr) ::MemHandleLock (strH);
			uae_strcpy (appName, strP);
			::MemHandleUnlock (strH);
			::DmReleaseResource (strH);
		}

		// Get the version from the 'tver' resource, using ID's 1 and 1000

		strH = ::DmGet1Resource (verRsc, appVersionID);
		if (strH == NULL)
			strH = DmGet1Resource (verRsc, appVersionAlternateID);
		if (strH)
		{
			uaecptr	strP = (uaecptr) ::MemHandleLock (strH);
			uae_strcpy (appVersion, strP);
			::MemHandleUnlock (strH);
			::DmReleaseResource (strH);
		}

//		::DmCloseDatabase (dbP);
	}

	if (appName[0] == 0)	// No 'tAIN' resource, so use database name
	{
		::DmDatabaseInfo (cardNo, dbID,
						appName,
						NULL, NULL, NULL, NULL, NULL,
						NULL, NULL, NULL, NULL, NULL);
	}

	// Copy the strings from the stack to their permanent homes.

	strcpy (Patches::fgCurAppInfo.fName, appName);
	strcpy (Patches::fgCurAppInfo.fVersion, appVersion);
}


#pragma mark -

// ===========================================================================
//		 SysHeadpatch
// ===========================================================================

/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::RecordTrapNumber
 *
 * DESCRIPTION:	Record the trap we're executing for our patch to
 *				SysEvGroupWait later.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::RecordTrapNumber (void)
{
	struct StackFrame
	{
	};

	uae_u8*	realMem = get_real_address (m68k_getpc ());

	assert (do_get_mem_word (realMem - 2) == (m68kTrapInstr + sysDispatchTrapNum));

	gLastEvtTrap = do_get_mem_word (realMem);

	return kExecuteROM;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::DbgMessage
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::DbgMessage (void)
{
	// void DbgMessage(CharPtr aStr);

	struct StackFrame
	{
		CharPtr aStr;
	};

	uaecptr	msg	= GET_PARAMETER (aStr);

	if (msg)
	{
		string	msgCopy;
		size_t	msgLen = uae_strlen (msg);

		if (msgLen > 0)
		{
			msgCopy.resize (msgLen);
			uae_strcpy (&msgCopy[0], msg);
		}

		Debug::SendMessage (msgCopy.c_str ());
	}

	return kSkipROM;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::DmInit
 *
 * DESCRIPTION:	After MemInit is called, we need to sync up with the
 *				initial state of the heap(s).  However, MemInit is not
 *				called via the trap table, so we can't easily tailpatch
 *				it.  DmInit is the first such function called after
 *				MemInit, so we headpatch *it* instead of tailpatching
 *				MemInit.
 *
 *				(Actually, MemHeapCompact is called as one of the last
 *				 things MemInit does which makes it an interesting
 *				 candidate for patching in order to sync up with the
 *				 heap state.  However, if we were to do a full sync on
 *				 that call, a full sync would occur on *every* call to
 *				 MemHeapCompact, which we don't really want to do.)
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::DmInit (void)
{
	MetaMemory::SyncWithHeap (0, true);

	return kExecuteROM;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::ErrDisplayFileLineMsg
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::ErrDisplayFileLineMsg (void)
{
	// void ErrDisplayFileLineMsg(CharPtr filename, UInt lineno, CharPtr msg)

	CEnableFullAccess	munge;	// Remove blocks on memory access.

	// Force this guy to true.  If it's false, ErrDisplayFileLineMsg will
	// just try to enter the debugger.

	Word	sysMiscFlags = LowMem_GetGlobal (sysMiscFlags);
	LowMem_SetGlobal (sysMiscFlags, sysMiscFlags | sysMiscFlagUIInitialized);

	// Clear this low-memory flag so that we force the dialog to display.
	// If this flag is true, ErrDisplayFileLineMsg will just try to enter
	// the debugger.

	LowMem_SetGlobal (dbgWasEntered, false);

	return kExecuteROM;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::EvtAddEventToQueue
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::EvtAddEventToQueue (void)
{
	// void EvtAddEventToQueue (const EventPtr event)

	struct StackFrame
	{
		const EventPtr	event;
	};

	uaecptr	event	= GET_PARAMETER (event);

	LogEvtAddEventToQueue (event);

	return kExecuteROM;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::EvtAddUniqueEventToQueue
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::EvtAddUniqueEventToQueue (void)
{
	// void EvtAddUniqueEventToQueue(const EventPtr eventP, const DWord id, const Boolean inPlace)

	struct StackFrame
	{
		const EventPtr	event;
		const DWord		id;
		const Boolean	inPlace;
	};

	uaecptr	event	= GET_PARAMETER (event);
	DWord	id		= GET_PARAMETER (id);
	Boolean	inPlace	= GET_PARAMETER (inPlace);

	LogEvtAddUniqueEventToQueue (event, id, inPlace);

	return kExecuteROM;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::EvtEnqueueKey
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::EvtEnqueueKey (void)
{
	// Err EvtEnqueueKey(UInt ascii, UInt keycode, UInt modifiers)

	struct StackFrame
	{
		UInt	ascii;
		UInt	keycode;
		UInt	modifiers;
	};

	UInt	ascii		= GET_PARAMETER (ascii);
	UInt	keycode		= GET_PARAMETER (keycode);
	UInt	modifiers	= GET_PARAMETER (modifiers);

	LogEvtEnqueueKey (ascii, keycode, modifiers);

	return kExecuteROM;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::EvtEnqueuePenPoint
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::EvtEnqueuePenPoint (void)
{
	// Err EvtEnqueuePenPoint(PointType* ptP)

	struct StackFrame
	{
		PointType*	ptP;
	};

	uaecptr	ptP		= GET_PARAMETER (ptP);

	LogEvtEnqueuePenPoint (ptP);

	return kExecuteROM;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::HostControl
 *
 * DESCRIPTION:	This one's kind of odd.  Originally, there was
 *				SysGremlins, which was declared as follows:
 *
 *					DWord SysGremlins(GremlinFunctionType selector,
 *										GremlinParamsType *params)
 *
 *				Also originally, the only defined selector was
 *				GremlinIsOn.
 *
 *				Now, SysGremlins is extended to be SysHostControl,
 *				which allows the Palm environment to access host
 *				functions if it's actually running under the simulator
 *				or emulator.
 *
 *				Because of this extension, functions implemented via
 *				this trap are not limited to pushing a selector and
 *				parameter block on the stack. Now, they will all push
 *				on a selector, but what comes after is dependent on the
 *				selector.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::HostControl (void)
{
	return HandleHostControlCall ();
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::HwrBatteryLevel
 *
 * DESCRIPTION:	Return that the battery is always full.  HwrBatteryLevel
 *				is the bottleneck function called to determine the
 *				battery level.  By patching it this way, we don't have
 *				to emulate the hardware registers that report the
 *				battery level.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::HwrBatteryLevel (void)
{
	// UInt HwrBatteryLevel(void)

	struct StackFrame
	{
	};

	m68k_dreg (regs, 0) = 255;	// Hardcode a maximum level

	return kSkipROM;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::HwrGetROMToken
 *
 * DESCRIPTION:	Patch this guy so that we never return the 'irda' token.
 *				We should take this out when some sort of IR support is
 *				added.
 *
 *				NOTE: This patch is useless for diverting the ROM. It
 *				calls HwrGetROMToken directly, which means that it will
 *				bypass this patch.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::HwrGetROMToken (void)
{
	//	Err HwrGetROMToken (Word cardNo, DWord token, BytePtr *dataP, WordPtr sizeP)

	struct StackFrame
	{
		Word cardNo;
		DWord token;
		BytePtr *dataP;
		WordPtr sizeP;
	};

	Word	cardNo = GET_PARAMETER (cardNo);
	DWord	token = GET_PARAMETER (token);
	uaecptr	dataP = GET_PARAMETER (dataP);
	uaecptr	sizeP = GET_PARAMETER (sizeP);

	if (cardNo == 0 && token == hwrROMTokenIrda)
	{
		if (dataP)
			put_long (dataP, 0);

		if (sizeP)
			put_long (sizeP, 0);

		m68k_dreg (regs, 0) = ~0;		// token not found.

		return kSkipROM;
	}

	return kExecuteROM;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::HwrSleep
 *
 * DESCRIPTION:	Record whether or not we are sleeping and update the
 *				boolean that determines if low-memory access is OK.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::HwrSleep (void)
{
	// void HwrSleep(Boolean untilReset, Boolean emergency)

	struct StackFrame
	{
		Boolean untilReset;
		Boolean dummy;
		Boolean emergency;
	};

	// HwrSleep changes the exception vectors for for the interrupts,
	// so temporarily unlock those.  We'll re-establish them in the
	// HwrSleep tailpatch.

	MetaMemory::UnmarkLowMemory (offsetof (M68KExcTableType, busErr),
								 offsetof (M68KExcTableType, busErr) + sizeof (uaecptr));

	MetaMemory::UnmarkLowMemory (offsetof (M68KExcTableType, addressErr),
								 offsetof (M68KExcTableType, addressErr) + sizeof (uaecptr));

	MetaMemory::UnmarkLowMemory (offsetof (M68KExcTableType, illegalInstr),
								 offsetof (M68KExcTableType, illegalInstr) + sizeof (uaecptr));

	MetaMemory::UnmarkLowMemory (offsetof (M68KExcTableType, autoVec1),
								 offsetof (M68KExcTableType, trapN[0]));

	return kExecuteROM;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::KeyCurrentState
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

// From Roger:
//
//	I was thinking a bit more about Gremlins and games yesterday.  Games which
//	use the buttons as input have been a special case for Gremlins because
//	Gremlins only faked events and pen points.  Normally games have to fake
//	their own button mask, and they get better results if have some buttons
//	held down more often than others.
//
//	Now I'm thinking different.  With Poser, the KeyCurrentState call should
//	be stolen when Gremlins is running.  All of the keys possible should have
//	their button bits on half the time.  This will allow users to test games.
//	By not tuning how often buttons should be held down, the testing process
//	will take longer to excerise all app functionality, but it's better than
//	now.  App developers can override the default Gremlin values with their
//	own for better results.
//
//	To successfully test this, A game like SubHunt should play on it's own for
//	a least a while.  HardBall is an example of a game which would really
//	benefit from recognizing Gremlins is running and tune itself to make the
//	testing more effective.  It should grant infinite balls until after the
//	last level is finished.
//
//	I actually think this is important enough to be for Acorn because it
//	enables users to test a large class of apps which otherwise can't.  I
//	think it's easy to implement.  Basically just have KeyCurrentState return
//	the int from the random number generator.  Each bit should be on about
//	half the time.
//
// Follow-up commentary: it turns out that this patch is not having the
// effect we hoped.  SubHunt was so overwhelmed with events from Gremlins
// that it rarely had the chance to call KeyCurrentState.  We're thinking
// of Gremlins backing off on posting events if the SysGetEvent sleep time
// is small (i.e., not "forever"), but we'll have to think about the impact
// on other apps first.

CallROMType SysHeadpatch::KeyCurrentState (void)
{
	// DWord KeyCurrentState(void)

	struct StackFrame
	{
	};

	if (Gremlins_IsOn ())
	{
		// Let's try setting each bit 1/4 of the time.

		uae_u32	bits = rand () & rand ();

		m68k_dreg (regs, 0) = bits;

		return kSkipROM;
	}

	return kExecuteROM;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::MemMove
 *
 * DESCRIPTION:	Implement MemMove with native code for speed.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::MemMove (void)
{
	// Err MemMove(const VoidPtr dstP, const void * const sP, ULong numBytes)

	struct StackFrame
	{
		const VoidPtr dstP;
		const void * const sP;
		ULong numBytes;
	};

#if !HAS_PROFILING
	uaecptr	dP			= GET_PARAMETER (dstP);
	uaecptr	sP			= GET_PARAMETER (sP);
	uae_u32	numBytes	= GET_PARAMETER (numBytes);

//	ErrFatalDisplayIf(dP == NULL, "MemMove to NULL");
//	ErrNonFatalDisplayIf(sP == NULL, "MemMove from NULL");

	if (numBytes != 0 && dP != sP)
	{
		uae_memmove (dP, sP, numBytes);
		MetaMemory::MoveUninitialized (dP, sP, numBytes);
	}

	m68k_dreg (regs, 0) = 0;

	return kSkipROM;
#else
	return kExecuteROM;
#endif
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::MemSet
 *
 * DESCRIPTION:	Implement MemSet with native code for speed.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::MemSet (void)
{
	// Err MemSet(const VoidPtr dstP, ULong numBytes, Byte value)

	struct StackFrame
	{
		const VoidPtr dstP;
		ULong numBytes;
		Byte value;
	};

#if !HAS_PROFILING
	uaecptr	dstP		= GET_PARAMETER (dstP);
	ULong	numBytes	= (ULong) GET_PARAMETER (numBytes);
	Byte	value		= (Byte) GET_PARAMETER (value);

//	ErrFatalDisplayIf (dstP == NULL, "MemSet to NULL");

	uae_memset (dstP, value, numBytes);

	MetaMemory::MarkInitialized (dstP, dstP + numBytes);

	m68k_dreg (regs, 0) = 0;

	return kSkipROM;
#else
	return kExecuteROM;
#endif
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::PenOpen
 *
 * DESCRIPTION:	This is where the pen calibration information is read.
 *				Preflight this call to add the calibration information
 *				to the preferences database if it doesn't exist.  That
 *				way, we don't have to continually calibrate the screen
 *				when we boot up.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::PenOpen (void)
{
	// Err PenOpen(void)

	struct StackFrame
	{
	};

#if !TIME_STARTUP

	ResetCalibrationInfo ();

#endif

	return kExecuteROM;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::RctPtInRectangle
 *
 * DESCRIPTION:	Implement RctPtInRectangle with native code for speed.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::RctPtInRectangle (void)
{
	// Boolean RctPtInRectangle (const SWord x, const SWord y, const RectanglePtr r)

	struct StackFrame
	{
		const SWord x;
		const SWord y;
		const RectanglePtr r;
	};

#if !HAS_PROFILING
	SWord	x	= (SWord) GET_PARAMETER (x);
	SWord	y	= (SWord) GET_PARAMETER (y);
	uaecptr	rP	= GET_PARAMETER (r);

	RectangleType	r;
	r.topLeft.x = get_word (rP + offsetof (RectangleType, topLeft) + offsetof (PointType, x));
	r.topLeft.y = get_word (rP + offsetof (RectangleType, topLeft) + offsetof (PointType, y));
	r.extent.x = get_word (rP + offsetof (RectangleType, extent) + offsetof (PointType, x));
	r.extent.y = get_word (rP + offsetof (RectangleType, extent) + offsetof (PointType, y));

	Boolean result = (x >= r.topLeft.x &&
			  y >= r.topLeft.y &&
			  x < r.topLeft.x + r.extent.x &&
			  y < r.topLeft.y + r.extent.y);

	m68k_dreg (regs, 0) = result;

	return kSkipROM;
#else
	return kExecuteROM;
#endif
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::SysAppExit
 *
 * DESCRIPTION:	If the application calling SysAppExit was launched as a
 *				full application, then "forget" any information we have
 *				about it.  When the next application is launched, we'll
 *				collect information on it in SysAppStartup.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::SysAppExit (void)
{
	// Err SysAppExit(SysAppInfoPtr appInfoP, Ptr prevGlobalsP, Ptr globalsP)

	struct StackFrame
	{
		SysAppInfoPtr appInfoP;
		Ptr prevGlobalsP;
		Ptr globalsP;
	};

	uaecptr	appInfoP		= GET_PARAMETER (appInfoP);
	uaecptr	prevGlobalsP	= GET_PARAMETER (prevGlobalsP);
	uaecptr	globalsP		= GET_PARAMETER (globalsP);

	if (!appInfoP)
		return kExecuteROM;

	Int	cmd = get_word (appInfoP + offsetof (SysAppInfoType, cmd));

	if (cmd == sysAppLaunchCmdNormalLaunch)
	{
		if (Patches::fgQuitStage == kWaitingForSysAppExit)
		{
			Patches::fgQuitStage = kTimeToQuit;
		}

		memset (&Patches::fgCurAppInfo, 0, sizeof (Patches::fgCurAppInfo));
	}

	return kExecuteROM;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::SysBinarySearch
 *
 * DESCRIPTION:	There's a bug in pre-3.0 versions of SysBinarySearch
 *				that cause it to call the user callback function with a
 *				pointer just past the array to search.  Make a note of
 *				when we enter SysBinarySearch so that we can make
 *				allowances for that in MetaMemory's memory checking.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::SysBinarySearch (void)
{
	assert (Patches::fgSysBinarySearchCount < 10);
	Patches::fgSysBinarySearchCount++;

	return kExecuteROM;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::SysEvGroupWait
 *
 * DESCRIPTION:	We patch SysEvGroupWait as the mechanism for feeding
 *				the Palm OS new events.  See comments in
 *				Patches::PuppetString.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::SysEvGroupWait (void)
{
	// Err SysEvGroupWait(DWord evID, DWord mask, DWord value, SDWord matchType,
	//						 SDWord timeout)

	struct StackFrame
	{
		DWord evID;
		DWord mask;
		DWord value;
		SDWord matchType;
		SDWord timeout;
	};

	// Only do this under 2.0 and later.  Under Palm OS 1.0, EvtGetSysEvent
	// called SysSemaphoreWait instead.  See our headpatch of that function
	// for a chunk of pretty similar code.

	if (Patches::OSMajorVersion () == 1)
	{
		return kExecuteROM;
	}

	CallROMType	result;
	Bool		clearTimeout;

	Patches::PuppetString (result, clearTimeout);

	// If timeout is infinite, the kernel wants 0.
	// If timeout is none, the kernel wants -1.

	if (clearTimeout && GET_PARAMETER (timeout) == 0)
	{
		SET_PARAMETER (timeout, (uae_u32) -1);
	}

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::SysFatalAlert
 *
 * DESCRIPTION:	Intercept this and show the user a dialog.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::SysFatalAlert (void)
{
	// UInt SysFatalAlert (CharPtr msg)

	struct StackFrame
	{
		CharPtr msg;
	};

	uaecptr	msg	= GET_PARAMETER (msg);

	string	msgString;

	if (msg)
	{
		msgString = PrvToString (msg);
	}
	else
	{
		msgString = Platform::GetString (kStr_UnknownFatalError);
	}

	int	button = Errors::SysFatalAlert (msgString.c_str ());

	switch (button)
	{
		case Errors::kDebug:	m68k_dreg (regs, 0) = fatalEnterDebugger;	break;
		case Errors::kReset:	m68k_dreg (regs, 0) = fatalReset;			break;
		case Errors::kContinue:	m68k_dreg (regs, 0) = fatalDoNothing;		break;
	}

	// Turn off Gremlins.  Do this _after_ calling SysFatalAlert
	// so that that function can pick up any active Gremlin information.

	Gremlins_Stop ();

	return kSkipROM;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::SysLaunchConsole
 *
 * DESCRIPTION:	Stub this function out so that it doesn't do anything.
 *				We completely handle the console in DebugMgr, so there's
 *				no need to get the ROM all heated up.  Also, there are
 *				problems with actually letting the ROM try to launch its
 *				console task, at least on the Mac.  It will try to open
 *				a serial port socket and do stuff with it.  That attempt
 *				will fail, as much of the serial port processing on the
 *				Mac is handled at idle time, and idle time processing is
 *				inhibited when handling debugger packets
 *				(SysLaunchConsole is usually called by a debugger via
 *				the RPC packet).  Since serial port processing doesn't
 *				occur, SysLaunchConsole hangs.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::SysLaunchConsole (void)
{
	// Err SysLaunchConsole(void)

	struct StackFrame
	{
	};

	m68k_dreg (regs, 0) = 0;		// no error

	return kSkipROM;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::SysSemaphoreWait
 *
 * DESCRIPTION:	We patch SysSemaphoreWait as the mechanism for feeding
 *				the Palm OS new events. See comments in
 *				Patches::PuppetString.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::SysSemaphoreWait (void)
{
	// Err SysSemaphoreWait(DWord smID, DWord priority, SDWord timeout)

	struct StackFrame
	{
		DWord smID;
		DWord priority;
		SDWord timeout;
	};

	// Only do this under 1.0.  Under Palm OS 2.0 and later, EvtGetSysEvent
	// calls SysEvGroupWait instead.  See our headpatch of that function
	// for a chunk of pretty similar code.

	if (Patches::OSMajorVersion () != 1)
	{
		return kExecuteROM;
	}

	CallROMType	result;
	Bool		clearTimeout;

	Patches::PuppetString (result, clearTimeout);

	// If timeout is infinite, the kernel wants 0.
	// If timeout is none, the kernel wants -1.

	if (clearTimeout && GET_PARAMETER (timeout) == 0)
	{
		SET_PARAMETER (timeout, (uae_u32) -1);
	}

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::SysUIAppSwitch
 *
 * DESCRIPTION:	SysUIAppSwitch is called from the following locations
 *				for the given reasons.  When running Gremlins, we want
 *				to prevent SysUIAppSwitch from doing its job, which is
 *				to record information about the application to switch
 *				to and to then post an appStopEvent to the current
 *				application.
 *
 *				There are a couple of places where SysUIAppSwitch is
 *				called to quit and re-run the current application.
 *				Therefore, we want to stub out SysUIAppSwitch only when
 *				the application being switched to is not the currently
 *				running application.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

//	Places where this is called:
//
//	- LauncherMain.c (AppsViewSwitchApp) to launch new app.
//	- PrefApp (PilotMain) to launch a panel.
//	- SyncApp (prvLaunchExternalModule) to launch a "service owner"
//	- AppLaunchCmd.h (AppLaunchWithCommand) ???
//	- ModemPanel.h (ModemPanelLaunch) ???
//	- callApp.h (LaunchApp) ???
//	- ExgMgr.c (ExgNotifyReceive) to send a sysAppLaunchCmdGoTo message.
//	- GraffitiGlue.c (PrvLaunchGraffitiDemo) to launch Graffiti demo.
//	- Keyboard.c (PrvLaunchGraffitiDemo) to launch Graffiti demo.
//	- Launcher.c (LauncherFormHandleEvent) handles taps on launcher icons.
//	- SystemMgr.c (SysHandleEvent) to send sysAppLaunchCmdSystemLock
//	  in response to seeing a lockChr keyboard message.
//	- SystemMgr.c (SysHandleEvent) to switch apps in response to hard#Chr
//	  keyboard messages.
//	- Find.c (Find) to send sysAppLaunchCmdGoTo message.
//
//	- ButtonsPanel.c (ButtonsFormHandleEvent) switch to another panel.
//	- FormatsPanel.c (FormatsFormHandleEvent) switch to another panel.
//	- GeneralPanel.c (GeneralFormHandleEvent) switch to another panel.
//	- ModemPanel.c (ModemFormHandleEvent) switch to another panel.
//	- NetworkPanel.c (NetworkFormHandleEvent) switch to another panel.
//	- OwnerPanel.c (OwnerViewHandleEvent) switch to another panel.
//	- ShortCutsPanel.c (ShortCutFormHandleEvent) switch to another panel.

CallROMType SysHeadpatch::SysUIAppSwitch (void)
{
	// Err SysUIAppSwitch(UInt cardNo, LocalID dbID, Word cmd, Ptr cmdPBP)

	struct StackFrame
	{
		UInt cardNo;
		LocalID dbID;
		Word cmd;
		Ptr cmdPBP;
	};

	UInt	cardNo	= (UInt) GET_PARAMETER (cardNo);
	LocalID	dbID	= (LocalID) GET_PARAMETER (dbID);
	Word	cmd		= (Word) GET_PARAMETER (cmd);

	// Get the current application card and db.  If we are attempting to switch
	// to the same app, allow it.

	UInt	currentCardNo;
	LocalID	currentDbID;
	Err		err = SysCurAppDatabase (&currentCardNo, &currentDbID);

	// Got an error? Give up and let default handling occur.

	if (err != 0)
		return kExecuteROM;

	// Switching to current app; let default handling occur.

	if ((cardNo == currentCardNo) && (dbID == currentDbID))
		return kExecuteROM;

	// OK, we're switching to a different application.  If Gremlins is running
	// and running in "single app" mode, then stub out SysUIAppSwitch to do nothing.

	if (Gremlins_IsOn () && !Gremlins_CanSwitchToApp (cardNo, dbID))
	{
		m68k_dreg (regs, 0) = 0;
		return kSkipROM;
	}

	// Do the normal app switch.

	return kExecuteROM;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::SysUIBusy
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

extern uae_u32	gStart;

CallROMType SysHeadpatch::SysUIBusy (void)
{
	// Word SysUIBusy(Boolean set, Boolean value)

	struct StackFrame
	{
		Boolean set;
		Boolean value;
	};


#if TIME_STARTUP

	static calledOnce;

	if (!calledOnce)
	{
		calledOnce = true;

		uae_u32	now = Platform::GetMilliseconds ();
		uae_u32	elapsed = now - gStart;

		char	buffer[200];

		sprintf (buffer, "startup = %ld milliseconds  "
						"instructions = %ld "
						elapsed, Emulator::GetInstructionCount ());

		Platform::ReportString (buffer);
	}

#endif

	return kExecuteROM;
}


/***********************************************************************
 *
 * FUNCTION:	SysHeadpatch::TimInit
 *
 * DESCRIPTION:	TimInit is where the date is read from non-volatile
 *				memory into private Time Manager memory.  After that,
 *				the Time Manager uses its private copy to update the
 *				non-volatile copy.  Thus, stashing correct values of
 *				the date into non-volatile memory should be done just
 *				before TimInit.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType SysHeadpatch::TimInit (void)
{
	// Err TimInit(void)

	struct StackFrame
	{
	};

	// Get the current date.

	long	year, month, day;
	::GetHostDate (&year, &month, &day);

	// Get the current non-volatile variables.

	SysNVParamsType	params[100];	// Allocate much more memory than we need
									// in case the ROM's notion of SysNVParamsType
									// is different from ours.
	::MemNVParams (false, &params[0]);

	// Update the "hours adjustment" value to contain the current date.

	params[0].rtcHours = ::DateToDays (year, month, day) * 24;

	// Re-write the non-volatile variables.

	::MemNVParams (true, &params[0]);

	return kExecuteROM;
}


#pragma mark -

// ===========================================================================
//		 SysTailpatch
// ===========================================================================

/***********************************************************************
 *
 * FUNCTION:	SysTailpatch::EvtGetEvent
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void SysTailpatch::EvtGetEvent (void)
{
	// void EvtGetEvent(const EventPtr event, SDWord timeout);

	struct StackFrame
	{
		const EventPtr event;
		SDWord timeout;
	};

	uaecptr	event = GET_PARAMETER (event);
	SDWord	timeout = GET_PARAMETER (timeout);

	LogEvtGetEvent (event, timeout);
}


/***********************************************************************
 *
 * FUNCTION:	SysTailpatch::EvtGetPen
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void SysTailpatch::EvtGetPen (void)
{
	// void EvtGetPen(SWord *pScreenX, SWord *pScreenY, Boolean *pPenDown)

	struct StackFrame
	{
		SWord *pScreenX;
		SWord *pScreenY;
		Boolean *pPenDown;
	};

	uaecptr	pScreenX = GET_PARAMETER (pScreenX);
	uaecptr	pScreenY = GET_PARAMETER (pScreenY);
	uaecptr	pPenDown = GET_PARAMETER (pPenDown);

	LogEvtGetPen (pScreenX, pScreenY, pPenDown);
}


/***********************************************************************
 *
 * FUNCTION:	SysTailpatch::EvtGetSysEvent
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void SysTailpatch::EvtGetSysEvent (void)
{
	// void EvtGetSysEvent(EventPtr eventP, Long timeout)


	struct StackFrame
	{
		EventPtr event;
		Long timeout;
	};

	uaecptr	event = GET_PARAMETER (event);
	Long	timeout = GET_PARAMETER (timeout);

	LogEvtGetSysEvent (event, timeout);
}


/***********************************************************************
 *
 * FUNCTION:	SysTailpatch::FtrInit
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void SysTailpatch::FtrInit (void)
{
	// void FtrInit(void)

	struct StackFrame
	{
	};

	// Get information about the current OS so that we know
	// what features are implemented (for those cases when we
	// can't make other tests, like above where we test for
	// the existance of a trap before calling it).

	// Read the version into a local variable; the ROM Stub facility
	// automatically maps local variables into Palm space so that ROM
	// functions can get to them.  If we were to pass &gOSVersion,
	// the DummyBank functions would complain about an invalid address.

	DWord	version;
	Err		err = FtrGet (sysFtrCreator, sysFtrNumROMVersion, &version);

	if (!err)
	{
		gOSVersion = version;
	}
	else
	{
		gOSVersion = kOSUndeterminedVersion;
	}
}


/***********************************************************************
 *
 * FUNCTION:	SysTailpatch::HwrMemReadable
 *
 * DESCRIPTION:	Patch this function so that it returns non-zero if the
 *				address is in the range of memory that we've mapped in
 *				to emulated space.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void SysTailpatch::HwrMemReadable (void)
{
	// DWord HwrMemReadable(VoidPtr address)

	struct StackFrame
	{
		VoidPtr address;
	};

	uaecptr	address = GET_PARAMETER (address);
	DWord	result = m68k_dreg (regs, 0);

	if (result == 0)
	{
		void*	addrStart;
		uae_u32	addrLen;

		Memory::GetMappingInfo ((void*) address, &addrStart, &addrLen);

		m68k_dreg (regs, 0) = addrLen;
	}
}


/***********************************************************************
 *
 * FUNCTION:	SysTailpatch::HwrSleep
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void SysTailpatch::HwrSleep (void)
{
	// void HwrSleep(Boolean untilReset, Boolean emergency)

	struct StackFrame
	{
		Boolean untilReset;
		Boolean dummy;
		Boolean emergency;
	};

	MetaMemory::MarkLowMemory (offsetof (M68KExcTableType, busErr),
							   offsetof (M68KExcTableType, busErr) + sizeof (uaecptr));

	MetaMemory::MarkLowMemory (offsetof (M68KExcTableType, addressErr),
							   offsetof (M68KExcTableType, addressErr) + sizeof (uaecptr));

	MetaMemory::MarkLowMemory (offsetof (M68KExcTableType, illegalInstr),
							   offsetof (M68KExcTableType, illegalInstr) + sizeof (uaecptr));

	MetaMemory::MarkLowMemory (offsetof (M68KExcTableType, autoVec1),
							   offsetof (M68KExcTableType, trapN[0]));
}


/***********************************************************************
 *
 * FUNCTION:	SysTailpatch::SysAppStartup
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void SysTailpatch::SysAppStartup (void)
{
	// Err SysAppStartup(SysAppInfoPtr* appInfoPP, Ptr* prevGlobalsP, Ptr* globalsPtrP)

	struct StackFrame
	{
		SysAppInfoPtr* appInfoPP;
		Ptr* prevGlobalsP;
		Ptr* globalsPtrP;
	};

	uaecptr	appInfoPP = GET_PARAMETER (appInfoPP);
	if (!appInfoPP)
		return;

	uaecptr	appInfoP = get_long (appInfoPP);
	if (!appInfoP)
		return;

	Int	cmd = get_word (appInfoP + offsetof (SysAppInfoType, cmd));

	if (cmd == sysAppLaunchCmdNormalLaunch)
	{
		Patches::CollectCurrentAppInfo ();

		if (Patches::fgQuitStage == kWaitingForSysAppStartup)
		{
			Patches::fgQuitStage = kWaitingForSysAppExit;
		}

		// Clear the flags that tell us which warnings we've already issued.
		// We reset these flags any time a new application is launched.

		Errors::ClearWarningFlags ();
	}
}


/***********************************************************************
 *
 * FUNCTION:	SysTailpatch::SysBinarySearch
 *
 * DESCRIPTION:	There's a bug in pre-3.0 versions of SysBinarySearch
 *				that cause it to call the user callback function with a
 *				pointer just past the array to search.  Make a note of
 *				when we enter SysBinarySearch so that we can make
 *				allowances for that in MetaMemory's memory checking.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void SysTailpatch::SysBinarySearch (void)
{
	Patches::fgSysBinarySearchCount--;
	assert (Patches::fgSysBinarySearchCount >= 0);
}


/***********************************************************************
 *
 * FUNCTION:	SysTailpatch::SysLibFind
 *
 * DESCRIPTION:	SysTailpatch to remember the refnums returned for
 *				various libraries.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void SysTailpatch::SysLibFind (void)
{
	// Err SysLibFind(CharPtr nameP, UIntPtr refNumP)

	struct StackFrame
	{
		CharPtr	nameP;
		UIntPtr	refNumP;
	};

	string	name = PrvToString (GET_PARAMETER (nameP));
	UInt	refNum = get_word (GET_PARAMETER (refNumP));

	if (refNum != sysInvalidRefNum)
	{
		if (refNum >= gLibPatches.size ())
		{
			gLibPatches.resize (refNum + 1);
		}

		if (gLibPatches[refNum] == NULL)
		{
			Patches::InstallLibPatch (name, refNum, gLibPatches[refNum]);
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	SysTailpatch::TimInit
 *
 * DESCRIPTION:	TimInit is where a special boolean is set to trigger a
 *				bunch of RTC bug workarounds in the ROM (that is, there
 *				are RTC bugs in the Dragonball that the ROM needs to
 *				workaround).  We're not emulating those bugs, so turn
 *				off that boolean.  Otherwise, we'd have to add support
 *				for the 1-second, 1-minute, and 1-day RTC interrupts in
 *				order to get those workarounds to work.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void SysTailpatch::TimInit (void)
{
	// Err TimInit(void)

	struct StackFrame
	{
	};

	// Turn off the RTC bug workaround flag.

	uaecptr	timGlobalsP = LowMem_GetGlobal (timGlobalsP);

	if (get_byte (timGlobalsP + 8) == 0x01)
	{
		put_byte (timGlobalsP + 8, 0x00);
	}
}


/***********************************************************************
 *
 * FUNCTION:	SysTailpatch::UIInitialize
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void SysTailpatch::UIInitialize (void)
{
	// Void UIInitialize (void)

	struct StackFrame
	{
	};

	Patches::fgUIInitialized = true;

	// Prevent the device from going to sleep.

	SysSetAutoOffTime (0);

	// Can't call PrefSetPreference on 1.0 devices....

	if (LowMem::TrapExists (sysTrapPrefSetPreference))
	{
		PrefSetPreference (prefAutoOffDuration, 0);
	}

	// Install a 'pose' feature so that people can tell if they're running
	// under the emulator or not.  We *had* added a HostControl function
	// for this, but people can call that only under Poser or under 3.0
	// and later ROMs.  Which means that there's no way to tell under 2.0
	// and earlier ROMs when running on an actual device.
	//
	// Note that we do this here instead of in a tailpatch to FtrInit because
	// of a goofy inter-dependency.  FtrInit is called early in the boot process
	// before a valid stack has been allocated and switched to. During this time,
	// the ROM is using a range of memory that happens to reside in a free
	// memory chunk.  Routines in MetaMemory know about this and allow the use
	// of this unallocate range of memory.  However, in order to work, they need
	// to know the current stack pointer value.  If we were to call FtrSet from
	// the FtrInit tailpatch, the stack pointer will point to the stack set up
	// by the ATrap object, not the faux stack in use at the time of FtrInit.
	// Thus, the MetaMemory routines get confused and end up getting in a state
	// where accesses to the temporary OS stack are flagged as invalid.  Hence,
	// we defer adding these Features until much later in the boot process (here).

	FtrSet ('pose', 0, 0);

#if defined (_WINDOWS)
	// If we're listening on a socket, install the 'gdbS' feature.  The
	// existance of this feature causes programs written with the prc tools
	// to enter the debugger when they're launched.

	if (Socket::IsConnected ())
	{
		FtrSet ('gdbS', 0, 0x12BEEF34);
	}
#endif

	// Install the HotSync user-name.
	string	userName = gPrefs->GetUserName ();
	SetHotSyncUserName (userName.c_str ());

	// Auto-load any files in the Autoload[Foo] directories.

	PrvAutoload ();
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	HtalLibHeadpatch::HtalLibSendReply
 *
 * DESCRIPTION:	Ohhh...I'm going to Programmer Hell for this one...
 *				We call DlkDispatchRequest to install the user name in
 *				our UIInitialize patch.  DlkDispatchRequest will
 *				eventually call HtalLibSendReply to return a result
 *				code.  Well, we haven't fired up the Htal library, and
 *				wouldn't want it to send a response even if we had.
 *				Therefore, I'll subvert the whole process by setting
 *				the HTAL library refNum passed in to the Desktop Link
 *				Manager to an invalid value.  I'll look for this
 *				value in the SysTrap handling code and no-op the call
 *				by calling this stub.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

CallROMType HtalLibHeadpatch::HtalLibSendReply (void)
{
	m68k_dreg (regs, 0) = errNone;

	return kSkipROM;
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	PrvToString
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

string PrvToString (uaecptr s)
{
	string	result;

	size_t	sLen = uae_strlen (s);
	if (sLen > 0)
	{
		result.resize (sLen);
		uae_strcpy (&result[0], s);
	}

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	PrvAutoload
 *
 * DESCRIPTION:	Install the files in the various Autoload directories.
 *				If there are any designated to be run, pick on to run.
 *				If the emulator should exit when the picked application
 *				quits, schedule it to do so.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void PrvAutoload (void)
{
	LoadableList	fileList;
	long			numAutoloads;
	long			numAutoruns;
	long			numAutorunAndQuits;

	// Get the list of files to load into "fileList".  Keep track of what
	// parts of the list contain what files.

	Platform::GetLoadableFileList (Platform::GetString (kStr_Autoload), fileList);
	numAutoloads = fileList.size ();

	Platform::GetLoadableFileList (Platform::GetString (kStr_Autorun), fileList);
	numAutoruns = fileList.size () - numAutoloads;

	Platform::GetLoadableFileList (Platform::GetString (kStr_AutorunAndQuit), fileList);
	numAutorunAndQuits = fileList.size () - numAutoruns - numAutoloads;


	// Load all the files in one blow.

	::LoadPalmFileList (fileList);


	// Grovel over the list, looking for one that needs to be auto-started.
	// Also keep track of whether or no Poser needs to quit when the app quits.

	for (long ii = fileList.size () - 1; ii >= numAutoloads; --ii)
	{
		UInt	dbAttrs;
		ULong	type, creator;

		Err	err = ::DmDatabaseInfo (
					fileList[ii].cardNo,
					fileList[ii].dbID,
					NULL,				/*name*/
					&dbAttrs,
					NULL,				/*version*/
					NULL,				/*create date*/
					NULL,				/*modDate*/
					NULL,				/*backup date*/
					NULL,				/*modNum*/
					NULL,				/*appInfoID*/
					NULL,				/*sortInfoID*/
					&type,
					&creator);

		if (::IsExecutable (type, creator, dbAttrs))
		{
			Patches::SetSwitchApp (fileList[ii].cardNo, fileList[ii].dbID);

			if (ii >= numAutoloads + numAutoruns)
			{
				Patches::QuitOnAppExit (true);
			}
		}
	}
}

#include "PalmPackPop.h"
