/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "Settings.h"

#include "windowsx.h"			// ComboBox_SelectString

#include "CPU_MT.h"				// CPU_setsounds
#include "CPU_REG.h"			// Emulator
#include "Emulator.h"			// gInstance
#include "Gremlins.h"			// Gremlins_IsOn
#include "LCD.h"				// CenterWindow
#include "Registry.h"			// GetRegistryDword, etc.
#include "Platform.h"			// StTempString
#include "TrapPatches.h"		// SetHotSyncUserName
#include "Strings.r.h"			// kStr_NoPort

#include "resource.h"			// IDD_PROPERTIES


// ======================================================================
//	Globals and constants
// ======================================================================

TSettings	gSettings;
TSettings	gTempSettings;

TSettings*	gHostPrefs = &gSettings;


// ===========================================================================
//		 TSettings
// ===========================================================================

deque<string>	TSettings::fgCachedPorts;


/***********************************************************************
 *
 * FUNCTION:	TSettings constructor
 *
 * DESCRIPTION:	Loads the settings from the registry.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

TSettings::TSettings (void)
{
	this->Load ();
}


/***********************************************************************
 *
 * FUNCTION:	TSettings destructor
 *
 * DESCRIPTION:	Saves the current settings to the registry.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

TSettings::~TSettings (void)
{
	this->Save ();
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::Load
 *
 * DESCRIPTION:	Loads the settings from the registry.  Validates them
 *				and cleans up obsolete settings.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void TSettings::Load (void)
{
	DWORD	x;
	BOOL	b;
	string	s;

	// Migrate over old Registry entries.

	if (Registry::GetBool ("Realistic", b))
	{
		if (b)
		{
			fScreenBackground = kGrayBackground;
			fWindowFrame = kCaseFrame;
		}
		else
		{
			fScreenBackground = kWhiteBackground;
			fWindowFrame = kNoFrame;
		}
	}

	Registry::GetBool ("CheckLowMemoryAccess", fReportLowMemoryAccess);
	Registry::GetBool ("CheckGlobalMemoryAccess", fReportSystemGlobalAccess);
	Registry::GetBool ("CheckScreenMemoryAccess", fReportScreenAccess);
	Registry::GetBool ("CheckRegisterMemoryAccess", fReportHardwareRegisterAccess);
	Registry::GetBool ("CheckValidPC", fReportInvalidPC);
	Registry::GetBool ("CleanStack", fFillStack);

	// Remove old Registry entries

	Registry::RemoveSlot ("Realistic");
	Registry::RemoveSlot ("Sounds");
	Registry::RemoveSlot ("FPS");
	Registry::RemoveSlot ("DebugAtStartup");
	Registry::RemoveSlot ("CheckLowMemoryAccess");
	Registry::RemoveSlot ("CheckGlobalMemoryAccess");
	Registry::RemoveSlot ("CheckScreenMemoryAccess");
	Registry::RemoveSlot ("CheckRegisterMemoryAccess");
	Registry::RemoveSlot ("CheckHeapBlockAccess");
	Registry::RemoveSlot ("CheckValidPC");
	Registry::RemoveSlot ("CleanStack");
	Registry::RemoveSlot ("GremlinLogEvents");
	Registry::RemoveSlot ("GremlinLogTraps");
	Registry::RemoveSlot ("GremlinLogErrors");

	for (int ii = MRU_COUNT; ii < 100; ++ii)
	{
		char	buffer[100];
		sprintf (buffer, "PrcMru.%d", ii);
		Registry::RemoveSlot (buffer);
	}

	// Get the current supported settings.

	Registry::GetInt	("RamSize",						fLastConfiguration.fRAMSize);
	Registry::GetInt	("DeviceType",					fLastConfiguration.fDeviceType);
	Registry::GetInt	("Country",						fLastConfiguration.fSilkscreenType);
	Registry::GetInt	("Scale",						fScale);
	Registry::GetString	("CommPort",					fCommPort);
	Registry::GetInt	("CloseAction",					fCloseAction);
	Registry::GetInt	("ScreenBackground",			fScreenBackground);

	Registry::GetInt	("WindowX",						fWindowLocation.h);
	Registry::GetInt	("WindowY",						fWindowLocation.v);

	Registry::GetInt	("GCW_X",						fGCWLocation.h);
	Registry::GetInt	("GCW_Y",						fGCWLocation.v);

	Registry::GetBool	("RedirectNetLib",				fRedirectNetLib);

	Registry::GetBool	("fReportCorruptedHeap",			fReportCorruptedHeap);
	Registry::GetBool	("fReportFreeChunkAccess",			fReportFreeChunkAccess);
	Registry::GetBool	("fReportHardwareRegisterAccess",	fReportHardwareRegisterAccess);
	Registry::GetBool	("fReportInvalidPC",				fReportInvalidPC);
	Registry::GetBool	("fReportLowMemoryAccess",			fReportLowMemoryAccess);
	Registry::GetBool	("fReportLowStackAccess",			fReportLowStackAccess);
	Registry::GetBool	("fReportMemMgrDataAccess",			fReportMemMgrDataAccess);
	Registry::GetBool	("fReportMemMgrSemaphore",			fReportMemMgrSemaphore);
	Registry::GetBool	("fReportScreenAccess",				fReportScreenAccess);
	Registry::GetBool	("fReportStackAlmostOverflow",		fReportStackAlmostOverflow);
	Registry::GetBool	("fReportStackOverflow",			fReportStackOverflow);
	Registry::GetBool	("fReportStorageHeapAccess",		fReportStorageHeapAccess);
	Registry::GetBool	("fReportSysFatalAlert",			fReportSysFatalAlert);
	Registry::GetBool	("fReportSystemGlobalAccess",		fReportSystemGlobalAccess);
	Registry::GetBool	("fReportUnhandledException",		fReportUnhandledException);
	Registry::GetBool	("fReportUnimplementedTrap",		fReportUnimplementedTrap);
	Registry::GetBool	("fReportUninitializedChunkAccess",	fReportUninitializedChunkAccess);
	Registry::GetBool	("fReportUninitializedStackAccess",	fReportUninitializedStackAccess);
	Registry::GetBool	("fReportUnlockedChunkAccess",		fReportUnlockedChunkAccess);

	Registry::GetInt	("LogErrorMessages",			fLogErrorMessages);
	Registry::GetInt	("LogWarningMessages",			fLogWarningMessages);
	Registry::GetInt	("LogGremlins",					fLogGremlins);
	Registry::GetInt	("LogCPUOpcodes",				fLogCPUOpcodes);
	Registry::GetInt	("LogEnqueuedEvents",			fLogEnqueuedEvents);
	Registry::GetInt	("LogDequeuedEvents",			fLogDequeuedEvents);
	Registry::GetInt	("LogSystemCalls",				fLogSystemCalls);
	Registry::GetInt	("LogApplicationFunctions",		fLogApplicationCalls);
	Registry::GetInt	("LogSerial",					fLogSerial);
	Registry::GetInt	("LogSerialData",				fLogSerialData);
	Registry::GetInt	("LogNetLib",					fLogNetLib);
	Registry::GetInt	("LogNetLibData",				fLogNetLibData);
	Registry::GetInt	("LogExgMgr",					fLogExgMgr);
	Registry::GetInt	("LogExgMgrData",				fLogExgMgrData);
	Registry::GetInt	("LogHLDebugger",				fLogHLDebugger);
	Registry::GetInt	("LogHLDebuggerData",			fLogHLDebuggerData);
	Registry::GetInt	("LogLLDebugger",				fLogLLDebugger);
	Registry::GetInt	("LogLLDebuggerData",			fLogLLDebuggerData);

	Registry::GetBool	("ReportMemoryLeaks",			fReportMemoryLeaks);
	Registry::GetBool	("ReportLockedRecords",			fReportLockedRecords);

	Registry::GetBool	("FillNewBlocks",				fFillNewBlocks);
	Registry::GetBool	("FillDisposedBlocks",			fFillDisposedBlocks);
	Registry::GetBool	("FillStack",					fFillStack);

	Registry::GetInt	("GremlinNumber",				fGremlinInfo.fNumber);
	Registry::GetInt	("GremlinSteps",				fGremlinInfo.fSteps);
	Registry::GetBool	("GremlinSkipWarnings",			fGremlinInfo.fContinuePastWarnings);

	Registry::GetString	("UserName",					fUserName);

	if (Registry::GetString ("PrcPath", s))
		fLastPRCFile = FileReference (s);

	if (Registry::GetString ("RamFile", s))
		fLastRAMFile = FileReference (s);

	if (Registry::GetString ("RomFile", s))
		fLastConfiguration.fROMFile = FileReference (s);

	ii = 0;
	char buf[20];
	while (1)
	{
		sprintf (buf, "PrcMru.%d", ii++);
		if (!Registry::GetString (buf, s))
			break;

		fPRC_MRU.push_back (FileReference (s));
	}

	ii = 0;
	while (1)
	{
		sprintf (buf, "RamMru.%d", ii++);
		if (!Registry::GetString (buf, s))
			break;

		fRAM_MRU.push_back (FileReference (s));
	}

	// Validate the settings.

	this->ValidateSettings ();

		// Make sure fCommPort is valid.
	// !!! (Should figure out a way to put this into ValidateSettings,
	// but right now, the call to GetIndCommPort is messing us up).

	ii = 0;
	while (1)
	{
		s = TSettings::GetIndCommPort (ii);
		if (s.empty ())
		{
			fCommPort.erase ();	// No match, default to none.
			break;
		}

		if (s == fCommPort)
		{
			break;		// Match...break out.
		}

		++ii;
	};

	// Copy the "standard" settings into the "current" settings
	// (assumes that Gremlins is not running at the time...)

	this->UseLoggingOptions (kNormalLogging);
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::Save
 *
 * DESCRIPTION:	Save the settings to the registry.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void TSettings::Save (void)
{
	Registry::SetInt	("RamSize",						fLastConfiguration.fRAMSize);
	Registry::SetBool	("DeviceType",					fLastConfiguration.fDeviceType);
	Registry::SetInt	("Country",						fLastConfiguration.fSilkscreenType);
	Registry::SetInt	("Scale",						fScale);
	Registry::SetString	("CommPort",					fCommPort);
	Registry::SetInt	("CloseAction",					fCloseAction);
	Registry::SetInt	("ScreenBackground",			fScreenBackground);

	Registry::SetInt	("WindowX",						fWindowLocation.h);
	Registry::SetInt	("WindowY",						fWindowLocation.v);

	Registry::SetInt	("GCW_X",						fGCWLocation.h);
	Registry::SetInt	("GCW_Y",						fGCWLocation.v);

	Registry::SetBool	("RedirectNetLib",				fRedirectNetLib);

	Registry::SetBool	("fReportCorruptedHeap",			fReportCorruptedHeap);
	Registry::SetBool	("fReportFreeChunkAccess",			fReportFreeChunkAccess);
	Registry::SetBool	("fReportHardwareRegisterAccess",	fReportHardwareRegisterAccess);
	Registry::SetBool	("fReportInvalidPC",				fReportInvalidPC);
	Registry::SetBool	("fReportLowMemoryAccess",			fReportLowMemoryAccess);
	Registry::SetBool	("fReportLowStackAccess",			fReportLowStackAccess);
	Registry::SetBool	("fReportMemMgrDataAccess",			fReportMemMgrDataAccess);
	Registry::SetBool	("fReportMemMgrSemaphore",			fReportMemMgrSemaphore);
	Registry::SetBool	("fReportScreenAccess",				fReportScreenAccess);
	Registry::SetBool	("fReportStackAlmostOverflow",		fReportStackAlmostOverflow);
	Registry::SetBool	("fReportStackOverflow",			fReportStackOverflow);
	Registry::SetBool	("fReportStorageHeapAccess",		fReportStorageHeapAccess);
	Registry::SetBool	("fReportSysFatalAlert",			fReportSysFatalAlert);
	Registry::SetBool	("fReportSystemGlobalAccess",		fReportSystemGlobalAccess);
	Registry::SetBool	("fReportUnhandledException",		fReportUnhandledException);
	Registry::SetBool	("fReportUnimplementedTrap",		fReportUnimplementedTrap);
	Registry::SetBool	("fReportUninitializedChunkAccess",	fReportUninitializedChunkAccess);
	Registry::SetBool	("fReportUninitializedStackAccess",	fReportUninitializedStackAccess);
	Registry::SetBool	("fReportUnlockedChunkAccess",		fReportUnlockedChunkAccess);

	Registry::SetInt	("LogErrorMessages",			fLogErrorMessages);
	Registry::SetInt	("LogWarningMessages",			fLogWarningMessages);
	Registry::SetInt	("LogGremlins",					fLogGremlins);
	Registry::SetInt	("LogCPUOpcodes",				fLogCPUOpcodes);
	Registry::SetInt	("LogEnqueuedEvents",			fLogEnqueuedEvents);
	Registry::SetInt	("LogDequeuedEvents",			fLogDequeuedEvents);
	Registry::SetInt	("LogSystemCalls",				fLogSystemCalls);
	Registry::SetInt	("LogApplicationFunctions",		fLogApplicationCalls);
	Registry::SetInt	("LogSerial",					fLogSerial);
	Registry::SetInt	("LogSerialData",				fLogSerialData);
	Registry::SetInt	("LogNetLib",					fLogNetLib);
	Registry::SetInt	("LogNetLibData",				fLogNetLibData);
	Registry::SetInt	("LogExgMgr",					fLogExgMgr);
	Registry::SetInt	("LogExgMgrData",				fLogExgMgrData);
	Registry::SetInt	("LogHLDebugger",				fLogHLDebugger);
	Registry::SetInt	("LogHLDebuggerData",			fLogHLDebuggerData);
	Registry::SetInt	("LogLLDebugger",				fLogLLDebugger);
	Registry::SetInt	("LogLLDebuggerData",			fLogLLDebuggerData);

	Registry::SetBool	("ReportMemoryLeaks",			fReportMemoryLeaks);
	Registry::SetBool	("ReportLockedRecords",			fReportLockedRecords);

	Registry::SetBool	("FillNewBlocks",				fFillNewBlocks);
	Registry::SetBool	("FillDisposedBlocks",			fFillDisposedBlocks);
	Registry::SetBool	("FillStack",					fFillStack);

	Registry::SetInt	("GremlinNumber",				fGremlinInfo.fNumber);
	Registry::SetInt	("GremlinSteps",				fGremlinInfo.fSteps);
	Registry::SetBool	("GremlinSkipWarnings",			fGremlinInfo.fContinuePastWarnings);

	Registry::SetString ("PrcPath",						fLastPRCFile.GetFilePath ());
	Registry::SetString ("RamFile",						fLastRAMFile.GetFilePath ());
	Registry::SetString ("RomFile",						fLastConfiguration.fROMFile.GetFilePath ());

	Registry::SetString	("UserName",					fUserName);

	deque<FileReference>::iterator	iter;
	int								ii;

	for (iter = fPRC_MRU.begin(), ii = 0; iter != fPRC_MRU.end (); ++iter, ++ii)
	{
		if ((*iter).IsSpecified ())
		{
			char buf[20];
			sprintf (buf, "PrcMru.%d", ii);
			Registry::SetString (buf, (*iter).GetFilePath ());
		}
	}

	for (iter = fRAM_MRU.begin(), ii = 0; iter != fRAM_MRU.end (); ++iter, ++ii)
	{
		if ((*iter).IsSpecified ())
		{
			char buf[20];
			sprintf (buf, "RamMru.%d", ii);
			Registry::SetString (buf, (*iter).GetFilePath ());
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::EditProperties
 *
 * DESCRIPTION:	Display a dialog for editting the general properties.
 *				These include serial port for emulation, screen size/
 *				appearance, save-on-quit characteristics, etc.
 *
 * PARAMETERS:	hwnd - parent window
 *
 * RETURNED:	True if the emulator window needs to be updated.
 *
 ***********************************************************************/

Bool TSettings::EditProperties (HWND hwnd)
{
	Bool	result = false;

	gTempSettings = *this;

	int	choice = ::DialogBox (	gInstance, MAKEINTRESOURCE (IDD_PROPERTIES),
								hwnd, EditPropertiesDlgProcSt);

	if (choice == IDOK)
	{
		if (this->fScale != gTempSettings.fScale)
			result = true;

		if (this->fScreenBackground != gTempSettings.fScreenBackground)
			result = true;

		// Re-establish the serial port.

		if (Platform::SerialPortOpen ())
		{
			Platform::CloseSerialPort ();
			Platform::OpenSerialPort ();
		}

		*this = gTempSettings;
		this->Save ();
	}

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::EditPropertiesDlgProcSt
 *
 * DESCRIPTION:	Static function for the Edit Properties dialog.  Calls
 *				the method for the settings object we're editing.
 *
 * PARAMETERS:	Standard DlgProc parameters
 *
 * RETURNED:	Standard DlgProc result.
 *
 ***********************************************************************/

BOOL CALLBACK TSettings::EditPropertiesDlgProcSt (	HWND hwnd,
													UINT msg,
													WPARAM wparam,
													LPARAM lparam)
{
	return gTempSettings.EditPropertiesDlgProc (hwnd, msg, wparam, lparam);
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::EditPropertiesDlgProc
 *
 * DESCRIPTION:	DlgProc for the Edit Properties dialog.
 *
 * PARAMETERS:	Standard DlgProc parameters
 *
 * RETURNED:	Standard DlgProc result.
 *
 ***********************************************************************/

BOOL TSettings::EditPropertiesDlgProc (	HWND hwnd,
										UINT msg,
										WPARAM wparam,
										LPARAM lparam)
{
	string	noPort (Platform::GetString (kStr_NoPort));

	switch (msg)
	{
		case WM_INITDIALOG:
		{
			::CenterWindow (hwnd);

			HWND	portCombo = GetDlgItem (hwnd, IDC_PORT);

			// Add the comm port strings to the port combo box.

			TSettings::PopulatePortMenu (portCombo);

			// Add the "no port" menu item.

			ComboBox_AddString (portCombo, noPort.c_str ());

			// Select the menu item for the chosen port.

			string	portName (this->GetCommPort ());
			if (portName.empty ())
				portName = noPort;

			ComboBox_SelectString (portCombo, -1, portName.c_str ());

			// Set NetLib redirection option

			if (this->RedirectNetLib ())
				::SendDlgItemMessage (hwnd, IDC_NETLIB, BM_SETCHECK, BST_CHECKED, 0);
			else
				::SendDlgItemMessage (hwnd, IDC_NETLIB, BM_SETCHECK, BST_UNCHECKED, 0);

			// Select either "Gray Background" or "White background"

			if (this->GrayBackground ())
				::SendDlgItemMessage (hwnd, IDC_GRAYBACKGROUND, BM_SETCHECK, BST_CHECKED, 0);
			else
				::SendDlgItemMessage (hwnd, IDC_WHITEBACKGROUND, BM_SETCHECK, BST_CHECKED, 0);

			// Select either "Single Size" or "Double Size".

			if (this->SingleSize ())
				::SendDlgItemMessage (hwnd, IDC_SINGLESIZE, BM_SETCHECK, BST_CHECKED, 0);
			else
				::SendDlgItemMessage (hwnd, IDC_DOUBLESIZE, BM_SETCHECK, BST_CHECKED, 0);

			// Select "Always Save", "Ask Save", or "Never Save".

			if (this->SaveAlways ())
				::SendDlgItemMessage (hwnd, IDC_SAVEALWAYS, BM_SETCHECK, BST_CHECKED, 0);
			else if (this->SaveAsk ())
				::SendDlgItemMessage (hwnd, IDC_SAVEASK, BM_SETCHECK, BST_CHECKED, 0);
			else
				::SendDlgItemMessage (hwnd, IDC_SAVENEVER, BM_SETCHECK, BST_CHECKED, 0);

			// Install the user name.

			HWND	userNameItem = ::GetDlgItem (hwnd, IDC_USERNAME);
			::SetWindowText (userNameItem, fUserName.c_str ());

			break;
		}

		case WM_COMMAND:
		{
			switch (LOWORD (wparam))
			{
				case IDOK:
				{
					// Get the selected port

					HWND	portCombo = ::GetDlgItem (hwnd, IDC_PORT);
					LONG	index = ComboBox_GetCurSel (portCombo);
					char	tempBuffer[256];
					ComboBox_GetLBText (portCombo, index, tempBuffer);
					fCommPort = tempBuffer;

					if (noPort == fCommPort)
						fCommPort.erase ();

					// Get the NetLib redirection setting

					if (::SendDlgItemMessage (hwnd, IDC_NETLIB, BM_GETCHECK, 0, 0) == BST_CHECKED)
						fRedirectNetLib = true;
					else
						fRedirectNetLib = false;

					// Get the background setting

					if (::SendDlgItemMessage (hwnd, IDC_GRAYBACKGROUND, BM_GETCHECK, 0, 0) == BST_CHECKED)
						fScreenBackground = kGrayBackground;
					else
						fScreenBackground = kWhiteBackground;

					// Get the display size setting

					if (::SendDlgItemMessage (hwnd, IDC_DOUBLESIZE, BM_GETCHECK, 0, 0) == BST_CHECKED)
						fScale = 2;
					else
						fScale = 1;

					// Get the "close action"

					if (::SendDlgItemMessage (hwnd, IDC_SAVEALWAYS, BM_GETCHECK, 0, 0) == BST_CHECKED)
						fCloseAction = kSaveAlways;
					else if (::SendDlgItemMessage (hwnd, IDC_SAVEASK, BM_GETCHECK, 0, 0) == BST_CHECKED)
						fCloseAction = kSaveAsk;
					else
						fCloseAction = kSaveNever;

					// Get the user name.  Install it in the ROM.

					HWND	userNameItem = ::GetDlgItem (hwnd, IDC_USERNAME);
					::GetWindowText (userNameItem, tempBuffer, sizeof (tempBuffer));
					fUserName = tempBuffer;

					CPUStopper	stopper (LCD_GetWindow (), "set the user name");

					if (stopper.Stopped ())
					{
						::SetHotSyncUserName (fUserName.c_str ());
					}
				}

					// Fall thru

				case IDCANCEL:
					::EndDialog (hwnd, LOWORD (wparam));
					break;
			}
			break;
		}

		default:
			return FALSE;
	}

	return TRUE;
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::EditLoggingOptions
 *
 * DESCRIPTION:	Present a dialog box so that the use can edit the set
 *				of logging options.  This dialog contains two pages:
 *				one set of options to be used during normal use, and
 *				another set of options to take effect when a Gremlin
 *				is running.
 *
 * PARAMETERS:	hwnd - parent window
 *
 *				whichPanel - open the dialog to the given panel.  The
 *					panel number is 1-based.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void TSettings::EditLoggingOptions (HWND hwnd, int whichPanel)
{
	gTempSettings = *this;

	PROPSHEETPAGE	psp[2];
	::ZeroMemory (psp, sizeof (psp));

	psp[0].dwSize		= sizeof (PROPSHEETPAGE);
	psp[0].dwFlags		= PSP_DEFAULT | PSP_USETITLE;
	psp[0].hInstance	= gInstance;
	psp[0].pszTemplate	= MAKEINTRESOURCE (IDD_LOGGINGOPTIONS);
	psp[0].pszTitle		= MAKEINTRESOURCE (IDS_LOGGINGNORMAL);
	psp[0].pfnDlgProc	= &EditLoggingOptionsDlgProcSt1;
	psp[0].lParam		= (LPARAM) this;

	psp[1].dwSize		= sizeof (PROPSHEETPAGE);
	psp[1].dwFlags		= PSP_DEFAULT | PSP_USETITLE;
	psp[1].hInstance	= gInstance;
	psp[1].pszTemplate	= MAKEINTRESOURCE (IDD_LOGGINGOPTIONS);
	psp[1].pszTitle		= MAKEINTRESOURCE (IDS_LOGGINGGREMLINS);
	psp[1].pfnDlgProc	= &EditLoggingOptionsDlgProcSt2;
	psp[1].lParam		= (LPARAM) this;

	PROPSHEETHEADER psh;
	::ZeroMemory (&psh, sizeof (psh));

	psh.dwSize			= sizeof (psh);
	psh.dwFlags			= PSH_PROPSHEETPAGE | PSH_NOAPPLYNOW;
	psh.hwndParent		= hwnd;
	psh.hInstance		= gInstance;
	psh.pszCaption		= MAKEINTRESOURCE (IDS_LOGGINGOPTIONS);
	psh.nPages			= countof (psp);
	psh.nStartPage		= whichPanel - 1;
	psh.ppsp			= psp;

	(void) ::PropertySheet (&psh);

	if (::Gremlins_IsOn ())
	{
		this->UseLoggingOptions (kGremlinLogging);
	}
	else
	{
		this->UseLoggingOptions (kNormalLogging);
	}
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::EditLoggingOptionsDlgProcSt1
 *
 * DESCRIPTION:	DlgProc for page 1 of the Logging Options.  Calls a
 *				common DlgProc, passing in a parameter identifying
 *				which page is in use.
 *
 * PARAMETERS:	Standard DlgProc parameters
 *
 * RETURNED:	Standard DlgProc result.
 *
 ***********************************************************************/

BOOL CALLBACK TSettings::EditLoggingOptionsDlgProcSt1 (	HWND hwnd,
														UINT msg,
														WPARAM wparam,
														LPARAM lparam)
{
	return gTempSettings.EditLoggingOptionsDlgProc (hwnd, msg, wparam, lparam, 1);
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::EditLoggingOptionsDlgProcSt2
 *
 * DESCRIPTION:	DlgProc for page 1 of the Logging Options.  Calls a
 *				common DlgProc, passing in a parameter identifying
 *				which page is in use.
 *
 * PARAMETERS:	Standard DlgProc parameters
 *
 * RETURNED:	Standard DlgProc result.
 *
 ***********************************************************************/


BOOL CALLBACK TSettings::EditLoggingOptionsDlgProcSt2 (	HWND hwnd,
														UINT msg,
														WPARAM wparam,
														LPARAM lparam)
{
	return gTempSettings.EditLoggingOptionsDlgProc (hwnd, msg, wparam, lparam, 2);
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::EditLoggingOptionsDlgProc
 *
 * DESCRIPTION:	DlgProc to manage the pages of the Logging Options
 *				dialog.  Sets the checkboxes based on which page is
 *				currently active.
 *
 * PARAMETERS:	Standard DlgProc parameters
 *
 * RETURNED:	Standard DlgProc result.
 *
 ***********************************************************************/

BOOL TSettings::EditLoggingOptionsDlgProc (	HWND hDlg,
											UINT msg,
											WPARAM wParam,
											LPARAM lParam,
											int whichPage)
{
	// Except in response to the WM_INITDIALOG message, the dialog box procedure
	// should return nonzero if it processes the message, and zero if it does not.
	// In response to a WM_INITDIALOG message, the dialog box procedure should
	// return zero if it calls the SetFocus function to set the focus to one of
	// the controls in the dialog box. Otherwise, it should return nonzero, in
	// which case the system sets the focus to the first control in the dialog
	// box that can be given the focus.

	static TSettings*	original;

	int		bitMask = whichPage == 1 ? kNormalLogging : kGremlinLogging;

	switch (msg)
	{
		case WM_INITDIALOG:
		{
			original = (TSettings*) ((PROPSHEETPAGE*) lParam)->lParam;
			return TRUE;
		}

		case WM_NOTIFY:
		{
			switch (((NMHDR*) lParam)->code) 
			{
				case PSN_SETACTIVE:
				{
					// Returns zero to accept the activation, or -1 to activate 
					// the next or the previous page (depending on whether the 
					// user clicked the Next or Back button). To set the activation 
					// to a particular page, return the resource identifier of the page.

					UpdateCheckBox (hDlg, IDC_LOG_ERRORS,						fLogErrorMessages,		bitMask);
					UpdateCheckBox (hDlg, IDC_LOG_WARNINGS,						fLogWarningMessages,	bitMask);
					UpdateCheckBox (hDlg, IDC_LOG_GREMLINS,						fLogGremlins,			bitMask);
					UpdateCheckBox (hDlg, IDC_LOG_OPCODES,						fLogCPUOpcodes,			bitMask);
					UpdateCheckBox (hDlg, IDC_LOG_POSTED_EVENTS,				fLogEnqueuedEvents,		bitMask);
					UpdateCheckBox (hDlg, IDC_LOG_RECEIVED_EVENTS,				fLogDequeuedEvents,		bitMask);
					UpdateCheckBox (hDlg, IDC_LOG_SYSTEM_CALLS,					fLogSystemCalls,		bitMask);
					UpdateCheckBox (hDlg, IDC_LOG_APPLICATION_CALLS,			fLogApplicationCalls,	bitMask);
					UpdateCheckBox (hDlg, IDC_LOG_SERIAL_ACTIVITY,				fLogSerial,				bitMask);
					UpdateCheckBox (hDlg, IDC_LOG_SERIAL_DATA,					fLogSerialData,			bitMask);
					UpdateCheckBox (hDlg, IDC_LOG_NETLIB_ACTIVITY,				fLogNetLib,				bitMask);
					UpdateCheckBox (hDlg, IDC_LOG_NETLIB_DATA,					fLogNetLibData,			bitMask);
					UpdateCheckBox (hDlg, IDC_LOG_EXGMGR_ACTIVITY,				fLogExgMgr,				bitMask);
					UpdateCheckBox (hDlg, IDC_LOG_EXGMGR_DATA,					fLogExgMgrData,			bitMask);
					UpdateCheckBox (hDlg, IDC_LOG_HIGHLEVEL_DEBUGGER_ACTIVITY,	fLogHLDebugger,			bitMask);
					UpdateCheckBox (hDlg, IDC_LOG_HIGHLEVEL_DEBUGGER_DATA,		fLogHLDebuggerData,		bitMask);
					UpdateCheckBox (hDlg, IDC_LOG_LOWLEVEL_DEBUGGER_ACTIVITY,	fLogLLDebugger,			bitMask);
					UpdateCheckBox (hDlg, IDC_LOG_LOWLEVEL_DEBUGGER_DATA,		fLogLLDebuggerData,		bitMask);

					// Disable unsupported options.

					::EnableWindow (GetDlgItem (hDlg, IDC_LOG_ERRORS), false);
					::EnableWindow (GetDlgItem (hDlg, IDC_LOG_OPCODES), false);
					::EnableWindow (GetDlgItem (hDlg, IDC_LOG_APPLICATION_CALLS), false);
					::EnableWindow (GetDlgItem (hDlg, IDC_LOG_EXGMGR_ACTIVITY), false);
					::EnableWindow (GetDlgItem (hDlg, IDC_LOG_EXGMGR_DATA), false);
					
					::SetWindowLong (hDlg, DWL_MSGRESULT, 0);
					return TRUE;
				}

				case PSN_APPLY:
				{
					// Returns the PSNRET_INVALID_NOCHANGEPAGE value to prevent the
					// changes from taking effect and to return the focus to the page,
					// or the PSNRET_NOERROR value to accept the changes and allow the
					// property sheet to be destroyed.

					*original = *this;
					original->Save ();

					::SetWindowLong (hDlg, DWL_MSGRESULT, PSNRET_NOERROR);
					return TRUE;
				}

				case PSN_KILLACTIVE:
				{
					// Returns TRUE to prevent the page from losing the activation,
					// or FALSE to allow it.

					UpdatePreference (hDlg, IDC_LOG_ERRORS,							fLogErrorMessages,		bitMask);
					UpdatePreference (hDlg, IDC_LOG_WARNINGS,						fLogWarningMessages,	bitMask);
					UpdatePreference (hDlg, IDC_LOG_GREMLINS,						fLogGremlins,			bitMask);
					UpdatePreference (hDlg, IDC_LOG_OPCODES,						fLogCPUOpcodes,			bitMask);
					UpdatePreference (hDlg, IDC_LOG_POSTED_EVENTS,					fLogEnqueuedEvents,		bitMask);
					UpdatePreference (hDlg, IDC_LOG_RECEIVED_EVENTS,				fLogDequeuedEvents,		bitMask);
					UpdatePreference (hDlg, IDC_LOG_SYSTEM_CALLS,					fLogSystemCalls,		bitMask);
					UpdatePreference (hDlg, IDC_LOG_APPLICATION_CALLS,				fLogApplicationCalls,	bitMask);
					UpdatePreference (hDlg, IDC_LOG_SERIAL_ACTIVITY,				fLogSerial,				bitMask);
					UpdatePreference (hDlg, IDC_LOG_SERIAL_DATA,					fLogSerialData,			bitMask);
					UpdatePreference (hDlg, IDC_LOG_NETLIB_ACTIVITY,				fLogNetLib,				bitMask);
					UpdatePreference (hDlg, IDC_LOG_NETLIB_DATA,					fLogNetLibData,			bitMask);
					UpdatePreference (hDlg, IDC_LOG_EXGMGR_ACTIVITY,				fLogExgMgr,				bitMask);
					UpdatePreference (hDlg, IDC_LOG_EXGMGR_DATA,					fLogExgMgrData,			bitMask);
					UpdatePreference (hDlg, IDC_LOG_HIGHLEVEL_DEBUGGER_ACTIVITY,	fLogHLDebugger,			bitMask);
					UpdatePreference (hDlg, IDC_LOG_HIGHLEVEL_DEBUGGER_DATA,		fLogHLDebuggerData,		bitMask);
					UpdatePreference (hDlg, IDC_LOG_LOWLEVEL_DEBUGGER_ACTIVITY,		fLogLLDebugger,			bitMask);
					UpdatePreference (hDlg, IDC_LOG_LOWLEVEL_DEBUGGER_DATA,			fLogLLDebuggerData,		bitMask);

					::SetWindowLong (hDlg, DWL_MSGRESULT, FALSE);
					return TRUE;
				}

				case PSN_RESET:
				{
					// No return value.
					return TRUE;
				}
			}
		}
	}

	return FALSE;
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::EditDebugOptions
 *
 * DESCRIPTION:	Present a dialog for editting the so-called Debug
 *				Options.  These options are those that affect the
 *				display of messages reporting invalid program actions
 *				such as illegal memory accesses.
 *
 * PARAMETERS:	hwnd - parent window.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void TSettings::EditDebugOptions (HWND hwnd)
{
	gTempSettings = *this;

	int	choice = ::DialogBox (	gInstance, MAKEINTRESOURCE (IDD_DEBUGOPTIONS),
								hwnd, &EditDebugOptionsDlgProcSt);

	if (choice == IDOK)
	{
		*this = gTempSettings;
		this->Save ();
	}
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::EditDebugOptionsDlgProcSt
 *
 * DESCRIPTION:	Static function for the Debug Options dialog.  Calls
 *				the method for the settings object we're editing.
 *
 * PARAMETERS:	Standard DlgProc parameters.
 *
 * RETURNED:	Standard DlgProc result.
 *
 ***********************************************************************/

BOOL CALLBACK TSettings::EditDebugOptionsDlgProcSt (HWND hwnd,
													UINT msg,
													WPARAM wparam,
													LPARAM lparam)
{
	return gTempSettings.EditDebugOptionsDlgProc (hwnd, msg, wparam, lparam);
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::EditDebugOptionsDlgProc
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	Standard DlgProc parameters.
 *
 * RETURNED:	Standard DlgProc result.
 *
 ***********************************************************************/

BOOL TSettings::EditDebugOptionsDlgProc (	HWND hwnd,
											UINT msg,
											WPARAM wparam,
											LPARAM lparam)
{
	switch (msg)
	{
		case WM_INITDIALOG:
		{
			::CenterWindow (hwnd);

			UpdateCheckBox (hwnd, IDC_FREECHUNK,			fReportFreeChunkAccess);
			UpdateCheckBox (hwnd, IDC_REGISTERACCESS,		fReportHardwareRegisterAccess);
			UpdateCheckBox (hwnd, IDC_LOWMEMACCESS,			fReportLowMemoryAccess);
			UpdateCheckBox (hwnd, IDC_LOWSTACK,				fReportLowStackAccess);
			UpdateCheckBox (hwnd, IDC_MEMMGRDATA,			fReportMemMgrDataAccess);
			UpdateCheckBox (hwnd, IDC_MEMMGRSEMAPHORE,		fReportMemMgrSemaphore);
			UpdateCheckBox (hwnd, IDC_SCREENACCESS,			fReportScreenAccess);
			UpdateCheckBox (hwnd, IDC_STACKALMOSTOVERFLOW,	fReportStackAlmostOverflow);
			UpdateCheckBox (hwnd, IDC_STORAGEHEAP,			fReportStorageHeapAccess);
			UpdateCheckBox (hwnd, IDC_GLOBALSACCESS,		fReportSystemGlobalAccess);
			UpdateCheckBox (hwnd, IDC_UNINITCHUNK,			fReportUninitializedChunkAccess);
			UpdateCheckBox (hwnd, IDC_UNINITSTACK,			fReportUninitializedStackAccess);
			UpdateCheckBox (hwnd, IDC_UNLOCKEDCHUNK,		fReportUnlockedChunkAccess);

			// Disable unsupported options.

			::EnableWindow (GetDlgItem (hwnd, IDC_LOWSTACK), false);
			::EnableWindow (GetDlgItem (hwnd, IDC_STORAGEHEAP), false);
			::EnableWindow (GetDlgItem (hwnd, IDC_UNINITCHUNK), false);
			::EnableWindow (GetDlgItem (hwnd, IDC_UNINITSTACK), false);
			break;
		}

		case WM_COMMAND:
		{
			switch (LOWORD (wparam))
			{
				case IDOK:
					UpdatePreference (hwnd, IDC_FREECHUNK,				fReportFreeChunkAccess);
					UpdatePreference (hwnd, IDC_REGISTERACCESS,			fReportHardwareRegisterAccess);
					UpdatePreference (hwnd, IDC_LOWMEMACCESS,			fReportLowMemoryAccess);
					UpdatePreference (hwnd, IDC_LOWSTACK,				fReportLowStackAccess);
					UpdatePreference (hwnd, IDC_MEMMGRDATA,				fReportMemMgrDataAccess);
					UpdatePreference (hwnd, IDC_MEMMGRSEMAPHORE,		fReportMemMgrSemaphore);
					UpdatePreference (hwnd, IDC_SCREENACCESS,			fReportScreenAccess);
					UpdatePreference (hwnd, IDC_STACKALMOSTOVERFLOW,	fReportStackAlmostOverflow);
					UpdatePreference (hwnd, IDC_STORAGEHEAP,			fReportStorageHeapAccess);
					UpdatePreference (hwnd, IDC_GLOBALSACCESS,			fReportSystemGlobalAccess);
					UpdatePreference (hwnd, IDC_UNINITCHUNK,			fReportUninitializedChunkAccess);
					UpdatePreference (hwnd, IDC_UNINITSTACK,			fReportUninitializedStackAccess);
					UpdatePreference (hwnd, IDC_UNLOCKEDCHUNK,			fReportUnlockedChunkAccess);

					// Fall thru

				case IDCANCEL:
					::EndDialog (hwnd, LOWORD (wparam));
					break;
			}
			break;
		}

		default:
			return FALSE;
	}

	return TRUE;
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::GetLastPRCDir
 *
 * DESCRIPTION:	Returns the reference to the most recently loaded
 *				.prc file.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	The reference to the most recently loaded
 *				.prc file.  Weren't you listening?
 *
 ***********************************************************************/

FileReference& TSettings::GetLastPRCDir (void)
{
	return fLastPRCFile;
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::SetLastPRCDir
 *
 * DESCRIPTION:	Remember the given file reference to be used later when
 *				returning to the directory containing the most recently
 *				used PRC file.
 *
 * PARAMETERS:	ref - file reference for the most recently loaded .prc.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void TSettings::SetLastPRCDir (FileReference& ref)
{
	fLastPRCFile = ref;
	this->Save ();
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::PopulatePortMenu
 *
 * DESCRIPTION:	Add the list of serial ports to the given menu.
 *
 * PARAMETERS:	portCombo - menu to which to add the serial port names.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void TSettings::PopulatePortMenu (HWND portCombo)
{
	int		ii = 0;

	while (1)
	{
		string	portName (TSettings::GetIndCommPort (ii++));

		if (portName.empty ())
		{
			break;
		}

		if (!portName.empty ())
		{
			ComboBox_AddString (portCombo, portName.c_str ());
		}
	};
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::GetIndCommPort
 *
 * DESCRIPTION:	Get the name of the serial port indicated by the given
 *				index.  If the index is out of range, return an empty
 *				string.
 *
 * PARAMETERS:	index - zero-based index of the port for which we want
 *					then name.
 *
 * RETURNED:	The port's name, or an empty string.
 *
 ***********************************************************************/

string TSettings::GetIndCommPort (int index)
{
	if (fgCachedPorts.size () == 0)
	{
		TSettings::CreatePortNameCache ();
	}

	if (index >= fgCachedPorts.size ())
	{
		return string ();
	}

	return fgCachedPorts[index];
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::CreatePortNameCache
 *
 * DESCRIPTION:	By whatever means appropriate for this platform, find
 *				the set of available serial ports and store them in
 *				fgCachedPorts.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	Information is added to fgCachedPorts.
 *
 ***********************************************************************/

void TSettings::CreatePortNameCache (void)
{
	HKEY	curRootKey;

	if (::RegOpenKeyEx (HKEY_LOCAL_MACHINE, "Enum", 0, KEY_READ, &curRootKey) == ERROR_SUCCESS)
	{
		// Windows 95
		//
		// Gotta look for them the hard way. The NT way is not reliable. For instance,
		// using the NT way under NT will turn up COM1, COM2, and COM3 on my PC, but
		// under NT will only turn up COM1 and COM2.

		TSettings::SearchForPorts (HKEY_LOCAL_MACHINE, curRootKey);
	}
	else if (::RegOpenKeyEx (HKEY_LOCAL_MACHINE, "HARDWARE\\DEVICEMAP\\SERIALCOMM", 0, KEY_READ, &curRootKey) == ERROR_SUCCESS)
	{
		// Windows NT
		//
		// This one's easy: just iterate over HARDWARE\DEVICEMAP\SERIALCOMM to get
		// the names of all the ports.

		DWORD	ii = 0;
		while (1)
		{
			char	name[256], value[256];
			DWORD	nameSize = sizeof(name);
			DWORD	valueSize = sizeof(value);
			DWORD	type;

			if (::RegEnumValue (curRootKey, ii, name, &nameSize, NULL, &type, (BYTE*) value, &valueSize) != ERROR_SUCCESS)
			{
				break;
			}

			fgCachedPorts.push_back (string (value));
			++ii;
		}
	}

	::RegCloseKey (curRootKey);
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::SearchForPorts
 *
 * DESCRIPTION:	Search for serial ports, adding information on any
 *				found ports to fgCachedPorts.  This function is
 *				recursive; see details below.
 *
 * PARAMETERS:	HLM - hkey for HKEY_LOCAL_MACHINE
 *
 *				key - subkey to search.
 *
 * RETURNED:	Information is added to fgCachedPorts.
 *
 ***********************************************************************/

//	Look for serial port information.  This search consists of two steps:
//	first finding the hardware keys, and then using information there
//	to look up the software keys in order to determine which of those keys
//	are for serial ports.
//
//	To find the hardware keys, we have to do a recursive search over
//	HKEY_LOCAL_MACHINE\Enum\.  This tree contains the hardware keys either
//	found by Plug and Play, built-in enumerators, or installation programs.
//	We're looking for keys with sub-keys of "Class" that contain the
//	string "Ports".  When we find one of those, we can get the "display
//	name" from either "PortName" ("COM1") or "FRIENDLYNAAME" ("Communications
//	Port (COM1)").
//
//	After we have that information, we still have to find out of the port
//	we're looking at is for a serial port.  To determine that, we have to
//	look up the software key. Software keys are in HKEY_LOCAL_MACHINE\
//	System\CurrentControlSet\Services\Class\.  To determine which sub-key
//	to look up once we get there, we have to look back at the hardware key.
//	It contains a sub-key called "Driver" which contains the text of the
//	software sub-key to use.
//
//	Once we're at the right software sub-key, we have to look at the
//	"PortSubClass" sub-sub-key.  This key will contain 00 for parallel
//	ports, 01 for serial ports, and 02 for modem ports.
//
//	Whew!
//
//	By the way, a relatively "good" MSDN article talking about this is
//	"Installation and the Registry" in the "Windows 95 Device Driver Kit".
//	Just to nail things down, here are some sample registry entries for
//	you to look at:
//
//	Example Hardware Key:
//
//	[ HKEY_LOCAL_MACHINE\Enum\BIOS\*PNP0501\0C ]
//		Class			= "Ports"
//		PortName		= "COM1"
//		FRIENDLYNAME	= "Communications Port (COM1)"
//		Driver			= "Ports\0004"
//		...more...
//
//
//	Corresponding Software Key:
//
//	[ HKEY_LOCAL_MACHINE\System\CurrentControlSet\Services\Class\Ports\0004 ]
//		PortSubClass	= 01
//		...more...

void TSettings::SearchForPorts (HKEY HLM, HKEY key)
{
	char	buf[256];
	DWORD	n = sizeof(buf);
	DWORD	type;

	// See if we've got a "Class" key containing "Ports"

	if (::RegQueryValueEx (key, "Class", NULL, &type, (BYTE*) buf, &n) == ERROR_SUCCESS &&
		stricmp (buf, "Ports") == 0)
	{
		// We've got one. Now see if there's a "PortName" key. If there isn't
		// one, we assume we have a bogus hit. Otherwise, we use the value
		// as a comm port name.

		char	name[256];
		n = sizeof (name);

		if (::RegQueryValueEx (key, "PortName", NULL, &type, (BYTE*) name, &n) == ERROR_SUCCESS)
		{
			// We've got a "PortName" key/value. Now let's go for the "Driver"
			// key so that we can see if this port is a serial port. If we fail,
			// again bail out.

			n = sizeof (buf);

			if (::RegQueryValueEx (key, "Driver", NULL, &type, (BYTE*) buf, &n) == ERROR_SUCCESS)
			{
				// We now have the tag-end of the full software key. Catenate it
				// to the standard front-end, and go to the software key hierarchy.

				char kname[MAX_PATH];

				strcpy (kname, "System\\CurrentControlSet\\Services\\Class\\");
				strcat (kname, buf);

				HKEY softwareKey;

				if (::RegOpenKeyEx (HLM, kname, NULL, KEY_READ, &softwareKey) == ERROR_SUCCESS)
				{
					// Got this far...Now look up the "PortSubClass"

					n = sizeof(buf);
					if (::RegQueryValueEx (softwareKey, "PortSubClass", NULL, &type, (BYTE*) buf, &n) == ERROR_SUCCESS)
					{
						if (buf[0] == 1)
						{
							fgCachedPorts.push_back (string (name));
						}
					}

					::RegCloseKey (softwareKey);
				}
			}
		}
	}

	// No "Class" key at this level of the Registry hierarchy.
	// Iterate over the sub-keys.

	else
	{
		DWORD	ii = 0;
		while (1)
		{
			char		keyName[256], keyClass[256];
			DWORD		keyLen = sizeof(keyName);
			DWORD		classLen = sizeof(keyClass);
			FILETIME	ft;

			// Get the next key. If there are no more, break out of this loop.

			if (::RegEnumKeyEx (key, ii, keyName, &keyLen, NULL, keyClass, &classLen, &ft) != ERROR_SUCCESS)
			{
				break;
			}

			// Iterate over the next level by recursing.

			HKEY	k;
			if (::RegOpenKeyEx (key, keyName, NULL, KEY_READ, &k) == ERROR_SUCCESS)
			{
				SearchForPorts (HLM, k);
				::RegCloseKey (k);
			}

			++ii;
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::UpdateCheckBox
 *
 * DESCRIPTION:	Update the given checkbox based on the given value.
 *
 * PARAMETERS:	hwnd - dialog box handle.
 *
 *				id - id of the checkbox.
 *
 *				value - value determining the checkbox state.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void TSettings::UpdateCheckBox (HWND hwnd, UINT id, bool value)
{
	::SendDlgItemMessage (hwnd, id, BM_SETCHECK, (WPARAM) value, 0);
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::UpdatePreference
 *
 * DESCRIPTION:	Update the given value based on the setting of the
 *				given checkbox.
 *
 * PARAMETERS:	hwnd - dialog box handle.
 *
 *				id - id of the checkbox.
 *
 *				value - value to update.
 *
 * RETURNED:	"value" is updated in place.
 *
 ***********************************************************************/

void TSettings::UpdatePreference (HWND hwnd, UINT id, bool& value)
{
	value = ::SendDlgItemMessage (hwnd, id, BM_GETCHECK, 0, 0) == BST_CHECKED;
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::UpdateCheckBox
 *
 * DESCRIPTION:	Update the given checkbox based on the given bit in the
 *				given value.
 *
 * PARAMETERS:	hwnd - dialog box handle.
 *
 *				id - id of the checkbox.
 *
 *				value - value containing the bit to check
 *
 *				bitMask - mask determining the bit to check
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void TSettings::UpdateCheckBox (HWND hwnd, UINT id, uae_u8 value, int bitMask)
{
	::SendDlgItemMessage (hwnd, id, BM_SETCHECK, (WPARAM) ((value & bitMask) != 0), 0);
}


/***********************************************************************
 *
 * FUNCTION:	TSettings::UpdatePreference
 *
 * DESCRIPTION:	Update the given bit in the given value based on the
 *				setting of the given checkbox.
 *
 * PARAMETERS:	hwnd - dialog box handle.
 *
 *				id - id of the checkbox.
 *
 *				value - value to update.
 *
 *				bitmask - mask to use to update "value".
 *
 * RETURNED:	"value" is updated in place.
 *
 ***********************************************************************/

void TSettings::UpdatePreference (HWND hwnd, UINT id, uae_u8& value, int bitMask)
{
	if (::SendDlgItemMessage (hwnd, id, BM_GETCHECK, 0, 0) == BST_CHECKED)
		value |= bitMask;
	else
		value &= ~bitMask;
}


